import bpy

bl_info = {  
 "name": "Syb3R - Synthetic Benchmark Blender Tools",  
 "author": "Andreas Ley",  
 "version": (1, 0),  
 "blender": (2, 7, 6),
 "location": "Various",
 "category": "Syb3r",
 "description": "Collection of tools that aid the creation of scenes for synthetic benchmark datasets"
}


def syb3r_initializeRenderLayers():
    scene = bpy.context.scene
    
    def addLayer(name):
        layer = scene.render.layers.get(name)
        if layer == None:
            bpy.ops.scene.render_layer_add()
            layer = scene.render.layers.active
            layer.name = name
        return layer
        
    hdrRenderLayer = addLayer("syb3r_hdr_layer")
    hdrRenderLayer.use_pass_combined = True
    hdrRenderLayer.use_pass_z = False
    hdrRenderLayer.samples = 0

    attribRenderLayer = addLayer("syb3r_attrib_layer")
    attribRenderLayer.use_pass_combined = False
    attribRenderLayer.use_pass_z = True
    attribRenderLayer.use_pass_normal = True
    attribRenderLayer.use_pass_vector = True
    attribRenderLayer.use_pass_object_index = True
    attribRenderLayer.samples = 1


    # setup compositor
    scene.use_nodes = True
    node_tree = scene.node_tree
    node_tree.links.clear()
    node_tree.nodes.clear()
    
    def setupMultilayerExr(f):
        f.file_format = 'OPEN_EXR_MULTILAYER'
        f.color_mode = 'RGB'
        f.color_depth = '32'
        f.exr_codec = 'PIZ'
    
    
    rLayersHdr = node_tree.nodes.new('CompositorNodeRLayers')
    rLayersHdr.name = "syb3r_hdr_layer_data"
    rLayersHdr.layer = "syb3r_hdr_layer"
          
    
    fileOutputHdr = node_tree.nodes.new('CompositorNodeOutputFile')
    fileOutputHdr.name = "syb3r_hdr_file_output"
    fileOutputHdr.base_path = "/tmp/full_########.exr"
    fileOutputHdr.location[0] = 400
    setupMultilayerExr(fileOutputHdr.format)
    fileOutputHdr.file_slots.clear()
    fileOutputHdr.file_slots.new('HDR')
    node_tree.links.new(fileOutputHdr.inputs['HDR'], rLayersHdr.outputs['Image'])
    
    
    rLayersAttrib = node_tree.nodes.new('CompositorNodeRLayers')
    rLayersAttrib.name = "syb3r_attrib_layer_data"
    rLayersAttrib.layer = "syb3r_attrib_layer"    
    rLayersAttrib.location[1] = 300
    
    fileOutputAttrib = node_tree.nodes.new('CompositorNodeOutputFile')
    fileOutputAttrib.name = "syb3r_attrib_file_output"
    fileOutputAttrib.base_path = "/tmp/attrib_########.exr"
    fileOutputAttrib.location[0] = 400
    fileOutputAttrib.location[1] = 300
    setupMultilayerExr(fileOutputAttrib.format)
    fileOutputAttrib.file_slots.clear()
    fileOutputAttrib.file_slots.new('distance')
    fileOutputAttrib.file_slots.new('normal')
    fileOutputAttrib.file_slots.new('motion')
    fileOutputAttrib.file_slots.new('objectIdx')
    rLayersAttrib.outputs['Z'].enabled = True
    rLayersAttrib.outputs['Normal'].enabled = True
    rLayersAttrib.outputs['Speed'].enabled = True
    rLayersAttrib.outputs['IndexOB'].enabled = True
    node_tree.links.new(fileOutputAttrib.inputs['distance'], rLayersAttrib.outputs['Z'])
    node_tree.links.new(fileOutputAttrib.inputs['normal'], rLayersAttrib.outputs['Normal'])
    node_tree.links.new(fileOutputAttrib.inputs['motion'], rLayersAttrib.outputs['Speed'])
    node_tree.links.new(fileOutputAttrib.inputs['objectIdx'], rLayersAttrib.outputs['IndexOB'])


    # disable output
    hdrRenderLayer.use = False
    attribRenderLayer.use = False


def syb3r_initCamera(camera):
    camera.scale[0] = 1.0
    camera.scale[1] = 1.0
    camera.scale[2] = 1.0

    camera.data.type = 'PERSP' 
    camera.data.lens_unit = 'MILLIMETERS'
    camera.data.sensor_fit = 'HORIZONTAL'


def syb3r_initScene():
    bpy.context.scene.render.engine = 'CYCLES'
    bpy.context.scene.unit_settings.system = 'METRIC'
    bpy.context.scene.render.pixel_aspect_x = 1.0
    bpy.context.scene.render.pixel_aspect_y = 1.0
    #bpy.context.scene.render.image_settings.file_format = 'OPEN_EXR'
    #bpy.context.scene.render.image_settings.color_mode = 'RGB'
    #bpy.context.scene.render.image_settings.use_zbuffer = True
    bpy.context.scene.render.use_compositing = True
    bpy.context.scene.render.use_sequencer = False
    bpy.context.scene.cycles.progressive = 'BRANCHED_PATH' # the other one produces holes in the z/objID/... passes for alpha transparency
    bpy.context.scene.cycles.use_layer_samples = 'USE'

    syb3r_initCamera(bpy.context.scene.camera)

    syb3r_initializeRenderLayers()



#######################################################
#######################################################
#######################################################
#######################################################



class SYB3R_initScene(bpy.types.Operator):
    '''Initializes the scene for syb3r'''
    bl_idname = "syb3r.init_scene"
    bl_label = "Syb3r -> Init Scene"
    bl_options = {'REGISTER'}

    def execute(self, context):
        syb3r_initScene()
        return {'FINISHED'}


class SYB3R_initCamera(bpy.types.Operator):
    '''Initializes the selected cameras'''
    bl_idname = "syb3r.init_camera"
    bl_label = "Syb3r -> Initialize Camera"
    bl_options = {'REGISTER'}

    def execute(self, context):
        for obj in bpy.context.selected_objects:
            if obj.type == 'CAMERA':
                syb3r_initCamera(obj)

        return {'FINISHED'}

class SYB3R_setUniformObjectID(bpy.types.Operator):
    '''Set uniform object IDs'''
    bl_idname = "syb3r.set_uniform_objectid"
    bl_label = "Syb3r -> Set objectID(s) to uniform value"
    bl_options = {'REGISTER'}

    objId = bpy.props.IntProperty(
        name="objId",
        default=1,
        description="Object ID to apply to all selected objects. Zero is considered as background/unimportant, everything else as foreground."
    )

    def execute(self, context):
        for obj in bpy.context.selected_objects:
            obj.pass_index = self.objId

        return {'FINISHED'}

    def invoke(self, context, event):
        wm = context.window_manager
        return wm.invoke_props_dialog(self)


class SYB3R_setSequentialObjectID(bpy.types.Operator):
    '''Set sequential object IDs'''
    bl_idname = "syb3r.set_sequential_objectid"
    bl_label = "Syb3r -> Set objectID(s) sequentially"
    bl_options = {'REGISTER'}

    objId = bpy.props.IntProperty(
        name="first objId",
        default=1,
        description="Object ID at which to start the sequence of IDs that are applied to the selected objects. Zero is considered as background/unimportant, everything else as foreground."
    )

    def execute(self, context):
        nextObjId = self.objId
        for obj in bpy.context.selected_objects:
            obj.pass_index = nextObjId
            nextObjId = nextObjId + 1
        return {'FINISHED'}

    def invoke(self, context, event):
        wm = context.window_manager
        return wm.invoke_props_dialog(self)




def menu_func_initScene(self, context):
    self.layout.operator(SYB3R_initScene.bl_idname, text=SYB3R_initScene.__doc__, icon='PLUGIN')

def menu_func_initCamera(self, context):
    self.layout.operator(SYB3R_initCamera.bl_idname, text=SYB3R_initCamera.__doc__, icon='PLUGIN')

def menu_func_setUniformObjId(self, context):
    self.layout.operator(SYB3R_setUniformObjectID.bl_idname, text=SYB3R_setUniformObjectID.__doc__, icon='PLUGIN')

def menu_func_setSequentialObjId(self, context):
    self.layout.operator(SYB3R_setSequentialObjectID.bl_idname, text=SYB3R_setSequentialObjectID.__doc__, icon='PLUGIN')

def register():
   bpy.utils.register_class(SYB3R_initScene)
   bpy.utils.register_class(SYB3R_initCamera)
   bpy.utils.register_class(SYB3R_setUniformObjectID)
   bpy.utils.register_class(SYB3R_setSequentialObjectID)
   bpy.types.INFO_MT_file.append(menu_func_initScene)
   bpy.types.VIEW3D_MT_object.append(menu_func_initCamera)
   bpy.types.VIEW3D_MT_object.append(menu_func_setUniformObjId)
   bpy.types.VIEW3D_MT_object.append(menu_func_setSequentialObjId)
 
def unregister():
    bpy.utils.unregister_class(SYB3R_initScene)
    bpy.utils.unregister_class(SYB3R_initCamera)
    bpy.utils.unregister_class(SYB3R_setUniformObjectID)
    bpy.utils.unregister_class(SYB3R_setSequentialObjectID)
    bpy.types.INFO_MT_file.remove(menu_func_initScene)
    bpy.types.VIEW3D_MT_object.remove(menu_func_initCamera)
    bpy.types.VIEW3D_MT_object.remove(menu_func_setUniformObjId)
    bpy.types.VIEW3D_MT_object.remove(menu_func_setSequentialObjId)
 
if __name__ == "__main__":
    register()
    #syb3r_initScene()
