
#include <syb3r/models/CameraResponseCurve.h>
#include <syb3r/synthesis/ImagePostprocessor.h>
#include <syb3r/tools/RasterImage.h>
#include <syb3r/tools/FloatImage.h>
#include <syb3r/tools/TaskScheduler.h>
#include <iostream>
#include <fstream>

#include <thread>
#include <mutex>

#include <unsupported/Eigen/FFT>

#include <boost/format.hpp>
#include <boost/filesystem.hpp>
#include <boost/optional.hpp>


void extractChannel(const RasterImage &src, syb3r::tools::FloatImage<Eigen::Vector3f> &dst, unsigned channel)
{
    dst.resize(src.getWidth(), src.getHeight());
    for (unsigned i = 0; i < src.getWidth() * src.getHeight(); i++) {
        unsigned v = (src.getData()[i] >> (channel*8)) & 0xFF;
        dst[i][channel] = (v+0.5f) / 256.0f;
    }
}

void extractChannel(const RasterImage &src, syb3r::tools::FloatImage<Eigen::Vector3f> &dst, unsigned channel, const syb3r::models::CameraResponseCurve<256> &curve)
{
    dst.resize(src.getWidth(), src.getHeight());
    for (unsigned i = 0; i < src.getWidth() * src.getHeight(); i++) {
        unsigned v = (src.getData()[i] >> (channel*8)) & 0xFF;
        dst[i][channel] = curve.ldrToHdr(v);
    }
}


void printUsage(const char *cmd)
{
    std::cout << "Usage: " << cmd << " --generate-imgs path/" << std::endl
              << "   generates test images of random colors." << std::endl;

    std::cout << "Usage: " << cmd << " --tonemapCurve tonemappingCurve.xml [--generate-c++] --estimate-noise-jpg {filenames}" << std::endl
              << "   Estimates the noise model from jpg images." << std::endl;
}


int main(int argc, const char **argv)
{ 
    syb3r::tools::TaskScheduler::Init(std::thread::hardware_concurrency());
     
    boost::optional<boost::filesystem::path> generateImgsPath;    
    std::vector<std::string> noiseJpgImgs;
    
    bool generateCpp = false;
    
    boost::optional<boost::filesystem::path> tonemappingCurveFilename;
    
    for (unsigned i = 1; i < argc; i++) {
        std::string argument(argv[i]);
        if (argument == "--tonemapCurve") {
            if (i+1 >= argc) {
                std::cout << "Missing argument!" << std::endl;
                exit(1);
            }
            tonemappingCurveFilename = std::string(argv[++i]);
        } else 
        if (argument == "--generate-c++") {
            generateCpp = true;
        } else
        if (argument == "--generate-imgs") {
            if (i+1 >= argc) {
                std::cout << "Missing argument!" << std::endl;
                exit(1);
            }
            generateImgsPath = std::string(argv[++i]);
        } else 
        if (argument == "--estimate-noise-jpg") {
            if (i+1 >= argc) {
                std::cout << "Missing arguments!" << std::endl;
                exit(1);
            }
            while (++i < argc) {
                noiseJpgImgs.push_back(argv[i]);
            }
        } else {
            std::cout << "Unknown argument: " << argument << std::endl;
            printUsage(argv[0]);
            exit(1);
        }
    }
    
    if (!generateImgsPath && noiseJpgImgs.empty()) {
        printUsage(argv[0]);
        exit(1);
    }
    
    
    if (generateImgsPath) {
        boost::filesystem::create_directories(*generateImgsPath);
    
        RasterImage colorImg;
        colorImg.resize(1920, 1080);
        colorImg.clear(0xFFFFFFFF);
        colorImg.writeToFile((*generateImgsPath / "white.jpg").string().c_str());
        for (unsigned i = 0; i < 100; i++) {
            colorImg.clear(0xFF000000 | rand());
            colorImg.writeToFile((*generateImgsPath / (boost::format("color_%08d.jpg") % i).str()).string().c_str());
        }
    }    

    syb3r::models::CameraRGBResponseCurves<256> cameraCurves;
    if (!tonemappingCurveFilename) {
        std::cout << "No turnmapping curve given. Falling back to Eos400D_neutral_5200K_daylight." << std::endl;
        cameraCurves = syb3r::models::curves_Eos400D_neutral_5200K_daylight;
    } else {
        cameraCurves.readFromXML(*tonemappingCurveFilename);
    }
    
    
    if (!noiseJpgImgs.empty()) {    
        std::vector<float> gaussianKernel;
#if 0        
        gaussianKernel.resize(250);
        {
            for (unsigned i = 0; i < gaussianKernel.size(); i++)
                gaussianKernel[i] = 1.0f / (80.0f * std::sqrt(2.0f * (float)M_PI)) * std::exp(-(float)(i*i)/(2.0f * 80.0f*80.0f));
            syb3r::tools::normalizeFilterKernel(gaussianKernel);
        }
#else
        gaussianKernel.resize(80);
        {
            for (unsigned i = 0; i < gaussianKernel.size(); i++)
                gaussianKernel[i] = 1.0f / (30.0f * std::sqrt(2.0f * (float)M_PI)) * std::exp(-(float)(i*i)/(2.0f * 30.0f*30.0f));
            syb3r::tools::normalizeFilterKernel(gaussianKernel);
        }
#endif

        RasterImage image;

        Eigen::Vector3f avgVariance;

        Eigen::Matrix3f colorToVariance;
        Eigen::Vector3f varianceOffset;
        {
            std::cout << "Running matrix estimation" << std::endl;

            struct Sample {
                Eigen::Vector3f rgb;
                Eigen::Vector3f varComp;
            };

            std::vector<Sample> samples;
            std::mutex samplesMutex;

            syb3r::tools::TaskGroup group;
            
            for (const std::string &imgFilename : noiseJpgImgs) {
                group.add([imgFilename, &gaussianKernel, &cameraCurves, &samples, &samplesMutex]{
                    std::cout << "Processing " << imgFilename << std::endl;
                    RasterImage image;
                    image.loadFromFile(imgFilename.c_str());

                    syb3r::tools::FloatImage<Eigen::Vector3f> input;
                    extractChannel(image, input, 0, cameraCurves.r);
                    extractChannel(image, input, 1, cameraCurves.g);
                    extractChannel(image, input, 2, cameraCurves.b);


                    syb3r::tools::FloatImage<Eigen::Vector3f> avg, tmp;
                    syb3r::tools::blurHorizontal(tmp, input, gaussianKernel);
                    syb3r::tools::blurVertical(avg, tmp, gaussianKernel);
                    input -= avg;

                    {
                        std::lock_guard<std::mutex> lock(samplesMutex);
                        for (unsigned y = 0; y < input.getHeight(); y+=10) {
                            for (unsigned x = 0; x < input.getWidth(); x+=10) {
                                Sample sample;
                                sample.rgb = avg(x, y);

                                if (sample.rgb[0] > 2.5f) continue;
                                if (sample.rgb[1] > 2.5f) continue;
                                if (sample.rgb[2] > 2.5f) continue;

                                sample.varComp = input(x, y).cwiseProduct(input(x, y));

                                samples.push_back(sample);
                            }
                        }
                    }
                }, syb3r::tools::TaskScheduler::get());
            }

            syb3r::tools::TaskScheduler::get().waitFor(&group);

            std::cout << "collected " << samples.size() << " samples!" << std::endl;


            {
                Eigen::MatrixXf A;
                Eigen::VectorXf b;
                A.resize(3*samples.size(), 9 + 3);
                A.setZero();
                b.resize(3*samples.size());
                for (unsigned i = 0; i < samples.size(); i++) {
                    for (unsigned j = 0; j < 3; j++) {
                        A(i*3 + j, j*3 + 0) = samples[i].rgb[0];
                        A(i*3 + j, j*3 + 1) = samples[i].rgb[1];
                        A(i*3 + j, j*3 + 2) = samples[i].rgb[2];
                        b[i*3 + j] = samples[i].varComp[j];
                    }
                    A(i*3 + 0, 9) = 1.0f;
                    A(i*3 + 1, 10) = 1.0f;
                    A(i*3 + 2, 11) = 1.0f;
                }

                Eigen::MatrixXf AtA = A.transpose() * A;
                Eigen::VectorXf Atb = A.transpose() * b;

                Eigen::VectorXf x = AtA.fullPivHouseholderQr().solve(Atb);

                for (unsigned i = 0; i < 3; i++)
                    for (unsigned j = 0; j < 3; j++)
                        colorToVariance(i, j) = x[i*3+j];

                varianceOffset[0] = x[9+0];
                varianceOffset[1] = x[9+1];
                varianceOffset[2] = x[9+2];
            }
            {
                Eigen::Vector3f error;
                error.setZero();
                avgVariance.setZero();
                for (unsigned i = 0; i < samples.size(); i++) {
                    avgVariance += samples[i].varComp;

                    Eigen::Vector3f estimatedCompVar = colorToVariance * samples[i].rgb + varianceOffset;
                    estimatedCompVar[0] = std::max(estimatedCompVar[0], 0.0f);
                    estimatedCompVar[1] = std::max(estimatedCompVar[1], 0.0f);
                    estimatedCompVar[2] = std::max(estimatedCompVar[2], 0.0f);

                    Eigen::Vector3f diff = estimatedCompVar - samples[i].varComp;
                    error += diff.cwiseProduct(diff);
                }
                avgVariance /= samples.size();
                error /= samples.size();
                std::cout << "MSE Error: " << error << std::endl;
                std::cout << "avgVariance: " << avgVariance << std::endl;
            }
        }
        
        if (generateCpp) {        
            std::cout << "########################################" << std::endl << std::endl;
            std::cout << "Eigen::Vector3f avgVariance("
                    << avgVariance[0] << ", "
                    << avgVariance[1] << ", "
                    << avgVariance[2] << ");"
                    << std::endl << std::endl;
            std::cout << "########################################" << std::endl << std::endl;
            std::cout << "Eigen::Matrix3f color_to_variance;" << std::endl;
            std::cout << "color_to_variance << "
                    << "                     " << colorToVariance(0, 0) << ", " << colorToVariance(0, 1) << ", " << colorToVariance(0, 2) << ", " << std::endl
                    << "                     " << colorToVariance(1, 0) << ", " << colorToVariance(1, 1) << ", " << colorToVariance(1, 2) << ", " << std::endl
                    << "                     " << colorToVariance(2, 0) << ", " << colorToVariance(2, 1) << ", " << colorToVariance(2, 2) << ";" << std::endl
                    << std::endl;

            std::cout << "Eigen::Vector3f variance_offset("
                    << varianceOffset[0] << ", "
                    << varianceOffset[1] << ", "
                    << varianceOffset[2] << ");"
                    << std::endl;
        }
    }
}