#include <iostream>

#include <Eigen/Dense>

#include <boost/lexical_cast.hpp>
#include <boost/optional.hpp>
#include <boost/filesystem.hpp>

#include <syb3r/tools/RasterImage.h>
#include <syb3r/models/CameraShake.h>

#include <exiv2/exiv2.hpp>

#include <string>
#include <vector>
#include <algorithm>

struct ImageMetaData {
    float exposureTime;
    float focalLength;
    float sensorWidth;
};

ImageMetaData getImageMetaData(const std::string &filename)
{
    Exiv2::Image::AutoPtr image = Exiv2::ImageFactory::open(filename);
    if (image.get() == nullptr)
        throw std::runtime_error(std::string("Could not open file: ") + filename);

    image->readMetadata();

    Exiv2::ExifData &exifData = image->exifData();
    if (exifData.empty())
        std::runtime_error(std::string("No exif data in file: ") + filename);

    auto exposureTimeIt = exifData.findKey(Exiv2::ExifKey("Exif.Photo.ExposureTime"));
    if (exposureTimeIt == exifData.end())
        std::runtime_error(std::string("No exif tag 'Exif.Photo.ExposureTime' in file: ") + filename);

    auto focalLengthIt = exifData.findKey(Exiv2::ExifKey("Exif.Photo.FocalLength"));
    if (focalLengthIt == exifData.end())
        std::runtime_error(std::string("No exif tag 'Exif.Photo.FocalLength' in file: ") + filename);

    ImageMetaData metaData;
    metaData.exposureTime = exposureTimeIt->toFloat();
    metaData.focalLength = focalLengthIt->toFloat();

    return metaData;
}

struct ConnectedComponent {
    std::vector<Eigen::Vector2i> pixels;
};


std::vector<ConnectedComponent> getConnectedComponents(const RasterImage &img, float thresh)
{
    std::vector<ConnectedComponent> connectedComponents;

    float avg;
    {
        std::uint64_t avgAccu = 0;
        for (unsigned i = 0; i < img.getWidth()*img.getHeight(); i++)
            avgAccu += img.getData()[i] & 0xFF;
        avg = avgAccu / (float) (img.getWidth()*img.getHeight());
    }

    auto isDotPixel = [&img, avg, thresh](unsigned x, unsigned y)->bool{
        if (x < 100) return false;
        if (x > img.getWidth()-100) return false;
        if (y < 100) return false;
        if (y > img.getHeight()-100) return false;
        return (img.getData()[x+y*img.getWidth()] & 0xFF) > avg * thresh;
    };

    std::vector<Eigen::Vector2i> openList;
    std::vector<bool> handledPixels;
    handledPixels.resize(img.getWidth() * img.getHeight(), false);
    for (unsigned y = 0; y < img.getHeight(); y++)
        for (unsigned x = 0; x < img.getWidth(); x++) {
            if (handledPixels[x+y*img.getWidth()]) continue;
            if (!isDotPixel(x, y)) continue;

            openList.clear();

            openList.push_back(Eigen::Vector2i(x, y));
            ConnectedComponent component;

            while (!openList.empty()) {
                Eigen::Vector2i pixel = openList.back();
                openList.pop_back();

                if (handledPixels[pixel[0]+pixel[1]*img.getWidth()]) continue;
                if (!isDotPixel(pixel[0], pixel[1])) continue;
                component.pixels.push_back(pixel);
                handledPixels[pixel[0]+pixel[1]*img.getWidth()] = true;

                if (pixel[0] > 0)
                    openList.push_back(Eigen::Vector2i(pixel[0]-1, pixel[1]));
                if (pixel[1] > 0)
                    openList.push_back(Eigen::Vector2i(pixel[0], pixel[1]-1));
                if (pixel[0]+1 < (int) img.getWidth())
                    openList.push_back(Eigen::Vector2i(pixel[0]+1, pixel[1]));
                if (pixel[1]+1 < (int) img.getHeight())
                    openList.push_back(Eigen::Vector2i(pixel[0], pixel[1]+1));
            }

            if (!component.pixels.empty())
                connectedComponents.push_back(std::move(component));
        }


    return connectedComponents;
}


void removeAllComponentsSmallerThan(unsigned numPixels, std::vector<ConnectedComponent> connectedComponents)
{
    for (unsigned i = 0; i < connectedComponents.size();) {
        if (connectedComponents[i].pixels.size() < numPixels) {
            connectedComponents[i] = std::move(connectedComponents[connectedComponents.size()-1]);
            connectedComponents.resize(connectedComponents.size()-1);
        } else i++;
    }
}


void printUsage(const char *cmd)
{
    std::cout << cmd << " --generateImgs path/dotImg.png" << std::endl
              << "    Generate test images" << std::endl;
    std::cout << cmd << " [--cropFactor 1.0] [--generate-mat-plot] [--generate-xml filename.xml] [--generate-c++] --filenames {filenames}" << std::endl
              << "    --filenames: List of filenames to process" << std::endl
              << "    Compute motion blur statistics" << std::endl;
}

int main(int argc, const char **argv)
{
    std::vector<std::string> filenames;
    
    boost::optional<boost::filesystem::path> dotImagePath;

    float thresh = 4.0f;
    float cropFactor = 1.0f;

    bool generateMatPlot = false;
    bool generateCpp = false;
    boost::optional<boost::filesystem::path> xmlPath;
    
    for (int i = 1; i < argc; i++) {
        std::string argument(argv[i]);

        if (argument == "--generateImgs") {
            if (i+1 >= argc) {
                std::cout << "Missing argument after " << argument << std::endl;
                return -1;
            }
            dotImagePath = argv[++i];
        } else
        if (argument == "--filenames") {
            i++;
            for (; i < argc; i++) {
                filenames.push_back(argv[i]);
            }
        } else
        if (argument == "--generate-mat-plot") {
            generateMatPlot = true;
        } else
        if (argument == "--generate-xml") {
            if (i+1 >= argc) {
                std::cout << "Missing argument after " << argument << std::endl;
                return -1;
            }
            xmlPath = argv[++i];
        } else
        if (argument == "--generate-c++") {
            generateCpp = true;
        } else
        if (argument == "--cropFactor") {
            if (i+1 >= argc) {
                std::cout << "Missing argument after " << argument << std::endl;
                return -1;
            }
            try {
                cropFactor = boost::lexical_cast<float>(argv[++i]);
            } catch (...) {
                std::cout << "Invalid argument after " << argument << std::endl;
                return -1;
            }
        } else {
            std::cout << "Unknown argument: " << argument << std::endl;
        }
    }


    if (filenames.empty() && !dotImagePath) {
        printUsage(argv[0]);
        return -1;
    }
        
    if (dotImagePath) {
        boost::filesystem::create_directories(dotImagePath->parent_path());
    
        RasterImage dotImg;
        dotImg.resize(1920, 1080);
        dotImg.clear(0xFF000000);
        
        for (unsigned y = 0; y < dotImg.getHeight(); y += 128)
            for (unsigned x = 0; x < dotImg.getWidth(); x += 128) {
                dotImg.getData()[y*dotImg.getWidth()+x] = 0xFFFFFFFF;
            }                
        
        dotImg.writeToFile(dotImagePath->string().c_str());        
    }
    
    

    auto getMotionBlur = [thresh, cropFactor](const std::string &filename)->float{
        std::cout << filename << std::endl;
        ImageMetaData metaData = getImageMetaData(filename);


        RasterImage img;
        img.loadFromFile(filename.c_str());

        std::vector<ConnectedComponent> connectedComponents = getConnectedComponents(img, thresh);

        std::cout << "Num components: " << connectedComponents.size() << std::endl;

        unsigned numPixels = 0;
        for (unsigned i = 0; i < connectedComponents.size(); i++)
            numPixels += connectedComponents[i].pixels.size();

        float avgPixelsPerComponent = numPixels / (float)connectedComponents.size();
        std::cout << "avgPixelsPerComponent " << avgPixelsPerComponent << std::endl;

        removeAllComponentsSmallerThan(avgPixelsPerComponent * 1.0f, connectedComponents);

        std::cout << "Num components: " << connectedComponents.size() << std::endl;

#if 0
                RasterImage debugImage;
                debugImage.resize(img.getWidth(), img.getHeight());
                debugImage.clear(0xFF000000);

                for (unsigned i = 0; i < connectedComponents.size(); i++) {
                    uint32_t color = 0xFF000000 | (i+1) * 1234;

                    for (const Eigen::Vector2i &pixel : connectedComponents[i].pixels) {
                        debugImage.getData()[pixel[0] + pixel[1] * debugImage.getWidth()] = color;
                    }
                }
                debugImage.writeToFile("components.png");
#endif
        std::vector<float> lengths;
        for (unsigned i = 0; i < connectedComponents.size(); i++) {
            Eigen::Vector2f mean;
            mean.setZero();
            Eigen::Vector3f covarAccu;
            covarAccu.setZero();
            for (const Eigen::Vector2i &pixel : connectedComponents[i].pixels) {
                mean += Eigen::Vector2f(pixel[0], pixel[1]);
                covarAccu[0] += pixel[0]*pixel[0];
                covarAccu[1] += pixel[1]*pixel[1];
                covarAccu[2] += pixel[0]*pixel[1];
            }

            mean /= connectedComponents[i].pixels.size();

            Eigen::Matrix2f covar;
            covar(0, 0) = covarAccu[0] / connectedComponents[i].pixels.size() - mean[0]*mean[0];
            covar(0, 1) =
            covar(1, 0) = covarAccu[2] / connectedComponents[i].pixels.size() - mean[0]*mean[1];
            covar(1, 1) = covarAccu[1] / connectedComponents[i].pixels.size() - mean[1]*mean[1];

            Eigen::SelfAdjointEigenSolver<Eigen::Matrix2f> es;
            es.compute(covar);

            Eigen::Vector2f blurDirection = es.eigenvectors().col(0);
            Eigen::Vector2f movementDirection = es.eigenvectors().col(1);

            float movementMin;
            float movementMax;
            movementMin = movementMax = movementDirection.dot(Eigen::Vector2f(connectedComponents[i].pixels[0][0], connectedComponents[i].pixels[0][1]));

            float blurMin;
            float blurMax;
            blurMin = blurMax = blurDirection.dot(Eigen::Vector2f(connectedComponents[i].pixels[0][0], connectedComponents[i].pixels[0][1]));

            for (const Eigen::Vector2i &pixel : connectedComponents[i].pixels) {
                float movementF = movementDirection.dot(Eigen::Vector2f(pixel[0], pixel[1]));
                movementMin = std::min(movementMin, movementF);
                movementMax = std::max(movementMax, movementF);

                float blurF = blurDirection.dot(Eigen::Vector2f(pixel[0], pixel[1]));
                blurMin = std::min(blurMin, blurF);
                blurMax = std::max(blurMax, blurF);
            }
            float blur = blurMax - blurMin;
blur = 3.0f; // due to curvature, estimating the blur is much too fragile
            float movement = movementMax - movementMin - blur;
            lengths.push_back(movement);
        }

        std::sort(lengths.begin(), lengths.end());
        float length_pixels = lengths[lengths.size()*3/4];

        float sensorWidth_mm = 36.0f / cropFactor;
        float sensorWidth_pixels = img.getWidth();

        return length_pixels / sensorWidth_pixels * sensorWidth_mm / metaData.focalLength / metaData.exposureTime;
    };


    std::vector<float> moBlur;
    for (const std::string &filename : filenames) {
        float movement = getMotionBlur(filename);
        moBlur.push_back(movement);
    }

    float meanMoBlur = 0.0f;
    for (float f : moBlur)
        meanMoBlur += f;
    meanMoBlur /= moBlur.size();

    auto gamma = [](unsigned k)->unsigned{
        unsigned result = 1;
        for (unsigned i = 2; i+1 <= k; i++)
            result *= i;
        return result;
    };

    float maxLogLikelihood = -1e30f;
    unsigned maxK = 1;
    for (unsigned k = 1; k < 5; k++) {
        float theta = meanMoBlur / k;

        float logLikelihood = 0.0f;
        for (float f : moBlur)
            logLikelihood += -std::log(gamma(k)) - std::log(theta)*k + std::log(f)*(k-1) - f/theta;

        if (logLikelihood > maxLogLikelihood) {
            maxLogLikelihood = logLikelihood;
            maxK = k;
        }
    }


    if (generateCpp) {
        std::cout << "###############################" << std::endl;
        std::cout << "CameraShake cameraShake("<<maxK<<", "<< meanMoBlur/maxK <<"f);"<<std::endl;
    }
    
    if (xmlPath) {
        syb3r::models::LinearCameraShake cameraShake(maxK, meanMoBlur/maxK);
        cameraShake.writeToXML(*xmlPath);
    }

    if (generateMatPlot) {
        std::cout << "###############################" << std::endl;
        std::cout << "moBlur = [";
        for (float f : moBlur)
            std::cout << " " << f;
        std::cout << "];" << std::endl;

        std::cout << "k = " << maxK << std::endl;
        std::cout << "theta = mean(moBlur)/k" << std::endl;

        std::cout << "x = [0:0.1:60];" << std::endl;
        std::cout << "pdf = 1/(gamma(k)*theta^k)*x.^(k-1).*exp(-x./theta);" << std::endl;
        std::cout << "figure(1)" << std::endl;
        std::cout << "hist(moBlur,5);" << std::endl;
        std::cout << "xlabel('Blur amount [pixels]');" << std::endl;
        std::cout << "ylabel('Empirical probability');" << std::endl;
        std::cout << "axis ([0 60])" << std::endl;
        std::cout << "figure(2)" << std::endl;
        std::cout << "plot(x, pdf)" << std::endl;
        std::cout << "xlabel('Blur amount [pixels]');" << std::endl;
        std::cout << "ylabel('Fitted probability density');" << std::endl;
    }

}