
#include <syb3r/models/CameraResponseCurve.h>

#include <iostream>

#include <boost/optional.hpp>
#include <boost/lexical_cast.hpp>
#include <string>
#include <vector>


void printUsage(const char *cmd)
{
    std::cout << "This tool computes the nonlinear mapping of a camera between jpg colors and linear intensities."
              << "It requires a set of camera jpg images that depict the exact same content but acquired with different"
              << " exposure times (stored in the exif meta data). The images should be out of focus to reduce alignment problems."
              << " A blury white is sufficient, but lighting should be constant so that the brightness only changes because and according to"
              << " the camera's exposure settings." << std::endl;

    std::cout << "Usage: " << cmd << " [--numPixels 100] [--smoothness 300.0] [--generate-mat-plot] [--generate-c++] [--generate-xml filename.xml] --filenames {filenames}" << std::endl
              << "   --numPixels: Number of pixels to use from each image." << std::endl
              << "   --smoothness: Smoothness based regularization on the curve." << std::endl
              << "   --generate-mat-plot: Write matlab script to stdout that contains data and plotting instructions to display the curve." << std::endl
              << "   --generate-c++: Output results as c++ code" << std::endl
              << "   --generate-xml filename.xml: Write results to xml file" << std::endl
              << "   --filenames: List of jpg files." << std::endl;
}


int main(int argc, const char **argv)
{

    bool generateMatPlot = false;

    unsigned numPixels = 100;
    float smoothness = 300.0f;
    bool generateCpp = false;
    boost::optional<boost::filesystem::path> xmlPath;

    std::vector<std::string> filenames;

    for (int i = 1; i < argc; i++) {
        std::string argument(argv[i]);
        if (argument == "--numPixels") {
            if (i+1 >= argc) {
                std::cout << "Missing number after " << argument << std::endl;
                exit(1);
            }
            try {
                numPixels = boost::lexical_cast<unsigned>(argv[++i]);
            } catch (...) {
                std::cout << "Invalid number after " << argument << std::endl;
                exit(1);
            }
        } else
        if (argument == "--smoothness") {
            if (i+1 >= argc) {
                std::cout << "Missing number after " << argument << std::endl;
                exit(1);
            }
            try {
                smoothness = boost::lexical_cast<float>(argv[++i]);
            } catch (...) {
                std::cout << "Invalid number after " << argument << std::endl;
                exit(1);
            }
        } else
        if (argument == "--generate-mat-plot") {
            generateMatPlot = true;
        } else
        if (argument == "--generate-c++") {
            generateCpp = true;
        } else
        if (argument == "--generate-xml") {
            if (i+1 >= argc) {
                std::cout << "Missing number after " << argument << std::endl;
                exit(1);
            }
            xmlPath = argv[++i];
        } else
        if (argument == "--filenames") {
            i++;
            for (; i < argc; i++)
                filenames.push_back(argv[i]);
        } else {
            std::cout << "Unknown argument " << argument << std::endl;
            exit(1);
        }
    }

    if (filenames.size() < 2) {
        std::cout << "Needs at least two images!" << std::endl;
        printUsage(argv[0]);
        exit(1);
    }

    syb3r::models::CameraRGBResponseCurves<256> curves;
    
    try {
        syb3r::models::computeCameraResponseFromMultipleExposures(
                                                                filenames,
                                                                numPixels, smoothness,
                                                                curves.r, curves.g, curves.b);
    } catch (const std::exception &e) {
        std::cout << "Error: " << e.what() << std::endl;
    }

    if (generateCpp) {
        std::cout << "###################################################" << std::endl;
        std::cout << "syb3r::models::CameraRGBResponseCurves<256> curves = {" << std::endl;
        
        std::cout << "    {" << std::endl;
        
        for (unsigned i = 0; i < 16; i++) {
            std::cout << "        ";
            for (unsigned j = 0; j < 16; j++) {
                std::cout << curves.r.ldrToHdr(i*16+j) << ", ";
            }
            std::cout << std::endl;
        }
        std::cout << "    },"<<std::endl;

        std::cout << "    {" << std::endl;
        for (unsigned i = 0; i < 16; i++) {
            std::cout << "        ";
            for (unsigned j = 0; j < 16; j++) {
                std::cout << curves.g.ldrToHdr(i*16+j) << ", ";
            }
            std::cout << std::endl;
        }
        std::cout << "    },"<<std::endl;

        std::cout << "    {" << std::endl;
        for (unsigned i = 0; i < 16; i++) {
            std::cout << "    ";
            for (unsigned j = 0; j < 16; j++) {
                std::cout << curves.b.ldrToHdr(i*16+j) << ", ";
            }
            std::cout << std::endl;
        }
        std::cout << "    }"<<std::endl;
        std::cout << "};"<<std::endl;
    }
    
    if (xmlPath) {
        curves.writeToXML(*xmlPath);
    }


    if (generateMatPlot) {
        std::cout << "###################################################" << std::endl;

        auto printCurve = [](const char *name, const syb3r::models::CameraResponseCurve<256> &curve) {
            std::cout << name << " = [";
            for (unsigned i = 0; i < 256; i++) {
                if (i % 16 == 0)
                    std::cout << "..." << std::endl;
                std::cout << curve.ldrToHdr(i) << " ";
            }
            std::cout << "];" << std::endl;
        };

        printCurve("redCurve", curves.r);
        printCurve("greenCurve", curves.g);
        printCurve("blueCurve", curves.b);

        std::cout << "x = [0:1:255];" << std::endl
                  << "figure(1);" << std::endl
                  << "plot(x, redCurve, 'r', 'linewidth', 2, x, greenCurve, 'g', 'linewidth', 2, x, blueCurve, 'b', 'linewidth', 2);" << std::endl
                  << "xlabel('sRGB value');" << std::endl
                  << "ylabel('linear intensity');" << std::endl
                  << "set(gca, 'linewidth', 4, 'fontsize', 12)" << std::endl
                  << "figure(2);" << std::endl
                  << "semilogy(x, redCurve, 'r', 'linewidth', 2, x, greenCurve, 'g', 'linewidth', 2, x, blueCurve, 'b', 'linewidth', 2);" << std::endl
                  << "xlabel('sRGB value');" << std::endl
                  << "ylabel('log intensity');" << std::endl
                  << "set(gca, 'linewidth', 4, 'fontsize', 12)" << std::endl;
    }
}