

#include <iostream>

#include <boost/format.hpp>
#include <boost/optional.hpp>
#include <boost/filesystem.hpp>

#include <syb3r/synthesis/ImagePostprocessor.h>
#include <syb3r/data/Dataset.h>
#include <syb3r/tools/RasterImage.h>

#include <syb3r/tools/TaskScheduler.h>

#include <unordered_map>

#include <thread>
#include <fstream>


inline Eigen::Vector3f colorRamp(float f) {
    Eigen::Vector3f color;
    if (f < 0.0) {
        color[0] = 0;
        color[1] = 0;
        color[2] = 1.0;
        return color;
    }
    if (f < 0.25) {
        color[0] = 0;
        color[1] = f / 0.25;
        color[2] = 1.0;
        return color;
    }
    if (f < 0.5) {
        color[0] = 0;
        color[1] = 1.0;
        color[2] = (0.5-f) / 0.25;
        return color;
    }
    if (f < 0.75) {
        color[0] = (f-0.5) / 0.25;
        color[1] = 1.0;
        color[2] = 0;
        return color;
    }
    if (f < 1.0) {
        color[0] = 1.0;
        color[1] = (1.0-f) / 0.25;
        color[2] = 0;
        return color;
    }
    color[0] = 1.0;
    color[1] = 0;
    color[2] = 0;

    return color;
}


RasterImage colorRampFloatImg(const syb3r::data::Image::FloatArray &values, float min, float max)
{
    RasterImage result;
    result.resize(values.cols(), values.rows());
    for (unsigned y = 0; y < values.rows(); y++)
        for (unsigned x = 0; x < values.cols(); x++) {
            float lambda = std::min(std::max((values(y, x) - min) / (max - min), 0.0f), 1.0f);

            Eigen::Vector3f color = colorRamp(1.0f-lambda);

            int r = color[0] * 255.0f;
            int g = color[1] * 255.0f;
            int b = color[2] * 255.0f;

            result.getData()[x+y*values.cols()] =
                    (r << 0) |
                    (g << 8) |
                    (b << 16) |
                    (0xFF << 24);
        }

    return result;
}

RasterImage colorMapObjIdImg(const syb3r::data::Image::UIntArray &values, const std::unordered_map<unsigned, Eigen::Vector3f> &colorMap)
{
    RasterImage result;
    result.resize(values.cols(), values.rows());
    for (unsigned y = 0; y < values.rows(); y++)
        for (unsigned x = 0; x < values.cols(); x++) {
            int r, g, b;

            if (values(y, x) == ~0u) {
                r = g = b = ((x & 0b100) ==  (y & 0b100))?200:100;
            } else {
                Eigen::Vector3f color = colorMap.at(values(y, x));

                r = color[0] * 255.0f;
                g = color[1] * 255.0f;
                b = color[2] * 255.0f;
            }

            result.getData()[x+y*values.cols()] =
                    (r << 0) |
                    (g << 8) |
                    (b << 16) |
                    (0xFF << 24);
        }

    return result;
}

void renderBeautyImages(const syb3r::data::Dataset &dataset, const std::string &dstPath, syb3r::synthesis::ImagePostprocessor &imgPostprocessor, float minDepth = -1.0f, float maxDepth = -1.0f)
{

    if ((minDepth < 0.0f) || (maxDepth < 0.0f)) {
        std::vector<float> zValueSubset;

        std::cout << "Finding z value distribution:" << std::endl;
        for (unsigned i = 0; i < dataset.getImages().size(); i++) {
            std::cout << "\r    " << i << " out of " << dataset.getImages().size() << " images" << std::flush;
            const syb3r::data::Image &image = dataset.getImages()[i];

            syb3r::data::Image::FloatArray subsampledZBuffer = image.readZBuffer(5);

            for (unsigned y = 0; y < subsampledZBuffer.rows(); y++)
                for (unsigned x = 0; x < subsampledZBuffer.cols(); x++) {
                    float z = subsampledZBuffer(y, x);
                    if (z < 1e8f)
                        zValueSubset.push_back(z);
                }
        }

        std::sort(zValueSubset.begin(), zValueSubset.end());

        minDepth = zValueSubset[zValueSubset.size() / 100];
        maxDepth = zValueSubset[zValueSubset.size() * 99 / 100];
    }
    std::cout << std::endl;

    std::cout << "Finding object IDs:" << std::endl;
    std::unordered_map<unsigned, Eigen::Vector3f> colorMap;
    for (unsigned i = 0; i < dataset.getImages().size(); i++) {
        std::cout << "\r    " << i << " out of " << dataset.getImages().size() << " images" << std::flush;
        const syb3r::data::Image &image = dataset.getImages()[i];

        syb3r::data::Image::UIntArray objectIDs = image.readObjectID();

        for (unsigned y = 0; y < objectIDs.rows(); y++)
            for (unsigned x = 0; x < objectIDs.cols(); x++)
                colorMap[objectIDs(y,x)] = Eigen::Vector3f(0.0f, 0.0f, 0.0f);
    }
    std::cout << std::endl;
    std::cout << "Object IDs: " << std::endl;
    for (auto &color : colorMap)
        std::cout << "  " << color.first;
    std::cout << std::endl;

    {
        unsigned idx = 0;
        for (auto &color : colorMap)
            if (color.first != ~0u)
                color.second = colorRamp(idx++ / std::max<float>(colorMap.size()-2, 1));
    }


    std::vector<std::string> filenames;
    std::cout << "Rendering color images" << std::endl;
    imgPostprocessor.process(dataset, dstPath, filenames);
    std::cout << "Rendering zBuffer and object ID images" << std::endl;
    for (unsigned i = 0; i < dataset.getImages().size(); i++) {
        const syb3r::data::Image &image = dataset.getImages()[i];

        boost::filesystem::path imgPath(filenames[i]);
        //boost::filesystem::path imgPath((boost::format("%s%08d.jpg") % dstPath % i).str());

        boost::filesystem::path zBufferPath = imgPath.parent_path() / (imgPath.stem().string() + "_zBuffer.jpg");
        RasterImage zBufferImg = colorRampFloatImg(image.readZBuffer(), minDepth, maxDepth);
        zBufferImg.writeToFile(zBufferPath.string().c_str(), 100);

        boost::filesystem::path objIdPath = imgPath.parent_path() / (imgPath.stem().string() + "_objId.png");
        RasterImage objIdImg = colorMapObjIdImg(image.readObjectID(), colorMap);
        objIdImg.writeToFile(objIdPath.string().c_str());
    }
}

void printUsage(const char *cmd)
{
    std::cout << "Usage: " << cmd << " --dataset path/to/dataset.xml --outputDir path/to/outputdir/" << std::endl;
}

int main(int argc, const char **argv)
{
    syb3r::tools::TaskScheduler::Init(std::thread::hardware_concurrency());

    boost::optional<boost::filesystem::path> datasetFilename;
    boost::optional<boost::filesystem::path> destinationPath;

    for (unsigned i = 1; i < argc; i++) {
        std::string argument(argv[i]);
        if (argument == "--dataset") {
            if (i+1 < argc) {
                datasetFilename = argv[++i];
            } else {
                std::cout << "Missing argument after --dataset" << std::endl;
                return -1;
            }
        } else
        if (argument == "--outputDir") {
            if (i+1 < argc) {
                destinationPath = argv[++i];
            } else {
                std::cout << "Missing argument after --outputDir" << std::endl;
                return -1;
            }
        } else {
            std::cout << "Invalid argument: " << argument << std::endl;
            return -1;
        }
    }

    if (!datasetFilename) {
        std::cout << "Dataset is mandatory!" << std::endl;
        printUsage(argv[0]);
        return -1;
    }

    if (!destinationPath) {
        std::cout << "Output directory is mandatory!" << std::endl;
        printUsage(argv[0]);
        return -1;
    }

    boost::filesystem::create_directories(*destinationPath);

    try {
        syb3r::synthesis::ImagePostprocessor postprocessor;
        postprocessor.setupOldEOS400();
        postprocessor.initialProperties.set("jpegCompression", 98);

        syb3r::data::Dataset dataset;
        dataset.loadFromXML(*datasetFilename);

        renderBeautyImages(dataset,
                        destinationPath->string(),
                        postprocessor);

    } catch (const std::exception &e) {
        std::cout << "Error: " << e.what() << std::endl;
        return -1;
    }
    return 0;
}
