
#include <syb3r/synthesis/ImagePostprocessor.h>
#include <syb3r/wrappers/WrapperVSFM.h>
#include <syb3r/wrappers/WrapperPMVS.h>

#include <syb3r/data/Dataset.h>
#include <syb3r/data/EstimatedCameraPath.h>
#include <syb3r/data/PlyPointcloudFile.h>
#include <syb3r/data/Pointcloud.h>

#include <syb3r/analysis/CameraPathEvaluation.h>
#include <syb3r/analysis/PointCloudEvaluation.h>

#include <syb3r/tools/TmpDir.h>
#include <syb3r/tools/TaskScheduler.h>

#include <thread>

#include <iostream>

#include <fstream>

/////////////////////////////////////////////////////////
const char *path_to_vsfm = "path/to/vsfm/bin/";
const char *path_to_pmvs2 = "path/to/pmvs2";
/////////////////////////////////////////////////////////

void writePCResults(const syb3r::analysis::PointCloudEvaluation &eval, const syb3r::data::Dataset &referenceDataset, std::ostream &stream)
{
    stream << "correctness:" << std::endl;
    stream << "\tmin: " << eval.getPCCorrectness().min * *referenceDataset.getUnitsToMeters() << std::endl;
    stream << "\tavg: " << eval.getPCCorrectness().avg * *referenceDataset.getUnitsToMeters() << " +- " << eval.getPCCorrectness().stddev * *referenceDataset.getUnitsToMeters() << std::endl;
    stream << "\tmax: " << eval.getPCCorrectness().max * *referenceDataset.getUnitsToMeters() << std::endl;
    stream << "\tcorrect: " << eval.getPCCorrectness().numCorrect << " of " << eval.getPCCorrectness().numTotal << " ( " << (eval.getPCCorrectness().numCorrect/(float)eval.getPCCorrectness().numTotal*100) << "% )" << std::endl;
    stream << "completeness:" << std::endl;
    stream << "\tmin: " << eval.getPCCompleteness().min * *referenceDataset.getUnitsToMeters() << std::endl;
    stream << "\tavg: " << eval.getPCCompleteness().avg * *referenceDataset.getUnitsToMeters() << " +- " << eval.getPCCompleteness().stddev * *referenceDataset.getUnitsToMeters() << std::endl;
    stream << "\tmax: " << eval.getPCCompleteness().max * *referenceDataset.getUnitsToMeters() << std::endl;
    stream << "\tcorrect: " << eval.getPCCompleteness().numCorrect << " of " << eval.getPCCompleteness().numTotal << " ( " << (eval.getPCCompleteness().numCorrect/(float)eval.getPCCompleteness().numTotal*100) << "% )" << std::endl;
}


void runCompressionVsSFMPrecision(const char *dataset, const char *results)
{
    std::cout << "Loading scene" << std::endl;
    syb3r::data::Dataset referenceDataset;
    referenceDataset.loadFromXML(dataset);

    std::vector<std::pair<unsigned,  float>> compressionAvgPrecisionCurve;
    std::vector<std::pair<unsigned,  float>> compressionMinPrecisionCurve;
    std::vector<std::pair<unsigned,  float>> compressionMaxPrecisionCurve;

    syb3r::synthesis::ImagePostprocessor postProcessor;
    postProcessor.setupOldEOS400();

    for (unsigned i = 0; i < 10; i++) {
        syb3r::tools::TmpDir tmpDir;
        std::cout << "tmp dir: " << tmpDir.getPath().string() << std::endl;

        std::vector<std::string> ldrFilenames;

        std::cout << "Baking ldr images" << std::endl;
        postProcessor.initialProperties.set("jpegCompression", int(100-i*9));
        postProcessor.process(referenceDataset, tmpDir.getPath().string(), ldrFilenames);

        float error = 0.0f;
        float errorMin = 1e30f;
        float errorMax = -1e30f;
        for (unsigned j = 0; j < 20; j++) {
            std::cout << "Running sfm" << std::endl;
            syb3r::wrapper::WrapperVSFM sfm;
            sfm.executablePath = path_to_vsfm;
            sfm.nvmOutputFilename = (tmpDir.getPath() / "vsfm.nvm").string();
            syb3r::data::EstimatedCameraPath estimatedCameraPath;
            sfm.runSFM(referenceDataset, ldrFilenames, estimatedCameraPath);

            std::cout << "Analysing camera path" << std::endl;
            syb3r::analysis::CameraPathEvaluation pathEval(referenceDataset, estimatedCameraPath);

            error += pathEval.getCameraPositionAccuracy();
            errorMin = std::min(errorMin, pathEval.getCameraPositionAccuracy());
            errorMax = std::max(errorMax, pathEval.getCameraPositionAccuracy());
        }
        compressionAvgPrecisionCurve.push_back({postProcessor.initialProperties.getInt("jpegCompression"), error / 20});
        compressionMinPrecisionCurve.push_back({postProcessor.initialProperties.getInt("jpegCompression"), errorMin});
        compressionMaxPrecisionCurve.push_back({postProcessor.initialProperties.getInt("jpegCompression"), errorMax});
    }

    std::fstream file;
    file.exceptions(std::ifstream::failbit | std::ifstream::badbit);
    file.open(results, std::fstream::out);
    file << "compression = [";
    for (unsigned i = 0; i < compressionAvgPrecisionCurve.size(); i++)
        file << " " << compressionAvgPrecisionCurve[i].first;
    file << "];" << std::endl;
    file << "precision = [";
    for (unsigned i = 0; i < compressionAvgPrecisionCurve.size(); i++)
        file << " " << compressionAvgPrecisionCurve[i].second;
    file << "];" << std::endl;
    file << "precisionMax = [";
    for (unsigned i = 0; i < compressionMaxPrecisionCurve.size(); i++)
        file << " " << compressionMaxPrecisionCurve[i].second;
    file << "];" << std::endl;
    file << "precisionMin = [";
    for (unsigned i = 0; i < compressionMinPrecisionCurve.size(); i++)
        file << " " << compressionMinPrecisionCurve[i].second;
    file << "];" << std::endl;
    file << "unitToMeters = " << *referenceDataset.getUnitsToMeters() << ";" << std::endl;
    file << "plot(compression, precision * unitToMeters, compression, precisionMax * unitToMeters, compression, precisionMin * unitToMeters);" << std::endl;
    file << "legend(\"avg\", \"max\", \"min\");" << std::endl;
    file << "xlabel(\"jpeg compression\");" << std::endl;
    file << "ylabel(\"avg. location error [m]\");" << std::endl;
}



void runSFMAndMVS(const char *dataset, const char *results)
{
    std::cout << "Loading scene" << std::endl;
    syb3r::data::Dataset referenceDataset;
    referenceDataset.loadFromXML(dataset);

    std::vector<float> accuracies;

    syb3r::synthesis::ImagePostprocessor postProcessor;
    postProcessor.setupOldEOS400();

    syb3r::tools::TmpDir tmpDir;
    std::cout << "tmp dir: " << tmpDir.getPath().string() << std::endl;

    std::vector<std::string> ldrFilenames;

    std::cout << "Baking ldr images" << std::endl;
    postProcessor.initialProperties.set("jpegCompression", 100);
    postProcessor.process(referenceDataset, tmpDir.getPath().string(), ldrFilenames);

    syb3r::data::EstimatedCameraPath estimatedCameraPath;


    std::vector<Eigen::Vector3f> estPcld;
    std::vector<Eigen::Vector3f> estColors;


    std::cout << "Running sfm" << std::endl;
    syb3r::wrapper::WrapperVSFM sfm;
    sfm.executablePath = path_to_vsfm;    
    sfm.nvmOutputFilename = (tmpDir.getPath() / "vsfm.nvm").string();
    sfm.producePMVSOutput = true;
    sfm.runSFM(referenceDataset, ldrFilenames, estimatedCameraPath);

    std::cout << "Running mvs" << std::endl;
    syb3r::wrapper::WrapperPMVS pmvs;
    pmvs.executable = path_to_pmvs2;
    pmvs.mipLevel = 1;
    pmvs.generateOptionsFile = true;
    pmvs.numImages = estimatedCameraPath.getCameras().size();
    pmvs.runPreparedSet((tmpDir.getPath() / "vsfm.nvm.cmvs" / "00").string(), "options.txt", estPcld, &estColors);

    std::cout << "Analysing camera path" << std::endl;
    syb3r::analysis::CameraPathEvaluation pathEval(referenceDataset, estimatedCameraPath);

    accuracies.push_back(pathEval.getCameraPositionAccuracy());

    std::fstream file;
    file.exceptions(std::ifstream::failbit | std::ifstream::badbit);
    file.open(results, std::fstream::out);
    file << "accuracies = [";
    for (float f : accuracies)
        file << " " << f;
    file << "];" << std::endl;
    file << "unitToMeters = " << *referenceDataset.getUnitsToMeters() << ";" << std::endl;
    file << "metricAccuracies = accuracies .* unitToMeters;" << std::endl;
    file << "hist(metricAccuracies);" << std::endl;
    file << "xlabel(\"Avg camera location error [meters]\");" << std::endl;
    file << "ylabel(\"Count\");" << std::endl;
    file << "meanAccuracy = mean(metricAccuracies)" << std::endl;
    file << "sdAccuracy = sqrt(mean(metricAccuracies.^2) - meanAccuracy^2)" << std::endl;


    std::vector<Eigen::Vector3f> refPcld;
    std::vector<unsigned> objIds;
    referenceDataset.calcPointCloudFromDepth(refPcld, objIds, 2);
    std::vector<bool> objOfInterest;
    objOfInterest.resize(objIds.size());
    for (unsigned i = 0; i < objIds.size(); i++)
        objOfInterest[i] = objIds[i] != 0;

    float thresh = 0.0005; // 5mm

    syb3r::analysis::PointCloudEvaluation eval(refPcld, objOfInterest, estPcld, pathEval.getEstToRefTransform(), thresh);

    {
        std::fstream file;
        file.exceptions(std::ifstream::failbit | std::ifstream::badbit);
        file.open(std::string(results)+"_mvs_summary.ply", std::fstream::out);
        writePCResults(eval, referenceDataset, file);
    }
    {
        std::fstream file;
        file.exceptions(std::ifstream::failbit | std::ifstream::badbit);
        file.open(std::string(results)+"_mvs_plots.ply", std::fstream::out);
        eval.createMatlabPlots(file, 0.01f, 0.01f, *referenceDataset.getUnitsToMeters());
    }
    {
        std::fstream file;
        file.exceptions(std::ifstream::failbit | std::ifstream::badbit);
        file.open(std::string(results)+"_completeness.ply", std::fstream::out);

        file
            << "ply" << std::endl
            << "format ascii 1.0" << std::endl
            << "element vertex " << refPcld.size() << std::endl
            << "property float x" << std::endl
            << "property float y" << std::endl
            << "property float z" << std::endl
            << "property float quality" << std::endl
            << "end_header" << std::endl;

        for (unsigned i = 0; i < refPcld.size(); i++) {
            file
                << refPcld[i][0] * *referenceDataset.getUnitsToMeters() << " "
                << refPcld[i][1] * *referenceDataset.getUnitsToMeters() << " "
                << refPcld[i][2] * *referenceDataset.getUnitsToMeters() << " "
                << eval.getPerPointCompleteness()[i] * *referenceDataset.getUnitsToMeters() << " "
                << std::endl;
        }
    }

    {
        std::fstream file;
        file.exceptions(std::ifstream::failbit | std::ifstream::badbit);
        file.open(std::string(results)+"_correctness.ply", std::fstream::out);

        file
            << "ply" << std::endl
            << "format ascii 1.0" << std::endl
            << "element vertex " << estPcld.size() << std::endl
            << "property float x" << std::endl
            << "property float y" << std::endl
            << "property float z" << std::endl
            << "property float quality" << std::endl
            << "property uchar red" << std::endl
            << "property uchar green" << std::endl
            << "property uchar blue" << std::endl
            << "end_header" << std::endl;

        for (unsigned i = 0; i < estPcld.size(); i++) {

            Eigen::Vector4f trafoProj = eval.getRefinedEstToRefTrafo() * (Eigen::Vector4f{} << estPcld[i], 1.0f).finished();
            Eigen::Vector3f trafo = trafoProj.head(3) / trafoProj[3];

            file
                << trafo[0] * *referenceDataset.getUnitsToMeters() << " "
                << trafo[1] * *referenceDataset.getUnitsToMeters() << " "
                << trafo[2] * *referenceDataset.getUnitsToMeters() << " "
                << eval.getPerPointCorrectness()[i] * *referenceDataset.getUnitsToMeters() << " "
                << std::min<int>(255, estColors[i][0]*255 + 0.5f) << " "
                << std::min<int>(255, estColors[i][1]*255 + 0.5f) << " "
                << std::min<int>(255, estColors[i][2]*255 + 0.5f) << " "
                << std::endl;
        }
    }

}


void runCompressionVsPcldPrecision(const char *dataset, const char *results)
{
    std::cout << "Loading scene" << std::endl;
    syb3r::data::Dataset referenceDataset;
    referenceDataset.loadFromXML(dataset);


    std::vector<syb3r::data::Camera> calibs;
    for (const auto &image : referenceDataset.getImages())
        calibs.push_back(image.getCamera());


    std::vector<Eigen::Vector3f> refPcld;
    std::vector<unsigned> objIds;
    referenceDataset.calcPointCloudFromDepth(refPcld, objIds, 4);
    std::vector<bool> objOfInterest;
    objOfInterest.resize(objIds.size());
    for (unsigned i = 0; i < objIds.size(); i++)
        objOfInterest[i] = objIds[i] != 0;


    std::vector<unsigned> compression;
    std::vector<float> precision;
    std::vector<float> completeness;

    for (unsigned i = 0; i < 4; i++) {
        syb3r::tools::TmpDir tmpDir;
        std::cout << "tmp dir: " << tmpDir.getPath().string() << std::endl;

        std::vector<std::string> ldrFilenames;

        std::cout << "Baking ldr images" << std::endl;
        syb3r::synthesis::ImagePostprocessor postProcessor;
        postProcessor.setupOldEOS400();
        postProcessor.initialProperties.set("jpegCompression", int(100-i*20));
        postProcessor.process(referenceDataset, tmpDir.getPath().string(), ldrFilenames);

        syb3r::wrapper::WrapperPMVS pmvs;
        pmvs.executable = path_to_pmvs2;        
        pmvs.mipLevel = 1;
        std::vector<Eigen::Vector3f> estPcld;
        pmvs.runUnpreparedSet(calibs, ldrFilenames, estPcld);

        std::cout << "start evaluation" << std::endl;
        float thresh = 0.0005; // 5mm
        syb3r::analysis::PointCloudEvaluation eval(refPcld, objOfInterest, estPcld, Eigen::Matrix4f::Identity(), thresh);

        writePCResults(eval, referenceDataset, std::cout);

        compression.push_back(postProcessor.initialProperties.getFloat("jpegCompression"));
        precision.push_back(eval.getPCCorrectness().avg);
        completeness.push_back(eval.getPCCompleteness().avg);
    }

    std::fstream file;
    file.exceptions(std::ifstream::failbit | std::ifstream::badbit);
    file.open(results, std::fstream::out);
    file << "compression = [";
    for (unsigned i = 0; i < compression.size(); i++)
        file << " " << compression[i];
    file << "];" << std::endl;
    file << "precision = [";
    for (unsigned i = 0; i < precision.size(); i++)
        file << " " << precision[i];
    file << "];" << std::endl;
    file << "completeness = [";
    for (unsigned i = 0; i < completeness.size(); i++)
        file << " " << completeness[i];
    file << "];" << std::endl;
    file << "unitToMeters = " << *referenceDataset.getUnitsToMeters() << ";" << std::endl;
    file << "plot(compression, precision * unitToMeters, compression, completeness * unitToMeters);" << std::endl;
    file << "legend(\"precision\", \"completeness\");" << std::endl;
    file << "xlabel(\"jpeg compression\");" << std::endl;
    file << "ylabel(\"avg. error [m]\");" << std::endl;
}


int main(int, const char**)
{
    syb3r::tools::TaskScheduler::Init(std::thread::hardware_concurrency());

    // run one of the tests like this:
    // Remember to set the paths at the top.
    
    runCompressionVsSFMPrecision("path/to/dataset/scene.xml", "path/to/results");
    
    
    return 0;
}