#include "CameraPathEvaluation.h"

#include <syb3r/data/Dataset.h>
#include <syb3r/data/EstimatedCameraPath.h>

#include <syb3r/math/HelmertTransformation.h>

#include <vector>

namespace syb3r {
namespace analysis {

CameraPathEvaluation::CameraPathEvaluation(const syb3r::data::Dataset &referenceDataset, const syb3r::data::EstimatedCameraPath &estimation){

    // get camera positions from reference and estimate
    std::vector<Eigen::Vector3f> camPos_ref, camPos_est;
    std::vector<unsigned> cameraRefIdx;
    camPos_ref.reserve(estimation.getCameras().size());
    camPos_est.reserve(estimation.getCameras().size());
    cameraRefIdx.reserve(estimation.getCameras().size());
    for (auto estCam : estimation.getCameras()) {
        cameraRefIdx.push_back(estCam.first);
        camPos_ref.push_back(referenceDataset.getImages()[estCam.first].getCamera().getFocalPoint());
        camPos_est.push_back(estCam.second.getFocalPoint());
    }

    // compute the transformation from estimate coordinate system to reference coordinate system
    m_estimationToReferenceTransform = math::computeHelmertTransformation(camPos_est.data(), camPos_ref.data(), camPos_est.size());

    // transform estimated positions to reference coordinate system and compute difference
    for(unsigned c = 0; c < camPos_est.size(); c++){

        Eigen::Matrix<float, 4, 1> cam_est; cam_est << camPos_est[c][0],camPos_est[c][1],camPos_est[c][2],1;
        cam_est = m_estimationToReferenceTransform*cam_est;
        Eigen::Matrix<float, 3, 1> acc;
        // some output
        std::cout << "Camera " << c << " :\t" << cam_est.transpose() << " <--> " << camPos_ref[c].transpose() << std::endl;
        acc << cam_est[0]-camPos_ref[c][0], cam_est[1]-camPos_ref[c][1], cam_est[2]-camPos_ref[c][2];
        m_cameraAccuracies[cameraRefIdx[c]].positionAccuracy = acc;
    }
    // compute average accuracy over all dimensions
    float s;
    float a = getCameraPositionAccuracy(-1, &s);
    // some output
    std::cout << "Mean position accuracy (in meters): " << a * *referenceDataset.getUnitsToMeters() << " +- " << s * *referenceDataset.getUnitsToMeters() << std::endl;
}


// returns the absolute difference or average distance (and standard deviation) between estimated and reference position for a specific camera or all cameras for all three dimensions
Eigen::Vector3f CameraPathEvaluation::getCameraPositionAccuracyPerDimension(int id, Eigen::Vector3f* std) const{
    if (id>0){
       Eigen::Vector3f a = m_cameraAccuracies.find(id)->second.positionAccuracy;
       a(0) = std::abs(a(0));a(1) = std::abs(a(1));a(2) = std::abs(a(2));
       return a;
   }else{
       Eigen::Vector3f avg;
       avg << 0,0,0;
       if (std) *std << 0,0,0;

       for(const auto &cam : m_cameraAccuracies){
         avg(0) += cam.second.positionAccuracy[0];avg(1) += cam.second.positionAccuracy[1];avg(2) += cam.second.positionAccuracy[2];
         if (std) {(*std)(0) += std::pow(cam.second.positionAccuracy[0],2);(*std)(1) += std::pow(cam.second.positionAccuracy[1],2);(*std)(2) += std::pow(cam.second.positionAccuracy[2],2);}
       }
       avg /= m_cameraAccuracies.size();
       if (std) {
          (*std)(0) = std::sqrt((*std)(0)/m_cameraAccuracies.size() - std::pow(avg(0),2));
          (*std)(1) = std::sqrt((*std)(1)/m_cameraAccuracies.size() - std::pow(avg(1),2));
          (*std)(2) = std::sqrt((*std)(2)/m_cameraAccuracies.size() - std::pow(avg(2),2));
       }
       return avg;
   }
}

// returns the (average) distance (and standard deviation) between estimated and reference position for a specific camera or all cameras over all three dimensions
float CameraPathEvaluation::getCameraPositionAccuracy(int id, float* std) const {
    if (id>0){
       return m_cameraAccuracies.find(id)->second.positionAccuracy.norm();
    }else{
         float avg = 0;
         if (std) *std = 0;
         for(const auto &cam : m_cameraAccuracies){
            avg += cam.second.positionAccuracy.norm();
            if (std) *std += cam.second.positionAccuracy.squaredNorm();
         }
         avg /= m_cameraAccuracies.size();
         if (std) *std = std::sqrt(*std/m_cameraAccuracies.size() - std::pow(avg,2));
         return avg;
    }
    
}

}
}
