#ifndef CAMERAPATHEVALUATION_H
#define CAMERAPATHEVALUATION_H

#include <map>

#include <Eigen/Dense>

namespace syb3r {

namespace data {
    class Dataset;
    class EstimatedCameraPath;
}

namespace analysis {

/**
 * @brief Compares the given camera position with the estimate from SfM
 */
class CameraPathEvaluation
{
    public:
        /**
          * @param referenceDataset Reference dataset including given camera position and orientation
          * @param estimation Estimated dataset from SfM
        */
        CameraPathEvaluation(const syb3r::data::Dataset &referenceDataset, const syb3r::data::EstimatedCameraPath &estimation);

        struct CameraAccuracy {
            /// distance between positions of each camera in reference and estimate in all three dimensions
            Eigen::Vector3f positionAccuracy;
            /// angular distance between pose of each camera in reference and estimate in all three dimensions
            Eigen::Vector3f rotationAccuracy;
        };

        /// method returns position error per dimensions of a specific camera (identified via cam ID) or all cameras
        /**
          * @param id Specifies for which camera the position error should be retrieved, if id=-1 the average error of all cameras will be returned
          * @param std will contain the standard deviation of the position error of all cameras if provided (note: only active if id=-1)
        */
        Eigen::Vector3f getCameraPositionAccuracyPerDimension(int id=-1, Eigen::Vector3f* std=0) const;
        /// method returns position error of a specific camera (identified via cam ID) or all cameras
        /**
          * @param id Specifies for which camera the position error should be retrieved, if id=-1 the average error of all cameras will be returned
          * @param std will contain the standard deviation of the position error of all cameras if provided (note: only active if id=-1)
        */
        float getCameraPositionAccuracy(int id=-1, float* std=0) const;

        inline const Eigen::Matrix4f &getEstToRefTransform() const { return m_estimationToReferenceTransform; }
    protected:
        /// 4x4 transformation matrix from the coordinate system of the estimate to the reference coordinate system
        Eigen::Matrix4f m_estimationToReferenceTransform;
        /// contains the individual errors of all cameras
        std::map<unsigned, CameraAccuracy> m_cameraAccuracies;
};


}
}

#endif // CAMERAPATHEVALUATION_H
