#include "PointCloudEvaluation.h"

#include <syb3r/math/FractionalICP.h>

namespace syb3r {
namespace analysis {

bool isFinite(float f)
{
    union {
        float f;
        unsigned int i;
    } floatToInt;

    floatToInt.f = f;
    return ((floatToInt.i >> 23) & 0xFF) != 0xFF;
}

PointCloudEvaluation::PointCloudEvaluation(const std::vector<Eigen::Vector3f> &refPcld, const std::vector<bool>& objectOfInterest,
                   const std::vector<Eigen::Vector3f> &estPcld, const Eigen::Matrix4f &estToRefTrafo, float thresh){

	// ********************************
	// POINT CLOUD

    float resolution = 0.01f;

    std::cout << "get reference point cloud: start" << std::endl;
    // convert to pcl cloud
    pcl::PointCloud<pcl::PointXYZ>::Ptr ref_cloud (new pcl::PointCloud<pcl::PointXYZ>);
    eigen2pcl(refPcld, ref_cloud);
    std::cout << "get reference point cloud: done" << std::endl;
    //ref.saveToPLY(ref_cloud, "pc_ref.ply");

    std::cout << "load estimated point cloud: start" << std::endl;
    // load estimated point cloud
    pcl::PointCloud<pcl::PointXYZ>::Ptr est_cloud (new pcl::PointCloud<pcl::PointXYZ>);

    std::vector<Eigen::Vector3f> alignedEstPcld = estPcld;
    transformPointCloud(alignedEstPcld, estToRefTrafo);
#if 0
    // if necessary, point clouds can be further aligned by ICP
    math::ICPTransformHelmert helmertIcp;
    math::fractionalICP(estPcld, refPcld, helmertIcp);

    std::cout << "Rotation: " << std::endl
                << helmertIcp.getRotation() << std::endl<< std::endl;
    std::cout << "Scale: " << std::endl
                << helmertIcp.getScale() << std::endl<< std::endl;
    std::cout << "Offset: " << std::endl
                << helmertIcp.getOffset() << std::endl<< std::endl;


    m_estToRefTransform = Eigen::Matrix4f::Identity();
    for (unsigned i = 0; i < 3; i++)
        for (unsigned j = 0; j < 3; j++)
            m_estToRefTransform(i, j) = helmertIcp.getRotation()(i, j) * helmertIcp.getScale();

    for (unsigned i = 0; i < 3; i++)
        m_estToRefTransform(i, 3) = helmertIcp.getOffset()[i];

    transformPointCloud(alignedEstPcld, m_estToRefTransform);
    m_estToRefTransform = m_estToRefTransform * estToRefTrafo;
#else
    m_estToRefTransform = estToRefTrafo;
#endif
    eigen2pcl(alignedEstPcld, est_cloud);
	
	// built octree
	pcl::octree::OctreePointCloudSearch<pcl::PointXYZ> ref_octree(resolution);
	ref_octree.setInputCloud(ref_cloud);
	ref_octree.addPointsFromInputCloud();
	
	pcl::octree::OctreePointCloudSearch<pcl::PointXYZ> est_octree(resolution);
	est_octree.setInputCloud(est_cloud);
	est_octree.addPointsFromInputCloud();
	
   // compute correctness and completeness
	// only points on the object will contribute
	std::cout << "calculate correctness" << std::endl;
	computeCorrectness(ref_octree, est_cloud, thresh, objectOfInterest);
	std::cout << "calculate completeness" << std::endl;
	computeCompleteness(est_octree, ref_cloud, thresh, objectOfInterest);

	// ********************************
}

void PointCloudEvaluation::transformPointCloud(pcl::PointCloud<pcl::PointXYZ>::Ptr cloud, const Eigen::Matrix4f& H){
	for(unsigned p=0; p<cloud->size(); p++){
		pcl::PointXYZ pt = cloud->points[p];
		cloud->points[p].x = H(0,0)*pt.x + H(0,1)*pt.y + H(0,2)*pt.z + H(0,3);
		cloud->points[p].y = H(1,0)*pt.x + H(1,1)*pt.y + H(1,2)*pt.z + H(1,3);
		cloud->points[p].z = H(2,0)*pt.x + H(2,1)*pt.y + H(2,2)*pt.z + H(2,3);
		float w = H(3,0)*pt.x + H(3,1)*pt.y + H(3,2)*pt.z + H(3,3);
		cloud->points[p].x /= w;
		cloud->points[p].y /= w;
		cloud->points[p].z /= w;
	}
}

void PointCloudEvaluation::transformPointCloud(std::vector<Eigen::Vector3f> &points, const Eigen::Matrix4f &transform)
{
    for (auto &point : points) {
        Eigen::Vector4f transformed = transform * (Eigen::Vector4f() << point, 1.0f).finished();
        point = transformed.head<3>() * (1.0f / transformed[3]);
    }
}

void PointCloudEvaluation::computeCorrectness(pcl::octree::OctreePointCloudSearch<pcl::PointXYZ>& octree, pcl::PointCloud<pcl::PointXYZ>::Ptr cloud, float thresh, const std::vector<bool>& objectOfInterest)
{
	nnAnalysis(octree, cloud, thresh, this->m_cloudCorrectness, objectOfInterest, m_correctness, false);
}

void PointCloudEvaluation::computeCompleteness(pcl::octree::OctreePointCloudSearch<pcl::PointXYZ>& octree, pcl::PointCloud<pcl::PointXYZ>::Ptr cloud, float thresh, const std::vector<bool>& objectOfInterest){
	nnAnalysis(octree, cloud, thresh, this->m_cloudCompleteness, objectOfInterest, m_completeness, true);
}
	
void PointCloudEvaluation::nnAnalysis(pcl::octree::OctreePointCloudSearch<pcl::PointXYZ>& octree, pcl::PointCloud<pcl::PointXYZ>::Ptr cloud, float thresh, Statistics& stats, const std::vector<bool>& objectOfInterest, std::vector<float> &distances, bool computeCompleteness){
	
   int K = 1;
    
   // some stats
   int n=0, nCorrect=0;
	float d, mind, maxd, sumd, sumd2;
	mind = 1000; maxd = sumd = sumd2 = 0;
	
	// for display
	int num = cloud->size();
	//int step = num/100;   // only needed for some progress output on std

    distances.resize(num, -1.0f);
	
	// loop through all points
	//std::cout << std::endl;
	for(int p = 0; p < num; p++){
		
		if (computeCompleteness && !objectOfInterest[p]) continue;
		
		/*if (p%step == 0){
			std::cout << ((p*100.)/num) << "% ( " << mind << " , " << (sumd/n) << " , " << maxd << ")\r";
			std::cout.flush();
		}*/
		std::vector<int> pointIdxNKNSearch;
		std::vector<float> pointNKNSquaredDistance;
		// get closest nearest neighbor
		if (octree.nearestKSearch(cloud->points[p], K, pointIdxNKNSearch, pointNKNSquaredDistance) > 0){
			
			if (!computeCompleteness && !objectOfInterest[pointIdxNKNSearch[0]]) continue;
			
			// save stats
			d = pointNKNSquaredDistance[0];
			sumd2 += d;
			d = std::sqrt(d);
         distances[p] = d;
			sumd += d;
			if (mind>d) mind = d;
			if (maxd<d) maxd = d;
			n++;
			if (d < thresh) nCorrect++;
		}
	}
	// std::cout << std::endl;
	// save stats
	stats.min = mind;
	stats.max = maxd;
	stats.avg = sumd/n;
	stats.stddev = std::sqrt(sumd2/n - std::pow(sumd/n,2));
	stats.numCorrect = nCorrect;
	stats.numTotal = n;
}

void PointCloudEvaluation::eigen2pcl(const std::vector<Eigen::Vector3f>& pc1, pcl::PointCloud<pcl::PointXYZ>::Ptr pc2){
	pc2->width = pc1.size();
	pc2->height = 1;
	pc2->points.resize(pc2->width * pc2->height);
	for(unsigned p = 0; p < pc1.size(); p++){
		Eigen::Matrix<float, 3, 1> pt = pc1.at(p);
		pc2->points[p].x = pt[0];
		pc2->points[p].y = pt[1];
		pc2->points[p].z = pt[2];
	}
}


float PointCloudEvaluation::get95PercentPrecision() const
{
    std::vector<float> tmp;
    tmp.reserve(m_correctness.size());
    for (float precision : m_correctness)
        if (precision >= 0.0f) tmp.push_back(precision);

    std::sort(tmp.begin(), tmp.end());
    return tmp[tmp.size()*95/100];
}

float PointCloudEvaluation::get95PercentCompleteness() const
{
    std::vector<float> tmp;
    tmp.reserve(m_completeness.size());
    for (float completeness : m_completeness)
        if (completeness >= 0.0f) tmp.push_back(completeness);

    std::sort(tmp.begin(), tmp.end());
    return tmp[tmp.size()*95/100];
}

void PointCloudEvaluation::createMatlabPlots(std::ostream &stream, float precCutoff, float complCutoff, float metricScale) const
{
    unsigned numBarsPrec = 50;
    float errorToBarPrec = numBarsPrec / precCutoff;

    std::vector<unsigned> precisionBars;
    precisionBars.resize(numBarsPrec, 0u);
    for (float precision : m_correctness) {
        if (precision < 0.0f) continue;
        if (precision > precCutoff) continue;
        int idx = std::min<int>(std::max<int>(precision * errorToBarPrec, 0), numBarsPrec-1);
        precisionBars[idx]++;
    }


    unsigned numBarsCompl = 50;
    float errorToBarCompl = numBarsCompl / complCutoff;

    std::vector<unsigned> completenessBars;
    completenessBars.resize(numBarsCompl, 0u);
    for (float completeness : m_completeness) {
        if (completeness < 0.0f) continue;
        if (completeness > complCutoff) continue;
        int idx = std::min<int>(std::max<int>(completeness * errorToBarCompl, 0), numBarsCompl-1);
        completenessBars[idx]++;
    }

    stream << "barErrorsPrec_mm = ([0:1:"<<numBarsPrec-1<<"] + 0.5) / "<<numBarsPrec<<" * "<<precCutoff * metricScale<<";"<<std::endl;
    stream << "precBars = [";
    for (unsigned count : precisionBars)
        stream << " " << count;
    stream << "];"<<std::endl;

    stream << "barErrorsCompl_mm = ([0:1:"<<numBarsCompl-1<<"] + 0.5) / "<<numBarsCompl<<" * "<<complCutoff * metricScale<<";"<<std::endl;
    stream << "complBars = [";
    for (unsigned count : completenessBars)
        stream << " " << count;
    stream << "];"<<std::endl;

    stream << "subplot(\"position\",[0.05, 0.15, 0.4, 0.8]);" << std::endl;
    stream << "bar(barErrorsPrec_mm, precBars, \"hist\");" << std::endl;
    stream << "xlabel('Error [meters]');" << std::endl;
    stream << "ylabel('#points');" << std::endl;
    stream << "title('Precision');" << std::endl;

    stream << "subplot(\"position\",[0.55, 0.15, 0.4, 0.8]);" << std::endl;
    stream << "bar(barErrorsCompl_mm, complBars, \"hist\");" << std::endl;
    stream << "xlabel('Error [meters]');" << std::endl;
    stream << "ylabel('#points');" << std::endl;
    stream << "title('Completeness');" << std::endl;

    stream << "subplot(\"position\",[0.05, 0.025, 0.4, 0.05]);" << std::endl;
    stream << "axis(\"off\");" << std::endl;
    stream << "box(\"off\");" << std::endl;
    stream << "text(0, 0.5, \" Avg. precision:\\n  " << m_cloudCorrectness.avg * metricScale << " (meters)\\nsigma:\\n  "<< m_cloudCorrectness.stddev * metricScale << " (meters)\");" << std::endl;

    stream << "subplot(\"position\",[0.55, 0.025, 0.4, 0.05]);" << std::endl;
    stream << "axis(\"off\");" << std::endl;
    stream << "box(\"off\");" << std::endl;
    stream << "text(0, 0.5, \" Avg. completeness:\\n  " << m_cloudCompleteness.avg * metricScale << " (meters)\\nsigma:\\n  "<< m_cloudCompleteness.stddev * metricScale << " (meters)\");" << std::endl;
}

}
}
