#ifndef POINTCLOUDEVALUATION_H
#define POINTCLOUDEVALUATION_H

#include <iostream>
#include <fstream>

#include <Eigen/Dense>
#include <stdlib.h>
#include <vector>

#include <pcl/point_cloud.h>
#include <pcl/octree/octree.h>
#include <pcl/registration/icp.h>

#include <syb3r/data/Dataset.h>
#include <syb3r/data/Image.h>
#include <syb3r/data/Camera.h>
#include <syb3r/math/HelmertTransformation.h>

namespace syb3r {
namespace analysis {

class PointCloudEvaluation
{
    public:
        /// computation of performance measures
        /**
          * @param refPcld Reference point cloud
          * @param objectOfInterest denotes which pixels belong to the object of interest.
          * @param estPcld Estimated point cloud
          * @param estToRefTrafo 4x4 tranformation matrix from coordinate system of the estimate to the reference
          * @param thresh Maximal distance between points in both point clouds to count them as "correct" match
        */
        PointCloudEvaluation(const std::vector<Eigen::Vector3f> &refPcld, const std::vector<bool>& objectOfInterest,
                    const std::vector<Eigen::Vector3f> &estPcld, const Eigen::Matrix4f &estToRefTrafo, float thresh);

        /// bundles several statistical values (minimal, maximal, average values as well as standard deviation, and number of all and of correctly estimated points
        struct Statistics {
            float min, max, avg, stddev;
            unsigned numCorrect, numTotal;
        };

        // access to computed performance
        // sparse/dense point cloud estimation
        inline const Statistics &getPCCorrectness() const { return m_cloudCorrectness; }
        inline const Statistics &getPCCompleteness() const { return m_cloudCompleteness; }
        // mesh might come later?

        /// returns correctness per point, i.e. distance to closest point in reference point cloud
        inline const std::vector<float> &getPerPointCorrectness() const { return m_correctness; }
        /// returns completeness per point, i.e. distance to closest point in estimated point cloud
        inline const std::vector<float> &getPerPointCompleteness() const { return m_completeness; }
        /// returns used transformation matrix from coordinate system of the estimate to the reference
        inline const Eigen::Matrix4f &getRefinedEstToRefTrafo() const { return m_estToRefTransform; }

        /// returns precision value p such that 95% of the points have a precision larger than p
        float get95PercentPrecision() const;
        /// returns completeness value c such that 95% of the points have a completeness larger than c
        float get95PercentCompleteness() const;
        /// creates a script to generate matlab plots summarizing the computed statistics
        void createMatlabPlots(std::ostream &stream, float precCutoff, float complCutoff, float metricScale) const;

    protected:
        /// nearest neighbor analyis between two point clouds
        void nnAnalysis(pcl::octree::OctreePointCloudSearch<pcl::PointXYZ>&, pcl::PointCloud<pcl::PointXYZ>::Ptr, float, Statistics&, const std::vector<bool>& objectOfInterest, std::vector<float> &distances, bool computeCompleteness);
        /// distance from each point in estimated pc to closest point in reference pc
        void computeCorrectness(pcl::octree::OctreePointCloudSearch<pcl::PointXYZ>&, pcl::PointCloud<pcl::PointXYZ>::Ptr, float, const std::vector<bool>& objectOfInterest);
        /// distance from each point in reference pc to closest point in estimated pc
        void computeCompleteness(pcl::octree::OctreePointCloudSearch<pcl::PointXYZ>&, pcl::PointCloud<pcl::PointXYZ>::Ptr, float, const std::vector<bool>& objectOfInterest);
        // convert vector of eigen-points to pcl point cloud
        void eigen2pcl(const std::vector<Eigen::Vector3f>&, pcl::PointCloud<pcl::PointXYZ>::Ptr);
        // apply 3D homography to pcl point cloud
        void transformPointCloud(pcl::PointCloud<pcl::PointXYZ>::Ptr, const Eigen::Matrix4f&);
        static void transformPointCloud(std::vector<Eigen::Vector3f> &points, const Eigen::Matrix4f &transform);

        /// statistics over correctness
        Statistics m_cloudCorrectness;
        /// statistics over completeness
        Statistics m_cloudCompleteness;

        std::vector<float> m_correctness;
        std::vector<float> m_completeness;

        /// 4x4 transformation matrix from coordinate system of the estimate to the reference
        Eigen::Matrix4f m_estToRefTransform;

};

}
}

#endif // POINTCLOUDEVALUATION_H
