#include "Camera.h"

#include <stdexcept>

#include <boost/format.hpp>
#include <tinyxml2.h>

#include "../math/CameraProjectionMath.h"

namespace syb3r {
namespace data {


Camera Camera::convertFrom_PMVS_Matrix(const Eigen::Matrix<float, 3, 4> &pmvsProjectionMatrix, unsigned width, unsigned height)
{
    Eigen::Matrix3f internalCalibration;
    Eigen::Matrix3f rotation;
    Eigen::Vector3f position;

    // PMVS stores its projection matrix the same way we do, so factortizing is sufficient
    math::decomposeProjectionMatrix(pmvsProjectionMatrix, internalCalibration, rotation, position);
    return Camera(internalCalibration, rotation, position, width, height);
}

Eigen::Matrix<float, 3, 4> Camera::convertTo_PMVS_Matrix() const
{
    // PMVS stores its projection matrix the same way we do, so multiplication is sufficient
    return m_internalCalibration * Eigen::Matrix<float, 3, 4>::Identity() * m_externalCalibration;
}


Camera Camera::convertFrom_VSFM_NVM(float focalLength, const Eigen::Quaternionf &cameraRotation, const Eigen::Vector3f &cameraPosition, unsigned width, unsigned height)
{
    // VSFM has the same view space coordinate system as we do
    Eigen::Matrix3f cameraRotationMat = cameraRotation.matrix();

    // VSFM assumes the image center to be the principal point. Focal length is in pixels.
    // @todo: Check on half pixel offset!
    Eigen::Matrix3f internalCalibration;
    internalCalibration <<
        focalLength,    0.0f,           width*0.5f,
        0.0f,           focalLength,    height*0.5f,
        0.0f,           0.0f,           1.0f;

    return Camera(internalCalibration, cameraRotationMat, cameraPosition, width, height);
}


Camera Camera::convertFrom_DFM_SubsBA(const Eigen::Matrix<float, 3, 4> &DFM_SubsBA_ProjectionMatrix, unsigned width, unsigned height)
{
    Eigen::Matrix3f internalCalibration;
    Eigen::Matrix3f rotation;
    Eigen::Vector3f position;

    math::decomposeProjectionMatrix(DFM_SubsBA_ProjectionMatrix, internalCalibration, rotation, position);

    Eigen::Matrix3f imageMatrix = Eigen::Matrix3f::Identity();
    imageMatrix(0, 0) = width * 0.5f;
    imageMatrix(1, 1) = width * 0.5f;

    imageMatrix(0, 2) = width * 0.5f - 0.5f;
    imageMatrix(1, 2) = height * 0.5f - 0.5f;

    return Camera(imageMatrix * internalCalibration, rotation, position, width, height);
}



template<unsigned rows, unsigned cols, unsigned readRows, unsigned readCols>
Eigen::Matrix<float, rows, cols> readMatrixFromXmlAttribs(const tinyxml2::XMLElement *node, const std::string &errorMsg)
{
    Eigen::Matrix<float, rows, cols> result;
    result = Eigen::Matrix<float, rows, cols>::Identity();

    for (unsigned row = 0; row < readRows; row++)
        for (unsigned col = 0; col < readCols; col++) {
            std::string attribName = (boost::format("mat_%01d%01d") % row % col).str();
            switch (node->QueryFloatAttribute(attribName.c_str(), &result(row, col))) {
                case tinyxml2::XML_NO_ATTRIBUTE:
                    throw std::runtime_error(errorMsg+" attribute '"+attribName+"' missing!");
                case tinyxml2::XML_WRONG_ATTRIBUTE_TYPE:
                    throw std::runtime_error(errorMsg+" attribute '"+attribName+"' is not a float value!");
                default:
                    // all good
                    break;
            }
        }

    return result;
}

Camera Camera::readFromSceneXML(const tinyxml2::XMLElement *cameraNode)
{
    const tinyxml2::XMLElement *extCalibNode = cameraNode->FirstChildElement("ext_calib");
    if (extCalibNode == nullptr) throw std::runtime_error("Missing 'ext_calib' node!");

    Eigen::Matrix4f externalCalibration = readMatrixFromXmlAttribs<4, 4, 3, 4>(extCalibNode, "Error reading 'ext_calib' node:");

    const tinyxml2::XMLElement *intCalibNode = cameraNode->FirstChildElement("int_calib");
    if (intCalibNode == nullptr) throw std::runtime_error("Missing 'int_calib' node!");

    Eigen::Matrix3f internalCalibration = readMatrixFromXmlAttribs<3, 3, 3, 3>(intCalibNode, "Error reading 'int_calib' node:");

    const tinyxml2::XMLElement *resolutionNode = cameraNode->FirstChildElement("resolution");
    if (resolutionNode == nullptr) throw std::runtime_error("Missing 'resolution' node!");

    unsigned width;
    if (resolutionNode->QueryUnsignedAttribute("width", &width) != tinyxml2::XML_NO_ERROR)
        throw std::runtime_error("Error reading 'width' attribute of 'resolution' node");

    unsigned height;
    if (resolutionNode->QueryUnsignedAttribute("height", &height) != tinyxml2::XML_NO_ERROR)
        throw std::runtime_error("Error reading 'height' attribute of 'resolution' node");

    return Camera(internalCalibration, externalCalibration, width, height);
}


Camera::Camera(const Eigen::Matrix3f &internalCalibration, const Eigen::Matrix4f &externalCalibration, unsigned width, unsigned height) :
        m_internalCalibration(internalCalibration), m_externalCalibration(externalCalibration), m_width(width), m_height(height)
{
    computeRotationPositionFromExtCalib();

    m_radialDistortion[0].setZero();
    m_radialDistortion[1].setZero();
    m_radialDistortion[2].setZero();
}

Camera::Camera(const Eigen::Matrix3f &internalCalibration, const Eigen::Matrix3f &cameraRotation, const Eigen::Vector3f &cameraPosition, unsigned width, unsigned height) :
        m_internalCalibration(internalCalibration), m_rotation(cameraRotation), m_focalPoint(cameraPosition), m_width(width), m_height(height)
{
    Eigen::Vector3f offset = m_rotation * -m_focalPoint;
    for (unsigned i = 0; i < 3; i++) {
        for (unsigned j = 0; j < 3; j++)
            m_externalCalibration(i, j) = m_rotation(i, j);
        m_externalCalibration(i, 3) = offset[i];
    }
    m_externalCalibration(3, 0) = 0.0f;
    m_externalCalibration(3, 1) = 0.0f;
    m_externalCalibration(3, 2) = 0.0f;
    m_externalCalibration(3, 3) = 1.0f;

    m_radialDistortion[0].setZero();
    m_radialDistortion[1].setZero();
    m_radialDistortion[2].setZero();
}


void Camera::computeRotationPositionFromExtCalib()
{
    for (unsigned i = 0; i < 3; i++) {
        for (unsigned j = 0; j < 3; j++)
            m_rotation(i, j) = m_externalCalibration(i, j);
        m_focalPoint[i] = m_externalCalibration(i, 3);
    }
    m_externalCalibration(3, 0) = 0.0f;
    m_externalCalibration(3, 1) = 0.0f;
    m_externalCalibration(3, 2) = 0.0f;
    m_externalCalibration(3, 3) = 1.0f;
    m_focalPoint = m_rotation.transpose() * -m_focalPoint;
}

}
}

