#ifndef CAMERA_H
#define CAMERA_H

#include <Eigen/Dense>

namespace tinyxml2 {
    class XMLElement;
}

namespace syb3r {
namespace data {

/**
 * @brief Internal and external camera parameters
 * @details To be used for the ground truth as well as for the estimations.
 */
class Camera
{
    public:
        /// @name Conversion
        /// @todo move
        ///@{
        static Camera convertFrom_PMVS_Matrix(const Eigen::Matrix<float, 3, 4> &pmvsProjectionMatrix, unsigned width, unsigned height);
        Eigen::Matrix<float, 3, 4> convertTo_PMVS_Matrix() const;

        static Camera convertFrom_VSFM_NVM(float focalLength, const Eigen::Quaternionf &cameraRotation, const Eigen::Vector3f &cameraPosition, unsigned width, unsigned height);

        static Camera convertFrom_DFM_SubsBA(const Eigen::Matrix<float, 3, 4> &DFM_SubsBA_ProjectionMatrix, unsigned width, unsigned height);

        static Camera readFromSceneXML(const tinyxml2::XMLElement *cameraNode);
        ///@}

        Camera() = delete;
        Camera(const Camera&) = default;
        Camera(const Eigen::Matrix3f &internalCalibration, const Eigen::Matrix4f &externalCalibration, unsigned width, unsigned height);
        Camera(const Eigen::Matrix3f &internalCalibration, const Eigen::Matrix3f &cameraRotation, const Eigen::Vector3f &cameraPosition, unsigned width, unsigned height);

        /**
         * @brief Transforms from view space to image space
         * @details Image space is [0 .. width] horizontally and [0 .. height] vertically, incrementing towards right and bottom respectively.
         * Pixels span whole number intervals, eg. the top left pixel spans the quad from 0..1 in both x and y directions, it's center thus
         * being located at (0.5, 0.5)
         */
        inline const Eigen::Matrix3f &getInternalCalibration() const { return m_internalCalibration; }

        /// Polynomial radial distortion with one vec3 for every coefficient to control the colors individually
        /// @todo This will probably have to be way more flexible!
        inline const Eigen::Vector3f &getRadialDistortion(unsigned colorChannel) const { return m_radialDistortion[colorChannel]; }

        /**
         * @brief Transforms from world space to view space
         * @details In view space, the camera is located at the origin and looks in the z+ direction. The x+ direction is to the right and y+ is downwards.
         */
        inline const Eigen::Matrix4f &getExternalCalibration() const { return m_externalCalibration; }

        /// Returns the position of the camera
        inline const Eigen::Vector3f &getFocalPoint() const { return m_focalPoint; }

        /// Returns the camera rotation from world space to view space
        inline const Eigen::Matrix3f &getRotation() const { return m_rotation; }

        /// Width (in pixels) of the produced images
        inline unsigned getWidth() const { return m_width; }
        /// Height (in pixels) of the produced images
        inline unsigned getHeight() const { return m_height; }

        /// Returns the horizontal focal length in 35mm equivalent
        /// @details This is not necessarily the lens that was actually used. The sensor might not be full frame.
        inline float getFocalLength35mm() const { return m_internalCalibration(0, 0) / m_width * 36.0f; }
    protected:
        Eigen::Matrix3f m_internalCalibration;
        Eigen::Vector3f m_radialDistortion[3];
        Eigen::Matrix4f m_externalCalibration;
        Eigen::Matrix3f m_rotation;
        Eigen::Vector3f m_focalPoint;

        unsigned m_width;
        unsigned m_height;

        void computeRotationPositionFromExtCalib();

};

}
}

#endif // CAMERA_H
