#include "Dataset.h"

#include "../math/CameraProjectionMath.h"

#include "../tools/TaskScheduler.h"

#include <mutex>

#include <boost/format.hpp>
#include <boost/filesystem.hpp>

namespace syb3r {
namespace data {

void Dataset::calcPointCloudFromDepth(std::vector<Eigen::Vector3f>& pc, std::vector<unsigned>& mask, unsigned subsamplingStride) const{
#if 0
    // for each frame do
    for (unsigned i = 0; i < m_images.size(); i++){
        std::cout << "get point cloud " << i << " : ";
        std::cout.flush();
        m_images[i].extractPointcloud(pc, &mask);
        std::cout << "done" << std::endl;
    }
#else
    tools::TaskGroup group;
    std::mutex pcmutex;
    // for each frame do
    for (unsigned i = 0; i < m_images.size(); i++){
        group.add([this, i, &pc, &pcmutex, &mask, &subsamplingStride]{
            std::vector<Eigen::Vector3f> partialCloud;
            std::vector<unsigned> partialMask;
            m_images[i].extractPointcloud(partialCloud, &partialMask, subsamplingStride);

            {
                std::lock_guard<std::mutex> lock(pcmutex);
                std::cout << "point cloud " << i << " done" << std::endl;

                assert(pc.size() == mask.size());
                unsigned offset = pc.size();
                pc.resize(offset + partialCloud.size());
                memcpy(pc.data()+offset, partialCloud.data(), partialCloud.size() * sizeof(Eigen::Vector3f));

                mask.resize(offset + partialMask.size());
                memcpy(mask.data()+offset, partialMask.data(), partialMask.size() * sizeof(unsigned));
            }

        }, tools::TaskScheduler::get());
    }
    tools::TaskScheduler::get().waitFor(&group);
#endif
}


void Dataset::loadFromXML(const boost::filesystem::path &xmlFilename){

    m_unitsToMeters.reset();
    m_images.clear();

    const boost::filesystem::path path = xmlFilename.parent_path();

    tinyxml2::XMLDocument xmlDoc;
    if (xmlDoc.LoadFile(xmlFilename.string().c_str()) != tinyxml2::XML_NO_ERROR)
        throw std::runtime_error(std::string("Could not open or parse scene xml file! (") + xmlFilename.string() + ")");

    tinyxml2::XMLElement* pRoot = xmlDoc.FirstChildElement("Scene");
    if (pRoot == nullptr) pRoot = xmlDoc.FirstChildElement("CamNetwork"); // fallback for older versions.
    if (pRoot == nullptr) throw std::runtime_error("No 'CamNetwork' or 'Scene' node in xml file!");

    tinyxml2::XMLElement* sceneScaleNode = pRoot->FirstChildElement("SceneScale");
    if (sceneScaleNode != nullptr) {
        float f;
        switch (sceneScaleNode->QueryFloatAttribute("units_to_meters", &f)) {
            case tinyxml2::XML_NO_ATTRIBUTE:
                break;
            case tinyxml2::XML_WRONG_ATTRIBUTE_TYPE:
                throw std::runtime_error("units_to_meters attribute is not a float number!");
            default:
                m_unitsToMeters = f;
        }
    }

    tinyxml2::XMLElement* pElement = pRoot->FirstChildElement("Cameras");
    if (pElement == nullptr) throw std::runtime_error("No 'Cameras' node in xml file!");

    for (tinyxml2::XMLElement* pListElement = pElement->FirstChildElement("Cam");
            pListElement != nullptr;
            pListElement = pListElement->NextSiblingElement("Cam")) {

        addImage(Image::readFromSceneXML(pListElement, path));
    }
}


}
}
