#include "Image.h"

#include "../math/CameraProjectionMath.h"

#include <OpenEXR/ImfInputFile.h>
#include <OpenEXR/ImfArray.h>
#include <OpenEXR/ImathBox.h>

#include <tinyxml2.h>

namespace syb3r {
namespace data {

namespace {

boost::filesystem::path readImageFilenameFromXmlNode(const tinyxml2::XMLElement* node, const std::string &errorMsg)
{
    if (node == nullptr)
        throw std::runtime_error(errorMsg + " node not found!");
    const char *attrib = node->Attribute("filename");
    if (attrib == nullptr)
        throw std::runtime_error(errorMsg + " no 'filename' attrib!");
    return boost::filesystem::path(attrib);
}

}

Image Image::readFromSceneXML(const tinyxml2::XMLElement *imageNode, const boost::filesystem::path &sceneRoot)
{
    Image result(Camera::readFromSceneXML(imageNode),
                 sceneRoot / readImageFilenameFromXmlNode(imageNode->FirstChildElement("hdr_image"), "Error reading hdr image filename:"),
                 sceneRoot / readImageFilenameFromXmlNode(imageNode->FirstChildElement("attrib_image"), "Error reading attrib image filename:"));

    if (imageNode->QueryUnsignedAttribute("frame", &result.m_frameIdx) == tinyxml2::XML_WRONG_ATTRIBUTE_TYPE) {
        throw std::runtime_error("Could not parse camera frame number!");
    }

    {
        const char *c = imageNode->Attribute("name");
        if (c != nullptr)
            result.m_cameraName = c;
    }


    const tinyxml2::XMLElement *lensNode = imageNode->FirstChildElement("lens");
    if (lensNode != nullptr) {
        float f;
        switch (lensNode->QueryFloatAttribute("focal_length_mm", &f)) {
            case tinyxml2::XML_NO_ATTRIBUTE:
                break;
            case tinyxml2::XML_WRONG_ATTRIBUTE_TYPE:
                throw std::runtime_error("Focal_length_mm attribute is not a float number!");
            default:
                result.m_focalLength = f;
        }

        switch (lensNode->QueryFloatAttribute("f_number", &f)) {
            case tinyxml2::XML_NO_ATTRIBUTE:
                break;
            case tinyxml2::XML_WRONG_ATTRIBUTE_TYPE:
                throw std::runtime_error("f_number attribute is not a float number!");
            default:
                result.m_fNumber = f;
        }
    }

    const tinyxml2::XMLElement *sensorNode = imageNode->FirstChildElement("sensor");
    if (sensorNode != nullptr) {
        float f;
        switch (sensorNode->QueryFloatAttribute("width_mm", &f)) {
            case tinyxml2::XML_NO_ATTRIBUTE:
                break;
            case tinyxml2::XML_WRONG_ATTRIBUTE_TYPE:
                throw std::runtime_error("width_mm attribute is not a float number!");
            default:
                result.m_sensorWidth = f;
        }
    }
    return result;
}




Image::Image(const Camera &camera, const boost::filesystem::path &hdrFilename, const boost::filesystem::path &attribFilename) :
        m_camera(camera), m_hdrFilename(hdrFilename), m_attribFilename(attribFilename)
{

}


Image::FloatArray Image::readDistanceMap() const
{
    // open input file
    Imf::InputFile inputFile(m_attribFilename.string().c_str());

    // read size
    Imath::Box2i dw = inputFile.header().dataWindow();
    unsigned width = dw.max.x - dw.min.x + 1;
    unsigned height = dw.max.y - dw.min.y + 1;
    const int dx = dw.min.x;
    const int dy = dw.min.y;

    if ((width != m_camera.getWidth()) ||
        (height != m_camera.getHeight()))
        throw std::runtime_error("The image size stored in the distance exr file does not match the size specified for the camera!");

    // resize local data structure
    FloatArray result;
    result.resize(height, width);
    Imf::FrameBuffer frameBuffer;
    frameBuffer.insert("distance.V", Imf::Slice(Imf::FLOAT, (char *) (result.data() - (dx + dy*width)), sizeof(float), sizeof(float) * width));

    // load data
    inputFile.setFrameBuffer (frameBuffer);
    inputFile.readPixels (dw.min.y, dw.max.y);

    return result;
}

Image::FloatArray Image::readZBuffer(unsigned subsamplingStride) const
{
    return computeZBuffer(readDistanceMap(), subsamplingStride);
}

Image::FloatArray Image::computeZBuffer(const FloatArray &distanceMap, unsigned subsamplingStride) const
{
    FloatArray zBuffer;
    zBuffer.resize(distanceMap.rows()/subsamplingStride, distanceMap.cols()/subsamplingStride);

    const Eigen::Matrix4f invInternal = math::augmentProjectionMatrix<float>(m_camera.getInternalCalibration() * Eigen::Matrix<float, 3, 4>::Identity()).inverse();
    for (unsigned y = 0; y < zBuffer.rows(); y++)
        for (unsigned x = 0; x < zBuffer.cols(); x++) {
            Eigen::Vector4f imgSpace(x * subsamplingStride + 0.5f, y * subsamplingStride + 0.5f, 1.0f, 1.0f);
            Eigen::Vector4f viewSpace = invInternal * imgSpace;
            float sqrDistance = viewSpace.block<3,1>(0,0).squaredNorm() / (viewSpace[3]*viewSpace[3]);
            float z = distanceMap(y * subsamplingStride, x * subsamplingStride) / std::sqrt(sqrDistance);

            zBuffer(y, x) = z;
        }

    return zBuffer;
}



Image::UIntArray Image::readObjectID() const
{
    // open input file
    Imf::InputFile inputFile(m_attribFilename.string().c_str());

    // read size
    Imath::Box2i dw = inputFile.header().dataWindow();
    unsigned width = dw.max.x - dw.min.x + 1;
    unsigned height = dw.max.y - dw.min.y + 1;
    const int dx = dw.min.x;
    const int dy = dw.min.y;

    if ((width != m_camera.getWidth()) ||
        (height != m_camera.getHeight()))
        throw std::runtime_error("The image size stored in the distance exr file does not match the size specified for the camera!");


    UIntArray objIdBuffer;
    objIdBuffer.resize(height, width);
    FloatArray distance;
    distance.resize(height, width);


    Imf::FrameBuffer frameBuffer;
    frameBuffer.insert("objectIdx.V", Imf::Slice(Imf::UINT, (char *) (objIdBuffer.data() - (dx + dy*width)), sizeof(std::uint32_t), sizeof(std::uint32_t) * width));
    frameBuffer.insert("distance.V", Imf::Slice(Imf::FLOAT, (char *) (distance.data() - (dx + dy*width)), sizeof(float), sizeof(float) * width));

    // load data
    inputFile.setFrameBuffer (frameBuffer);
    inputFile.readPixels (dw.min.y, dw.max.y);

    for (unsigned y = 0; y < objIdBuffer.rows(); y++) {
        for (unsigned x = 0; x < objIdBuffer.cols(); x++) {
            if (distance(y,x) > 1e8f)
                objIdBuffer(y, x) = ~0u;
        }
    }


    return objIdBuffer;
}


void Image::extractPointcloud(std::vector<Eigen::Vector3f> &position, std::vector<unsigned> *mask, unsigned subsamplingStride) const
{
    FloatArray distance = readDistanceMap();
    FloatArray zBuffer = computeZBuffer(distance, subsamplingStride);
    UIntArray objIdBuffer;
    if (mask != nullptr)
        objIdBuffer = readObjectID();

    Eigen::Matrix4f P = math::composeAugmentedProjectionMatrix<float>(
                                   m_camera.getInternalCalibration(),
                                   m_camera.getExternalCalibration());

    const Eigen::Matrix4f invP = P.inverse();

    for (unsigned y = 0; y < zBuffer.rows(); y++) {
        for (unsigned x = 0; x < zBuffer.cols(); x++) {

            float z = zBuffer(y, x);
            if (z>1e8f) continue;

            unsigned x_ = x * subsamplingStride;
            unsigned y_ = y * subsamplingStride;

            Eigen::Vector4f imgSpace(x_ + 0.5f, y_ + 0.5f, 1.0f/z, 1.0f);
            Eigen::Vector4f worldSpace = invP * imgSpace;

            position.push_back(worldSpace.block<3, 1>(0,0) / worldSpace[3]);
            if (mask != nullptr) {
                mask->push_back(objIdBuffer(y_, x_));
            }
        }
    }
}


}
}
