#include "PlyPointcloudFile.h"

#include "Pointcloud.h"
#include <boost/lexical_cast.hpp>

#include <iostream>

namespace syb3r {
namespace data {

void importPLYPointcloudFile(Pointcloud &dst, const std::string &filename)
{
    std::fstream inputStream;
    inputStream.exceptions(std::ifstream::failbit | std::ifstream::badbit);
    inputStream.open(filename, std::fstream::in);

    importPLYPointcloudFile(dst, inputStream);
}

void importPLYPointcloudFile(Pointcloud &dst, std::fstream &inputStream)
{
    inputStream.exceptions(std::ifstream::failbit | std::ifstream::badbit);

    std::string line;

    std::getline(inputStream, line, (char)0x0A);
    if (line != "ply")
        throw std::runtime_error("Missing header!");

    std::getline(inputStream, line, (char)0x0A);
    if (line != "format ascii 1.0")
        throw std::runtime_error("Invalid or unsupported format");

    const unsigned MODE_DEFAULT = 0;
    const unsigned MODE_VERTEX = 1;
    const unsigned MODE_FACE = 2;
    unsigned mode = MODE_DEFAULT;

    unsigned numVertices = ~0u;
    unsigned vertexStride = 0;
    unsigned posOffset[3] = {~0u, ~0u, ~0u};
    unsigned colorOffset[3] = {~0u, ~0u, ~0u};
    unsigned normalOffset[3] = {~0u, ~0u, ~0u};


    for (std::getline(inputStream, line, (char)0x0A); line != "end_header"; std::getline(inputStream, line, (char)0x0A)) {
        if (line.substr(0, 7) == "comment") continue;

        if (line.substr(0, 14) == "element vertex") {
            numVertices = boost::lexical_cast<unsigned>(line.substr(15));

            mode = MODE_VERTEX;
        } else if (line.substr(0, 12) == "element face") {
            mode = MODE_FACE;
        } else {
            switch (mode) {
                case MODE_VERTEX: {
                    if (line.substr(0, 14) == "property float") {

                        if (line.substr(15) == "x") {
                            posOffset[0] = vertexStride;
                        } else
                        if (line.substr(15) == "y") {
                            posOffset[1] = vertexStride;
                        } else
                        if (line.substr(15) == "z") {
                            posOffset[2] = vertexStride;
                        } else
                        if (line.substr(15) == "nx") {
                            normalOffset[0] = vertexStride;
                        } else
                        if (line.substr(15) == "ny") {
                            normalOffset[1] = vertexStride;
                        } else
                        if (line.substr(15) == "nz") {
                            normalOffset[2] = vertexStride;
                        } else
                            std::cout << "ignoring property " << line << std::endl;

                        vertexStride += 1;
                    } else
                    if (line.substr(0, 14) == "property uchar") {

                        if ((line.substr(15) == "diffuse_red") || (line.substr(15) == "red")) {
                            colorOffset[0] = vertexStride;
                        } else
                        if ((line.substr(15) == "diffuse_green") || (line.substr(15) == "green")) {
                            colorOffset[1] = vertexStride;
                        } else
                        if ((line.substr(15) == "diffuse_blue") || (line.substr(15) == "blue")) {
                            colorOffset[2] = vertexStride;
                        } else
                            std::cout << "ignoring property " << line << std::endl;
                        vertexStride += 1;
                    } else 
                        throw std::runtime_error("Invalid or unsupported vertex property type (can only do float and uchar)");
                } break;
                case MODE_FACE: {
                } break;
            }
        }
    }
    
    if (numVertices == ~0u)
        throw std::runtime_error("Missing vertex declaration!");

    bool hasVertexPositions = (posOffset[0] != 0xFFFFFFFF) &&
                              (posOffset[1] != 0xFFFFFFFF) &&
                              (posOffset[2] != 0xFFFFFFFF);
    bool hasVertexColors = (colorOffset[0] != 0xFFFFFFFF) &&
                              (colorOffset[1] != 0xFFFFFFFF) &&
                              (colorOffset[2] != 0xFFFFFFFF);
    bool hasVertexNormals = (normalOffset[0] != 0xFFFFFFFF) &&
                              (normalOffset[1] != 0xFFFFFFFF) &&
                              (normalOffset[2] != 0xFFFFFFFF);
                              
    dst.reset(numVertices);

    if (!hasVertexPositions)
        throw std::runtime_error("Missing vertex positions!");
/*
    if (!hasVertexNormals)
        throw std::runtime_error("Missing vertex normals!");
*/
    std::vector<float> vertexBlob;
    vertexBlob.resize(vertexStride);

    for (unsigned i = 0; i < numVertices; i++) {
        for (unsigned j = 0; j < vertexStride; j++)
            inputStream >> vertexBlob[j];

        Eigen::Vector3f position, color, normal;
        position.setZero();
        color.setZero();
        normal.setZero();

        if (hasVertexPositions) {
            position[0] = vertexBlob[posOffset[0]];
            position[1] = vertexBlob[posOffset[1]];
            position[2] = vertexBlob[posOffset[2]];
        }
        if (hasVertexColors) {
            color[0] = vertexBlob[colorOffset[0]] / 255.0f;
            color[1] = vertexBlob[colorOffset[1]] / 255.0f;
            color[2] = vertexBlob[colorOffset[2]] / 255.0f;
        }
        if (hasVertexNormals) {
            normal[0] = vertexBlob[normalOffset[0]];
            normal[1] = vertexBlob[normalOffset[1]];
            normal[2] = vertexBlob[normalOffset[2]];
        }

        dst.setPoint(i, position, normal, color);
    }
}


void importPLYPointcloudFile(Pointcloud &dst, std::vector<float> &vertexQuality, const std::string &filename)
{
    std::fstream inputStream;
    inputStream.exceptions(std::ifstream::failbit | std::ifstream::badbit);
    inputStream.open(filename, std::fstream::in);

    importPLYPointcloudFile(dst, vertexQuality, inputStream);
}

void importPLYPointcloudFile(Pointcloud &dst, std::vector<float> &vertexQuality, std::fstream &inputStream)
{
    inputStream.exceptions(std::ifstream::failbit | std::ifstream::badbit);

    std::string line;

    std::getline(inputStream, line, (char)0x0A);
    if (line != "ply")
        throw std::runtime_error("Missing header!");

    std::getline(inputStream, line, (char)0x0A);
    if (line != "format ascii 1.0")
        throw std::runtime_error("Invalid or unsupported format");

    const unsigned MODE_DEFAULT = 0;
    const unsigned MODE_VERTEX = 1;
    const unsigned MODE_FACE = 2;
    unsigned mode = MODE_DEFAULT;

    unsigned numVertices = ~0u;
    unsigned vertexStride = 0;
    unsigned posOffset[3] = {~0u, ~0u, ~0u};
    unsigned colorOffset[3] = {~0u, ~0u, ~0u};
    unsigned normalOffset[3] = {~0u, ~0u, ~0u};
    unsigned vertexQualityOffset = ~0u;


    for (std::getline(inputStream, line, (char)0x0A); line != "end_header"; std::getline(inputStream, line, (char)0x0A)) {
        if (line.substr(0, 7) == "comment") continue;

        if (line.substr(0, 14) == "element vertex") {
            numVertices = boost::lexical_cast<unsigned>(line.substr(15));

            mode = MODE_VERTEX;
        } else if (line.substr(0, 12) == "element face") {
            mode = MODE_FACE;
        } else {
            switch (mode) {
                case MODE_VERTEX: {
                    if (line.substr(0, 14) == "property float") {

                        if (line.substr(15) == "x") {
                            posOffset[0] = vertexStride;
                        } else
                        if (line.substr(15) == "y") {
                            posOffset[1] = vertexStride;
                        } else
                        if (line.substr(15) == "z") {
                            posOffset[2] = vertexStride;
                        } else
                        if (line.substr(15) == "nx") {
                            normalOffset[0] = vertexStride;
                        } else
                        if (line.substr(15) == "ny") {
                            normalOffset[1] = vertexStride;
                        } else
                        if (line.substr(15) == "nz") {
                            normalOffset[2] = vertexStride;
                        } else
                        if (line.substr(15) == "quality") {
                            vertexQualityOffset = vertexStride;
                        } else
                            std::cout << "ignoring property " << line << std::endl;

                        vertexStride += 1;
                    } else
                    if (line.substr(0, 14) == "property uchar") {

                        if ((line.substr(15) == "diffuse_red") || (line.substr(15) == "red")) {
                            colorOffset[0] = vertexStride;
                        } else
                        if ((line.substr(15) == "diffuse_green") || (line.substr(15) == "green")) {
                            colorOffset[1] = vertexStride;
                        } else
                        if ((line.substr(15) == "diffuse_blue") || (line.substr(15) == "blue")) {
                            colorOffset[2] = vertexStride;
                        } else
                            std::cout << "ignoring property " << line << std::endl;
                        vertexStride += 1;
                    } else
                        throw std::runtime_error("Invalid or unsupported vertex property type (can only do float and uchar)");
                } break;
                case MODE_FACE: {
                } break;
            }
        }
    }

    if (numVertices == ~0u)
        throw std::runtime_error("Missing vertex declaration!");

    bool hasVertexPositions = (posOffset[0] != 0xFFFFFFFF) &&
                              (posOffset[1] != 0xFFFFFFFF) &&
                              (posOffset[2] != 0xFFFFFFFF);
    bool hasVertexColors = (colorOffset[0] != 0xFFFFFFFF) &&
                              (colorOffset[1] != 0xFFFFFFFF) &&
                              (colorOffset[2] != 0xFFFFFFFF);
    bool hasVertexNormals = (normalOffset[0] != 0xFFFFFFFF) &&
                              (normalOffset[1] != 0xFFFFFFFF) &&
                              (normalOffset[2] != 0xFFFFFFFF);

    dst.reset(numVertices);

    if (!hasVertexPositions)
        throw std::runtime_error("Missing vertex positions!");
/*
    if (!hasVertexNormals)
        throw std::runtime_error("Missing vertex normals!");
*/

    if (vertexQualityOffset == ~0u)
        vertexQuality.resize(0);
    else
        vertexQuality.resize(numVertices);

    std::vector<float> vertexBlob;
    vertexBlob.resize(vertexStride);

    for (unsigned i = 0; i < numVertices; i++) {
        for (unsigned j = 0; j < vertexStride; j++)
            inputStream >> vertexBlob[j];

        Eigen::Vector3f position, color, normal;
        position.setZero();
        color.setZero();
        normal.setZero();

        if (hasVertexPositions) {
            position[0] = vertexBlob[posOffset[0]];
            position[1] = vertexBlob[posOffset[1]];
            position[2] = vertexBlob[posOffset[2]];
        }
        if (hasVertexColors) {
            color[0] = vertexBlob[colorOffset[0]] / 255.0f;
            color[1] = vertexBlob[colorOffset[1]] / 255.0f;
            color[2] = vertexBlob[colorOffset[2]] / 255.0f;
        }
        if (hasVertexNormals) {
            normal[0] = vertexBlob[normalOffset[0]];
            normal[1] = vertexBlob[normalOffset[1]];
            normal[2] = vertexBlob[normalOffset[2]];
        }

        if (vertexQualityOffset != ~0u)
            vertexQuality[i] = vertexBlob[vertexQualityOffset];

        dst.setPoint(i, position, normal, color);
    }
}



void writePLYPointcloudFile(const std::string &filename, unsigned numPoints, const std::function<void(unsigned, Eigen::Vector3f&)> &getPosition)
{
    std::fstream file;
    file.exceptions(std::ifstream::failbit | std::ifstream::badbit);
    file.open(filename, std::fstream::out);

    writePLYPointcloudFile(file, numPoints, getPosition);
}

void writePLYPointcloudFile(std::fstream &dst, unsigned numPoints, const std::function<void(unsigned, Eigen::Vector3f&)> &getPosition)
{
    dst
        << "ply" << std::endl
        << "format ascii 1.0" << std::endl
        << "element vertex " << numPoints << std::endl
        << "property float x" << std::endl
        << "property float y" << std::endl
        << "property float z" << std::endl
        << "end_header" << std::endl;

    for (unsigned i = 0; i < numPoints; i++) {
        Eigen::Vector3f position;
        getPosition(i, position);
        dst
            << position[0] << " "
            << position[1] << " "
            << position[2] << " "
            << std::endl;
    }
}


void writePLYPointcloudFile(const std::string &filename, unsigned numPoints, const std::function<void(unsigned, Eigen::Vector3f&, Eigen::Vector3f&)> &getPositionNormal)
{
    std::fstream file;
    file.exceptions(std::ifstream::failbit | std::ifstream::badbit);
    file.open(filename, std::fstream::out);

    writePLYPointcloudFile(file, numPoints, getPositionNormal);
}

void writePLYPointcloudFile(std::fstream &dst, unsigned numPoints, const std::function<void(unsigned, Eigen::Vector3f&, Eigen::Vector3f&)> &getPositionNormal)
{
    dst
        << "ply" << std::endl
        << "format ascii 1.0" << std::endl
        << "element vertex " << numPoints << std::endl
        << "property float x" << std::endl
        << "property float y" << std::endl
        << "property float z" << std::endl
        << "property float nx" << std::endl
        << "property float ny" << std::endl
        << "property float nz" << std::endl
        << "end_header" << std::endl;

    for (unsigned i = 0; i < numPoints; i++) {
        Eigen::Vector3f position, normal;
        getPositionNormal(i, position, normal);
        dst
            << position[0] << " "
            << position[1] << " "
            << position[2] << " "
            << normal[0] << " "
            << normal[1] << " "
            << normal[2] << " "
            << std::endl;
    }
}

void writePLYPointcloudFile(const std::string &filename, unsigned numPoints, const std::function<void(unsigned, Eigen::Vector3f&, Eigen::Vector3f&, Eigen::Vector3f&)> &getPositionNormalColor)
{
    std::fstream file;
    file.exceptions(std::ifstream::failbit | std::ifstream::badbit);
    file.open(filename, std::fstream::out);

    writePLYPointcloudFile(file, numPoints, getPositionNormalColor);
}

void writePLYPointcloudFile(std::fstream &dst, unsigned numPoints, const std::function<void(unsigned, Eigen::Vector3f&, Eigen::Vector3f&, Eigen::Vector3f&)> &getPositionNormalColor)
{
    dst
        << "ply" << std::endl
        << "format ascii 1.0" << std::endl
        << "element vertex " << numPoints << std::endl
        << "property float x" << std::endl
        << "property float y" << std::endl
        << "property float z" << std::endl
        << "property float nx" << std::endl
        << "property float ny" << std::endl
        << "property float nz" << std::endl
        << "property uchar red" << std::endl
        << "property uchar green" << std::endl
        << "property uchar blue" << std::endl
        << "end_header" << std::endl;

    for (unsigned i = 0; i < numPoints; i++) {
        Eigen::Vector3f position, normal, color;
        getPositionNormalColor(i, position, normal, color);
        dst
            << position[0] << " "
            << position[1] << " "
            << position[2] << " "
            << normal[0] << " "
            << normal[1] << " "
            << normal[2] << " "
            << std::min<int>(255, color[0]*255 + 0.5f) << " "
            << std::min<int>(255, color[1]*255 + 0.5f) << " "
            << std::min<int>(255, color[2]*255 + 0.5f) << " "
            << std::endl;
    }
}


void writePLYPointcloudFile(const std::string &filename, unsigned numPoints, const std::function<void(unsigned, Eigen::Vector3f&, Eigen::Vector3f&, std::uint32_t&)> &getPositionNormalColor)
{
    std::fstream file;
    file.exceptions(std::ifstream::failbit | std::ifstream::badbit);
    file.open(filename, std::fstream::out);

    writePLYPointcloudFile(file, numPoints, getPositionNormalColor);
}

void writePLYPointcloudFile(std::fstream &dst, unsigned numPoints, const std::function<void(unsigned, Eigen::Vector3f&, Eigen::Vector3f&, std::uint32_t&)> &getPositionNormalColor)
{
    dst
        << "ply" << std::endl
        << "format ascii 1.0" << std::endl
        << "element vertex " << numPoints << std::endl
        << "property float x" << std::endl
        << "property float y" << std::endl
        << "property float z" << std::endl
        << "property float nx" << std::endl
        << "property float ny" << std::endl
        << "property float nz" << std::endl
        << "property uchar red" << std::endl
        << "property uchar green" << std::endl
        << "property uchar blue" << std::endl
        << "end_header" << std::endl;

    for (unsigned i = 0; i < numPoints; i++) {
        Eigen::Vector3f position, normal;
        std::uint32_t color;
        getPositionNormalColor(i, position, normal, color);
        dst
            << position[0] << " "
            << position[1] << " "
            << position[2] << " "
            << normal[0] << " "
            << normal[1] << " "
            << normal[2] << " "
            << ((color << 0) & 0xFF) << " "
            << ((color << 8) & 0xFF) << " "
            << ((color << 16) & 0xFF) << " "
            << std::endl;
    }
}

}
}
