#ifndef CAMERAPROJECTIONMATH_H
#define CAMERAPROJECTIONMATH_H

#include <Eigen/Dense>
//#include <iostream>

namespace syb3r {
namespace math {

template<typename Type>
Eigen::Matrix<Type, 3, 3> dropColumn(const Eigen::Matrix<Type, 3, 4> &projectionMatrix, unsigned col)
{
    Eigen::Matrix<Type, 3, 3> P;
    for (unsigned i = 0; i < 3; i++) {
        for (unsigned j = 0; j < col; j++)
            P(i, j) = projectionMatrix(i, j);
        for (unsigned j = col+1; j < 4; j++)
            P(i, j-1) = projectionMatrix(i, j);
    }
    return P;
}

/**
 * @brief computes the focal point of a camera from a given projection-view matrix
 */
template<typename Type>
Eigen::Matrix<Type, 3, 1> computeFocalPoint(const Eigen::Matrix<Type, 3, 4> &projectionMatrix)
{

    Eigen::Matrix<Type, 3, 3> M = projectionMatrix.block(0, 0, 3, 3);

    Type lambda = M.row(2).norm();
    if (M.determinant() < 0.0f)
        lambda = -lambda;

    Eigen::Matrix<Type, 3, 4> scaledProjectionMatrix = projectionMatrix;
    scaledProjectionMatrix *= (1.0f /  lambda);
   // M *= (1.0f /  lambda);


    Type Cx = (dropColumn(scaledProjectionMatrix, 0).determinant());
    Type Cy = -(dropColumn(scaledProjectionMatrix, 1).determinant());
    Type Cz = (dropColumn(scaledProjectionMatrix, 2).determinant());
    Type Cw = -(dropColumn(scaledProjectionMatrix, 3).determinant());


    Eigen::Matrix<Type, 3, 1> cameraPosition(
        Cx / Cw,
        Cy / Cw,
        Cz / Cw
    );

    return cameraPosition;
}


template<typename Type, unsigned Rows, unsigned Cols>
Eigen::Matrix<Type, Rows, Cols> flipud(const Eigen::Matrix<Type, Rows, Cols> &A)
{
    Eigen::Matrix<Type, Rows, Cols> result;
    for (unsigned i = 0; i < Rows; i++)
        result.row(Rows-1-i) = A.row(i);
    return result;
}

extern void RQDecompose(const Eigen::Matrix<float, 3, 3> &A, Eigen::Matrix<float, 3, 3> &Q, Eigen::Matrix<float, 3, 3> &R);
/*{
    Eigen::Matrix<float, 3, 3> A_ = flipud<float, 3, 3>(A).transpose();
    Eigen::HouseholderQR<Eigen::Matrix<float, 3, 3>> qr(A_);
    Eigen::Matrix<float, 3, 3> R_ = qr.matrixQR().triangularView<Eigen::Upper>();

    R = flipud<float, 3, 3>(flipud<float, 3, 3>(R_).transpose());
    Q = flipud<float, 3, 3>(qr.householderQ().transpose());
}*/

/*

///@todo Figure out what the hell is wrong with this
template<typename Type, unsigned Dim, class QRDecompose>// = Eigen::FullPivHouseholderQR<Eigen::Matrix<Type, Dim, Dim>>>
void RQDecompose(const Eigen::Matrix<Type, Dim, Dim> &A, Eigen::Matrix<Type, Dim, Dim> &Q, Eigen::Matrix<Type, Dim, Dim> &R)
{
    Eigen::Matrix<Type, Dim, Dim> A_ = flipud<Type, Dim, Dim>(A.transpose());
    //QRDecompose qr(A_);
    Eigen::FullPivHouseholderQR<Eigen::Matrix<Type, Dim, Dim>> qr(A_);
    Eigen::Matrix<Type, Dim, Dim> R_ = qr.matrixQR().triangularView<Eigen::Upper>();

    R = flipud<Type, Dim, Dim>(flipud<Type, Dim, Dim>(R_).transpose());
    Q = qr.matrixQ().transpose();
}
*/

template<typename Type>
void decomposeProjectionMatrix(const Eigen::Matrix<Type, 3, 4> &projectionMatrix,
                                Eigen::Matrix<Type, 3, 3> &internalMatrix,
                                Eigen::Matrix<Type, 3, 3> &cameraRotation,
                                Eigen::Matrix<Type, 3, 1> &cameraPosition)
{
    Eigen::Matrix<Type, 3, 3> M = projectionMatrix.block(0, 0, 3, 3);

    Type lambda = M.row(2).norm();
    if (M.determinant() < 0.0f)
        lambda = -lambda;

    Eigen::Matrix<Type, 3, 4> scaledProjectionMatrix = projectionMatrix;
    scaledProjectionMatrix *= (1.0f /  lambda);
    M *= (1.0f /  lambda);


    Type Cx = (dropColumn(scaledProjectionMatrix, 0).determinant());
    Type Cy = -(dropColumn(scaledProjectionMatrix, 1).determinant());
    Type Cz = (dropColumn(scaledProjectionMatrix, 2).determinant());
    Type Cw = -(dropColumn(scaledProjectionMatrix, 3).determinant());

    cameraPosition = Eigen::Matrix<Type, 3, 1>(
        Cx / Cw,
        Cy / Cw,
        Cz / Cw
    );

    Eigen::Matrix<Type, 3, 3> R, Q;
    RQDecompose(M, Q, R);

    for (unsigned i = 0; i < 3; i++) {
        if (R(i,i) < 0.0f) {
            for (unsigned j = 0; j < 3; j++) {
                R(j, i) = -R(j, i);
                Q(i,j) = -Q(i,j);
            }
        }
    }

    internalMatrix = R;
    cameraRotation = Q;
}


template<typename Type>
Eigen::Matrix<Type, 4, 4> composeAugmentedProjectionMatrix(const Eigen::Matrix<Type, 3, 3> &internalMatrix, const Eigen::Matrix<Type, 4, 4> &externalMatrix)
{
    Eigen::Matrix<Type, 4, 4> augmentedInternal;
    augmentedInternal(0, 0) = internalMatrix(0, 0) * (1.0f / internalMatrix(2, 2));
    augmentedInternal(0, 1) = internalMatrix(0, 1) * (1.0f / internalMatrix(2, 2));
    augmentedInternal(0, 2) = internalMatrix(0, 2) * (1.0f / internalMatrix(2, 2));
    augmentedInternal(0, 3) = 0.0f;

    augmentedInternal(1, 0) = internalMatrix(1, 0) * (1.0f / internalMatrix(2, 2));
    augmentedInternal(1, 1) = internalMatrix(1, 1) * (1.0f / internalMatrix(2, 2));
    augmentedInternal(1, 2) = internalMatrix(1, 2) * (1.0f / internalMatrix(2, 2));
    augmentedInternal(1, 3) = 0.0f;

    augmentedInternal(2, 0) = 0.0f;
    augmentedInternal(2, 1) = 0.0f;
    augmentedInternal(2, 2) = 0.0f;
    augmentedInternal(2, 3) = 1.0f;

    augmentedInternal(3, 0) = 0.0f;
    augmentedInternal(3, 1) = 0.0f;
    augmentedInternal(3, 2) = 1.0f;
    augmentedInternal(3, 3) = 0.0f;

    return augmentedInternal * externalMatrix;
}


/**
 * @brief Augments a 3x4 projection matrix to a 4x4 projection matrix which also computes the reciprocal depth to the focal plane
 */
template<typename Type>
Eigen::Matrix<Type, 4, 4> augmentProjectionMatrix(const Eigen::Matrix<Type, 3, 4> &projectionMatrix)
{
    Eigen::Matrix<Type, 3, 3> internalMatrix;
    Eigen::Matrix<Type, 3, 3> cameraRotation;
    Eigen::Matrix<Type, 3, 1> cameraPosition;

    decomposeProjectionMatrix(projectionMatrix, internalMatrix, cameraRotation, cameraPosition);

    Eigen::Matrix<Type, 3, 1> revCameraPosition = cameraRotation * cameraPosition;
    Eigen::Matrix<Type, 4, 4> external;
    external(0, 0) = cameraRotation(0, 0);
    external(0, 1) = cameraRotation(0, 1);
    external(0, 2) = cameraRotation(0, 2);
    external(0, 3) = -revCameraPosition(0);

    external(1, 0) = cameraRotation(1, 0);
    external(1, 1) = cameraRotation(1, 1);
    external(1, 2) = cameraRotation(1, 2);
    external(1, 3) = -revCameraPosition(1);

    external(2, 0) = cameraRotation(2, 0);
    external(2, 1) = cameraRotation(2, 1);
    external(2, 2) = cameraRotation(2, 2);
    external(2, 3) = -revCameraPosition(2);

    external(3, 0) = 0.0f;
    external(3, 1) = 0.0f;
    external(3, 2) = 0.0f;
    external(3, 3) = 1.0f;

    return composeAugmentedProjectionMatrix<Type>(internalMatrix, external);
}

}
}

#endif // CAMERAPROJECTIONMATH_H
