#include "FractionalICP.h"

#include "HelmertTransformation.h"
#include "KNearestNeighbors.hpp"

#include "../tools/TaskScheduler.h"

#include <algorithm>

#include <iostream>

namespace syb3r {
namespace math {


void ICPTransformHelmert::transform(const Eigen::Vector3f *src, Eigen::Vector3f *dst, unsigned count) const
{
    for (unsigned i = 0; i < count; i++)
        dst[i] = m_rotation * src[i] * m_scale + m_offset;
}

void ICPTransformHelmert::computeLeastSquaresTransform(const Eigen::Vector3f *src, const Eigen::Vector3f *dst, unsigned count)
{
    Eigen::Matrix4f trafo = math::computeHelmertTransformation(src, dst, count);

    assert(trafo(3, 0) == 0.0f);
    assert(trafo(3, 1) == 0.0f);
    assert(trafo(3, 2) == 0.0f);
    assert(trafo(3, 3) == 1.0f);

    m_offset[0] = trafo(0, 3);
    m_offset[1] = trafo(1, 3);
    m_offset[2] = trafo(2, 3);

    m_rotation = trafo.block<3,3>(0,0);

    m_scale = std::sqrt(m_rotation.col(0).dot(m_rotation.col(0)));
    m_rotation *= 1.0f / m_scale;
}



namespace {

    class Point;

    struct FeatureSpace {
        typedef Eigen::Vector3f Features;
        enum SplitType {
            SPLIT_X = 0,
            SPLIT_Y = 1,
            SPLIT_Z = 2,
            NO_SPLIT = 3
        };

        struct Metric {
            typedef float Type;
            static inline Type compute(const Features &features, const Point &fact);
            static inline bool stillInRange(const Point &fact, const Features &features, Type metric, SplitType splittinFeature);
        };

        static inline SplitType chooseSplittingFeature(const Point &parent, const Point &child);
        static inline unsigned compare(const Point &nodePoint, const Features &queryPoint, SplitType splittinFeature);
        static inline unsigned compare(const Point &nodePoint, const Point &queryPoint, SplitType splittinFeature);
    };

    struct Point {
        typedef unsigned Handle;

        FeatureSpace::Features position;
        Handle handle;

        inline Handle getHandle() const { return handle; }
    };


    FeatureSpace::Metric::Type FeatureSpace::Metric::compute(const Features &features, const Point &fact) {
        Eigen::Vector3f D = features - fact.position;
        return D.dot(D);
    }

    bool FeatureSpace::Metric::stillInRange(const Point &fact, const Features &features, Type metric, SplitType splittinFeature) {
        Type distance = features[splittinFeature] - fact.position[splittinFeature];
        return distance*distance < metric;
    }

    FeatureSpace::SplitType FeatureSpace::chooseSplittingFeature(const Point &parent, const Point &child) {
        Eigen::Vector3f D = parent.position - child.position;
        float x = std::abs(D[0]);
        float y = std::abs(D[1]);
        float z = std::abs(D[2]);

        if ((x > y) && (x > z))
            return FeatureSpace::SplitType(0);

        if (y > z)
            return FeatureSpace::SplitType(1);

        return FeatureSpace::SplitType(2);
    }

    unsigned FeatureSpace::compare(const Point &nodePoint, const Features &queryPoint, SplitType splittinFeature) {
        return (queryPoint[splittinFeature] > nodePoint.position[splittinFeature])?0:1;
    }
    unsigned FeatureSpace::compare(const Point &nodePoint, const Point &queryPoint, SplitType splittinFeature) {
        return compare(nodePoint, queryPoint.position, splittinFeature);
    }
}


void fractionalICP(const std::vector<Eigen::Vector3f> &src, const std::vector<Eigen::Vector3f> &unshuffledDst, ICPTransform &transform)
{
    std::vector<Eigen::Vector3f> srcSubset = src;
    std::random_shuffle(srcSubset.begin(), srcSubset.end());
    srcSubset.resize(std::min<unsigned>(40000u, srcSubset.size()));

    std::vector<std::pair<float, unsigned>> sortedResiduals;
    sortedResiduals.resize(srcSubset.size());
    std::vector<Eigen::Vector3f> src2dstMapping;
    src2dstMapping.resize(srcSubset.size());
    std::vector<Eigen::Vector3f> transformed;
    transformed.resize(srcSubset.size());

    std::vector<Eigen::Vector3f> subsetSrc;
    subsetSrc.resize(srcSubset.size());
    std::vector<Eigen::Vector3f> subsetDst;
    subsetDst.resize(srcSubset.size());

    std::cout << "Building kd tree!" << std::endl;
    kNearestNeighbors::Dataset<FeatureSpace, Point> kNearestDst;

    std::vector<Eigen::Vector3f> dst = unshuffledDst;
    std::random_shuffle(dst.begin(), dst.end());

    for (unsigned i = 0; i < dst.size(); i++)
        kNearestDst.addFact(Point{dst[i], i});


    std::cout << "Running ICP" << std::endl;

    for (unsigned iter = 0; iter < 10; iter++) {
        std::cout << "Iter " << iter << " of 10" << std::endl;
        transform.transform(srcSubset.data(), transformed.data(), srcSubset.size());
        tools::TaskGroup group;
        for (unsigned block = 0; block < srcSubset.size(); block+=1000) {
            group.add([block, &srcSubset, &kNearestDst, &transformed, &src2dstMapping, &dst]{
                for (unsigned i = block; i < std::min<unsigned>(srcSubset.size(), block+1000); i++) {
                    kNearestNeighbors::QueueEntry<FeatureSpace, Point> kNearest;
                    kNearestDst.getNearestNeighbor(transformed[i], kNearest);
                    src2dstMapping[i] = dst[kNearest.handle];
                }
            }, tools::TaskScheduler::get());
        }
        tools::TaskScheduler::get().waitFor(&group);

        for (unsigned i = 0; i < srcSubset.size(); i++)
            sortedResiduals[i] = {(transformed[i] - src2dstMapping[i]).squaredNorm(), i};

        std::sort(sortedResiduals.begin(), sortedResiduals.end());

        unsigned bestFractionCount = 0;
        {
            float bestFractionScore = 1e30f;

            float runningSum = 0.0f;
            for (unsigned fractionCount = 0; fractionCount < srcSubset.size(); fractionCount++) {
                runningSum += sortedResiduals[fractionCount].first;

                if (fractionCount > 1000) {
                    float fraction = fractionCount / (float)srcSubset.size();
                    float fractionFactor = 1.0f / (fraction*fraction);
                    fractionFactor *= fractionFactor;
                    fractionFactor *= fractionFactor;// Lambda = 4

                    float score = runningSum / fractionCount * fractionFactor;
                    if (score < bestFractionScore) {
                        bestFractionScore = score;
                        bestFractionCount = fractionCount;
                    }
                }
            }
        }
        std::cout << "Best fraction count: " << bestFractionCount << std::endl;
        subsetSrc.resize(bestFractionCount);
        subsetDst.resize(bestFractionCount);
        for (unsigned i = 0; i < bestFractionCount; i++) {
            subsetSrc[i] = srcSubset[sortedResiduals[i].second];
            subsetDst[i] = src2dstMapping[sortedResiduals[i].second];
        }

        transform.computeLeastSquaresTransform(subsetSrc.data(), subsetDst.data(), subsetSrc.size());
    }
}

void ICP(const std::vector<Eigen::Vector3f> &src, const std::vector<Eigen::Vector3f> &dst, ICPTransform &transform, float residualPower, float *residual)
{
    std::vector<Eigen::Vector3f> src2dstMapping;
    src2dstMapping.resize(src.size());
    std::vector<Eigen::Vector3f> transformed;
    transformed.resize(src.size());

    kNearestNeighbors::Dataset<FeatureSpace, Point> kNearestDst;

    for (unsigned i = 0; i < dst.size(); i++)
        kNearestDst.addFact(Point{dst[i], i});

    kNearestNeighbors::QueueEntry<FeatureSpace, Point> kNearest;

    for (unsigned iter = 0; iter < 100; iter++) {
        transform.transform(src.data(), transformed.data(), src.size());
        for (unsigned i = 0; i < src.size(); i++) {
            kNearestDst.getNearestNeighbor(transformed[i], kNearest);
            src2dstMapping[i] = dst[kNearest.handle];
        }

        transform.computeLeastSquaresTransform(src.data(), src2dstMapping.data(), transformed.size());
    }

    if (residual != nullptr) {
        transform.transform(src.data(), transformed.data(), src.size());
        for (unsigned i = 0; i < src.size(); i++) {
            kNearestDst.getNearestNeighbor(transformed[i], kNearest);
            src2dstMapping[i] = dst[kNearest.handle];
        }

        float sum = 0.0f;
        for (unsigned i = 0; i < src.size(); i++)
            sum += std::pow((transformed[i] - src2dstMapping[i]).squaredNorm(), residualPower/2.0f);

        *residual = sum;
    }

}

}
}