#ifndef HELMERTTRANSFORMATION_H
#define HELMERTTRANSFORMATION_H

#include <Eigen/Dense>
#include <stdexcept>

namespace syb3r {
namespace math {


/**
 * @brief Computes the helmert transformation from a set of point pairs.
 * @details The resulting transformation tries (in a least squares way) to map every point in the src array to it's corresponding point in the dst array
 * while only using rotation, translation, and homogeneous scale.
 */
template<typename Type>
Eigen::Matrix<Type, 4, 4> computeHelmertTransformation(const Eigen::Matrix<Type, 3, 1> *src, const Eigen::Matrix<Type, 3, 1> *dst, unsigned count)
{
    if (count < 3)
        throw std::runtime_error("Insufficient points for helmert transformation!");

    Eigen::Matrix<Type, 3, 1> srcCenter = src[0];
    Eigen::Matrix<Type, 3, 1> dstCenter = dst[0];

    for (unsigned i = 1; i < count; i++) {
        srcCenter += src[i];
        dstCenter += dst[i];
    }

    srcCenter /= count;
    dstCenter /= count;

    Type scale;
    {
        Type srcSize = Type(0);
        Type dstSize = Type(0);
        for (unsigned i = 0; i < count; i++) {
            srcSize += (src[i] - srcCenter).squaredNorm();
            dstSize += (dst[i] - dstCenter).squaredNorm();
        }
        scale = std::sqrt(dstSize / srcSize);
    }

    Eigen::Matrix<Type, 3, 3> M;
    M.setZero();

    for (unsigned i = 0; i < count; i++) {
        Eigen::Matrix<Type, 3, 1> s = src[i] - srcCenter;
        Eigen::Matrix<Type, 3, 1> d = dst[i] - dstCenter;

        for (unsigned k = 0; k < 3; k++)
            for (unsigned l = 0; l < 3; l++)
                M(k, l) += s[k] * d[l];
    }

    Eigen::Matrix<Type, 4, 4> N;
    N(0, 0) = M(0, 0) + M(1, 1) + M(2, 2);
    N(0, 1) = M(1, 2) - M(2, 1);
    N(0, 2) = M(2, 0) - M(0, 2);
    N(0, 3) = M(0, 1) - M(1, 0);

    N(1, 0) = M(1, 2) - M(2, 1);
    N(1, 1) = M(0, 0) - M(1, 1) - M(2, 2);
    N(1, 2) = M(0, 1) + M(1, 0);
    N(1, 3) = M(2, 0) + M(0, 2);

    N(2, 0) = M(2, 0) - M(0, 2);
    N(2, 1) = M(0, 1) + M(1, 0);
    N(2, 2) = -M(0, 0) + M(1, 1) - M(2, 2);
    N(2, 3) = M(1, 2) + M(2, 1);

    N(3, 0) = M(0, 1) - M(1, 0);
    N(3, 1) = M(2, 0) + M(0, 2);
    N(3, 2) = M(1, 2) + M(2, 1);
    N(3, 3) = -M(0, 0) - M(1, 1) + M(2, 2);


    Eigen::SelfAdjointEigenSolver<decltype(N)> eigenSolver(N);

    Eigen::Quaternion<Type> rotation(
                        eigenSolver.eigenvectors()(0, 3),
                        eigenSolver.eigenvectors()(1, 3),
                        eigenSolver.eigenvectors()(2, 3),
                        eigenSolver.eigenvectors()(3, 3)
                    );

    return Eigen::Transform<Type, 3, Eigen::Affine>(Eigen::Translation<Type, 3>(dstCenter) * Eigen::Scaling(scale, scale, scale) * rotation * Eigen::Translation<Type, 3>(-srcCenter)).matrix();
}

extern template Eigen::Matrix<float, 4, 4> computeHelmertTransformation<float>(const Eigen::Matrix<float, 3, 1> *src, const Eigen::Matrix<float, 3, 1> *dst, unsigned count);
extern template Eigen::Matrix<double, 4, 4> computeHelmertTransformation<double>(const Eigen::Matrix<double, 3, 1> *src, const Eigen::Matrix<double, 3, 1> *dst, unsigned count);

}
}

#endif // HELMERTTRANSFORMATION_H
