#ifndef KNEARESTNEIGHBORS_HPP_
#define KNEARESTNEIGHBORS_HPP_

#include <vector>
#include <memory>

namespace syb3r {
namespace math {


namespace kNearestNeighbors {

template<class FeatureSpace, typename Fact>
struct QueueEntry
{
    typename FeatureSpace::Metric::Type metric;
    typename Fact::Handle handle;
    inline bool operator<(const QueueEntry &rhs) const {
        return metric < rhs.metric;
    }
};

template<class FeatureSpace, typename Fact>
class Node
{
    public:
        Node(const Fact &fact);
        void addFact(const Fact &fact);
        void getkNearestNeighbors(const typename FeatureSpace::Features &features, std::vector<QueueEntry<FeatureSpace, Fact>> &facts, unsigned k) const;
        void getNearestNeighbor(const typename FeatureSpace::Features &features, QueueEntry<FeatureSpace, Fact> &fact) const;

        inline const Fact &getFact() const { return m_fact; }
    private:
        typedef Node<FeatureSpace, Fact> NodeType;
        Fact m_fact;
        std::unique_ptr<NodeType> m_subNodes[2];
        typename FeatureSpace::SplitType m_splittinFeature;
};

template<class FeatureSpace, typename Fact>
class Dataset
{
    public:
        typedef Node<FeatureSpace, Fact> NodeType;
        void addFact(const Fact &fact);
        void getkNearestNeighbors(const typename FeatureSpace::Features &features, std::vector<QueueEntry<FeatureSpace, Fact>> &facts, unsigned k) const;
        void getNearestNeighbor(const typename FeatureSpace::Features &features, QueueEntry<FeatureSpace, Fact> &facts) const;
    private:
        std::unique_ptr<NodeType> m_rootNode;
};


template<class FeatureSpace, class Fact>
Node<FeatureSpace, Fact>::Node(const Fact &fact)
{
    m_fact = fact;
    m_splittinFeature = FeatureSpace::NO_SPLIT;
}

template<class FeatureSpace, class Fact>
void Node<FeatureSpace, Fact>::addFact(const Fact &fact)
{
    if (m_splittinFeature == FeatureSpace::NO_SPLIT)
        m_splittinFeature = FeatureSpace::chooseSplittingFeature(m_fact, fact);

    unsigned index = FeatureSpace::compare(m_fact, fact, m_splittinFeature);

    if (m_subNodes[index] == nullptr)
        m_subNodes[index].reset(new Node<FeatureSpace, Fact>(fact));
    else
        m_subNodes[index]->addFact(fact);
}

template<class FeatureSpace, typename Fact>
void insertSorted(std::vector<QueueEntry<FeatureSpace, Fact>> &facts,
                  const QueueEntry<FeatureSpace, Fact> &newFact)
{
    unsigned start = 0;
    unsigned end = facts.size()-1;

    while (start != end) {
        unsigned center = (start + end) / 2;
        if (facts[center] < newFact) {
            start = center+1;
        } else {
            end = center;
        }
    }

    unsigned index = start;
    for (unsigned i = facts.size()-1; i > index; i--) {
        facts[i] = facts[i-1];
    }
    facts[index] = newFact;
}


template<class FeatureSpace, class Fact>
void Node<FeatureSpace, Fact>::getkNearestNeighbors(const typename FeatureSpace::Features &features,
                                                     std::vector<QueueEntry<FeatureSpace, Fact>> &facts,
                                                     unsigned k) const
{
    typename FeatureSpace::Metric::Type metric = FeatureSpace::Metric::compute(features, m_fact);

    if (facts.size() < k) {
        facts.push_back(QueueEntry<FeatureSpace, Fact>{metric, m_fact.getHandle()});

        if (facts.size() == k)
            std::sort(facts.begin(), facts.end());
    } else {
        if (metric < facts[k-1].metric)
            insertSorted<FeatureSpace, Fact>(facts, QueueEntry<FeatureSpace, Fact>{metric, m_fact.getHandle()});
    }

    if (m_splittinFeature != FeatureSpace::NO_SPLIT) {
        unsigned first = FeatureSpace::compare(m_fact, features, m_splittinFeature);

        if (m_subNodes[first] != nullptr)
            m_subNodes[first]->getkNearestNeighbors(features, facts, k);

        if ((facts.size() < k) || FeatureSpace::Metric::stillInRange(m_fact, features, facts[k-1].metric, m_splittinFeature)) {
            unsigned second = first ^ 1;
            if (m_subNodes[second] != nullptr)
                m_subNodes[second]->getkNearestNeighbors(features, facts, k);
        }
    }
}


template<class FeatureSpace, class Fact>
void Node<FeatureSpace, Fact>::getNearestNeighbor(const typename FeatureSpace::Features &features,
                                                    QueueEntry<FeatureSpace, Fact> &fact) const
{
    typename FeatureSpace::Metric::Type metric = FeatureSpace::Metric::compute(features, m_fact);

    if (metric < fact.metric)
        fact = QueueEntry<FeatureSpace, Fact>{metric, m_fact.getHandle()};

    if (m_splittinFeature != FeatureSpace::NO_SPLIT) {
        unsigned first = FeatureSpace::compare(m_fact, features, m_splittinFeature);

        if (m_subNodes[first] != nullptr)
            m_subNodes[first]->getNearestNeighbor(features, fact);

        if (FeatureSpace::Metric::stillInRange(m_fact, features, fact.metric, m_splittinFeature)) {
            unsigned second = first ^ 1;
            if (m_subNodes[second] != nullptr)
                m_subNodes[second]->getNearestNeighbor(features, fact);
        }
    }
}



template<class FeatureSpace, typename Fact>
void Dataset<FeatureSpace, Fact>::addFact(const Fact &fact)
{
    if (m_rootNode == nullptr)
        m_rootNode.reset(new Node<FeatureSpace, Fact>(fact));
    else
        m_rootNode->addFact(fact);
}

template<class FeatureSpace, typename Fact>
void Dataset<FeatureSpace, Fact>::getkNearestNeighbors(const typename FeatureSpace::Features &features,
                                                        std::vector<QueueEntry<FeatureSpace, Fact>> &facts,
                                                        unsigned k) const
{
    facts.clear();
    facts.reserve(k);

    if (m_rootNode != nullptr)
        m_rootNode->getkNearestNeighbors(features, facts, k);
}

template<class FeatureSpace, typename Fact>
void Dataset<FeatureSpace, Fact>::getNearestNeighbor(const typename FeatureSpace::Features &features,
                                                     QueueEntry<FeatureSpace, Fact> &fact) const
{
    fact.metric = 1e30f;

    if (m_rootNode != nullptr)
        m_rootNode->getNearestNeighbor(features, fact);
}

}
}
}

#endif