#include "CameraResponseCurve.h"

#ifdef SYB3R_HAS_EXIV2
#include <exiv2/exiv2.hpp>
#endif
#include <tinyxml2.h>

#include "../tools/RasterImage.h"

#include <sstream>
#include <stdexcept>
#include <random>

namespace syb3r {
namespace models {

template<unsigned numLdrValues>
CameraResponseCurve<numLdrValues> CameraResponseCurve<numLdrValues>::readFromXML(tinyxml2::XMLElement *rootNode, const char *nodeName)
{
    tinyxml2::XMLElement *curveNode = rootNode->FirstChildElement(nodeName);
    if (curveNode == nullptr)
        throw std::runtime_error("Xml tag for curve not found!");

    CameraResponseCurve<numLdrValues> result;
    
    std::stringstream values(curveNode->GetText());
    for (unsigned i = 0; i < numLdrValues; i++)
        values >> result.m_hdrValues[i];
    
    if (!values)
        throw std::runtime_error("Error parsing values of CameraResponseCurve!");
    
    return result;
}


template<unsigned numLdrValues>
void CameraResponseCurve<numLdrValues>::writeToXML(tinyxml2::XMLElement *rootNode, const char *nodeName) const
{
    std::stringstream values;
    for (unsigned i = 0; i < numLdrValues; i++)
        values << m_hdrValues[i] << ' ';
    
    tinyxml2::XMLElement *curveNode = rootNode->GetDocument()->NewElement(nodeName);
    rootNode->LinkEndChild(curveNode);
    curveNode->SetText(values.str().c_str());
}


template<unsigned numLdrValues>
CameraRGBResponseCurves<numLdrValues> CameraRGBResponseCurves<numLdrValues>::readFromXML(tinyxml2::XMLElement *node)
{
    CameraRGBResponseCurves<numLdrValues> result;
    result.r = CameraResponseCurve<numLdrValues>::readFromXML(node, "CameraResponseCurve_R");
    result.g = CameraResponseCurve<numLdrValues>::readFromXML(node, "CameraResponseCurve_G");
    result.b = CameraResponseCurve<numLdrValues>::readFromXML(node, "CameraResponseCurve_B");    
   
    return result;
}

template<unsigned numLdrValues>
CameraRGBResponseCurves<numLdrValues> CameraRGBResponseCurves<numLdrValues>::readFromXML(const boost::filesystem::path &filename)
{
    tinyxml2::XMLDocument xmlDoc;
    
    if (xmlDoc.LoadFile(filename.string().c_str()) != tinyxml2::XML_NO_ERROR)
        throw std::runtime_error(std::string("Could not open or parse xml file! (") + filename.string() + ")");

    tinyxml2::XMLElement* pRoot = xmlDoc.FirstChildElement();
    if (pRoot == nullptr) throw std::runtime_error("No root node in xml file!");
    
    return readFromXML(pRoot);
}

template<unsigned numLdrValues>
void CameraRGBResponseCurves<numLdrValues>::writeToXML(tinyxml2::XMLElement *node) const
{
    r.writeToXML(node, "CameraResponseCurve_R");
    g.writeToXML(node, "CameraResponseCurve_G");
    b.writeToXML(node, "CameraResponseCurve_B");
}

template<unsigned numLdrValues>
void CameraRGBResponseCurves<numLdrValues>::writeToXML(const boost::filesystem::path &filename) const
{
    tinyxml2::XMLDocument xmlDoc;

    tinyxml2::XMLElement* pRoot = xmlDoc.NewElement("CameraRGBResponseCurves");
    xmlDoc.InsertEndChild(pRoot);

    writeToXML(pRoot);
    
    if (xmlDoc.SaveFile(filename.string().c_str()) != tinyxml2::XML_NO_ERROR)
        throw std::runtime_error("Could not save scene xml file!");        
}





#ifdef SYB3R_HAS_EXIV2

/// Based on Paul Debevec "Recovering High Dynamic Range Radiance Maps from Photographs"
void computeCameraResponseFromMultipleExposures(const std::vector<std::string> &jpegFilenames, unsigned numPixels, float smoothness,
                                                CameraResponseCurve<256> &redCurve, CameraResponseCurve<256> &greenCurve, CameraResponseCurve<256> &blueCurve)

{
    std::vector<float> exposureFactors;
    exposureFactors.resize(jpegFilenames.size());

    float iso = -1.0f; // Prevent uninitialized var warning...
    float fNumber = -1.0f;

    for (unsigned i = 0; i < jpegFilenames.size(); i++) {
        Exiv2::Image::AutoPtr image = Exiv2::ImageFactory::open(jpegFilenames[i]);
        if (image.get() == nullptr)
            throw std::runtime_error(std::string("Could not open file: ") + jpegFilenames[i]);

        image->readMetadata();

        Exiv2::ExifData &exifData = image->exifData();
        if (exifData.empty())
            std::runtime_error(std::string("No exif data in file: ") + jpegFilenames[i]);

        auto exposureTimeIt = exifData.findKey(Exiv2::ExifKey("Exif.Photo.ExposureTime"));
        if (exposureTimeIt == exifData.end())
            std::runtime_error(std::string("No exif tag 'Exif.Photo.ExposureTime' in file: ") + jpegFilenames[i]);

        auto fNumberIt = exifData.findKey(Exiv2::ExifKey("Exif.Photo.FNumber"));
        if (fNumberIt == exifData.end())
            std::runtime_error(std::string("No exif tag 'Exif.Photo.FNumber' in file: ") + jpegFilenames[i]);

        auto isoSpeedIt = exifData.findKey(Exiv2::ExifKey("Exif.Photo.ISOSpeedRatings"));
        if (isoSpeedIt == exifData.end())
            std::runtime_error(std::string("No exif tag 'Exif.Photo.ISOSpeedRatings' in file: ") + jpegFilenames[i]);


        if (i == 0) {
            iso = isoSpeedIt->toFloat();
            fNumber = fNumberIt->toFloat();
        } else {
            if (iso != isoSpeedIt->toFloat())
                std::cout << "Warning: ISO speed changed!" << std::endl;
            if (fNumber != fNumberIt->toFloat())
                std::cout << "Warning: FNumber changed!" << std::endl;
        }

        float exposure = exposureTimeIt->toFloat() * isoSpeedIt->toFloat() / (fNumberIt->toFloat()*fNumberIt->toFloat());

        //std::cout << i << "   " << exposureTimeIt->toFloat() << "  " << fNumberIt->toFloat() << "  " << isoSpeedIt->toFloat() << std::endl;
        //std::cout << "Exposure factor: " << exposure << std::endl;

        exposureFactors[i] = exposure;
/*
        Exiv2::ExifData::const_iterator end = exifData.end();
        for (Exiv2::ExifData::const_iterator i = exifData.begin(); i != end; ++i) {
            const char* tn = i->typeName();
            std::cout << std::setw(44) << std::setfill(' ') << std::left
                    << i->key() << " "
                    << "0x" << std::setw(4) << std::setfill('0') << std::right
                    << std::hex << i->tag() << " "
                    << std::setw(9) << std::setfill(' ') << std::left
                    << (tn ? tn : "Unknown") << " "
                    << std::dec << std::setw(3)
                    << std::setfill(' ') << std::right
                    << i->count() << "  "
                    << std::dec << i->value()
                    << "\n";
        }
*/
    }

    std::vector<Eigen::Vector2i> pixelLocations;
    pixelLocations.resize(numPixels);
    std::mt19937 rng(31337);

    unsigned width, height;
    RasterImage::getImageFileDimensions(jpegFilenames[0].c_str(), width, height);

    std::uniform_int_distribution<unsigned> randomX(0, width-1);
    std::uniform_int_distribution<unsigned> randomY(0, height-1);

    for (unsigned i = 0; i < pixelLocations.size(); i++) {
        pixelLocations[i][0] = randomX(rng);
        pixelLocations[i][1] = randomY(rng);
    }


    std::vector<unsigned> pixelValues[3];
    pixelValues[0].resize(numPixels * jpegFilenames.size());
    pixelValues[1].resize(numPixels * jpegFilenames.size());
    pixelValues[2].resize(numPixels * jpegFilenames.size());

    for (unsigned i = 0; i < jpegFilenames.size(); i++) {
        RasterImage image;
        image.loadFromFile(jpegFilenames[i].c_str());

        if ((image.getWidth() != width) || (image.getHeight() != height))
            throw std::runtime_error(std::string("All images must have the same resolution, but ") + jpegFilenames[i] + " differs");

        for (unsigned j = 0; j < numPixels; j++) {
            uint32_t c = image.getData()[pixelLocations[j][0] + pixelLocations[j][1] * width];

            pixelValues[0][i * numPixels + j] = (c >> 0) & 0xFF;
            pixelValues[1][i * numPixels + j] = (c >> 8) & 0xFF;
            pixelValues[2][i * numPixels + j] = (c >> 16) & 0xFF;
        }
    }

    computeCameraResponseFromMultipleExposures<256>(pixelValues[0], exposureFactors, smoothness, redCurve);
    computeCameraResponseFromMultipleExposures<256>(pixelValues[1], exposureFactors, smoothness, greenCurve);
    computeCameraResponseFromMultipleExposures<256>(pixelValues[2], exposureFactors, smoothness, blueCurve);
}


#endif



///@todo Move this somewhere

syb3r::models::CameraRGBResponseCurves<256> curves_Eos400D_neutral_5200K_daylight = {
    {
        0.0233742, 0.0263379, 0.0296775, 0.0334078, 0.0375327, 0.0420487, 0.0469466, 0.0522129, 0.0578269, 0.0637607, 0.0699807, 0.0764472, 0.0831191, 0.0899556, 0.0969146, 0.103954,
        0.111032, 0.118111, 0.125156, 0.132134, 0.139018, 0.145785, 0.152416, 0.158897, 0.165218, 0.171373, 0.177358, 0.183174, 0.188824, 0.194317, 0.19966, 0.204864,
        0.209939, 0.214899, 0.219756, 0.224523, 0.229214, 0.233843, 0.238425, 0.242972, 0.247497, 0.25201, 0.256521, 0.261039, 0.26557, 0.270121, 0.274698, 0.279306,
        0.283949, 0.288632, 0.293361, 0.29814, 0.302972, 0.307861, 0.312812, 0.317828, 0.322915, 0.328077, 0.333319, 0.338645, 0.344059, 0.349567, 0.355174, 0.360885,
        0.366706, 0.372643, 0.3787, 0.384885, 0.391202, 0.397657, 0.404256, 0.411003, 0.417903, 0.424957, 0.432172, 0.439548, 0.447089, 0.454796, 0.462671, 0.470716,
        0.478929, 0.48731, 0.495857, 0.504566, 0.513432, 0.522451, 0.531617, 0.540922, 0.550361, 0.559932, 0.569631, 0.579456, 0.589401, 0.599464, 0.60964, 0.619925,
        0.630311, 0.640792, 0.651365, 0.662026, 0.67277, 0.683592, 0.694487, 0.705454, 0.716491, 0.727598, 0.738778, 0.750033, 0.761369, 0.772791, 0.784302, 0.795912,
        0.807626, 0.819453, 0.831396, 0.843463, 0.85566, 0.867991, 0.880465, 0.893086, 0.905863, 0.918801, 0.931906, 0.94518, 0.958627, 0.972246, 0.986037, 1,
        1.01413, 1.02844, 1.04291, 1.05755, 1.07235, 1.08732, 1.10245, 1.11773, 1.13317, 1.14878, 1.16455, 1.18048, 1.19658, 1.21284, 1.22926, 1.24584,
        1.26257, 1.27947, 1.29654, 1.31377, 1.33119, 1.34878, 1.36656, 1.38454, 1.40272, 1.42111, 1.43972, 1.45856, 1.47764, 1.49697, 1.51657, 1.53646,
        1.55664, 1.57712, 1.59794, 1.61909, 1.6406, 1.66248, 1.68473, 1.70737, 1.73042, 1.75389, 1.7778, 1.80214, 1.82695, 1.85224, 1.87801, 1.90429,
        1.93107, 1.95838, 1.98622, 2.0146, 2.04352, 2.07299, 2.10302, 2.13362, 2.16478, 2.19651, 2.22882, 2.2617, 2.29517, 2.32924, 2.36391, 2.3992,
        2.43512, 2.47167, 2.50887, 2.54673, 2.58527, 2.6245, 2.66443, 2.70511, 2.74654, 2.78874, 2.83176, 2.87561, 2.92033, 2.96594, 3.0125, 3.06003,
        3.10858, 3.15818, 3.20886, 3.26065, 3.3136, 3.36775, 3.42316, 3.47988, 3.53798, 3.59751, 3.65855, 3.72114, 3.78533, 3.85117, 3.91872, 3.98804,
        4.05919, 4.13225, 4.2073, 4.28442, 4.3637, 4.44524, 4.52915, 4.61551, 4.70442, 4.79598, 4.89027, 4.98738, 5.08739, 5.19035, 5.29635, 5.40544,
        5.51765, 5.63303, 5.75161, 5.87343, 5.99852, 6.12695, 6.25878, 6.39407, 6.53285, 6.67515, 6.82098, 6.97033, 7.1232, 7.27958, 7.43947, 7.60288,
    },
    {
        0.0156389, 0.0179012, 0.0204907, 0.0234481, 0.0267979, 0.030557, 0.0347349, 0.0393348, 0.0443537, 0.0497797, 0.0555919, 0.0617602, 0.0682475, 0.0750102, 0.0820028, 0.0891791,
        0.0964909, 0.10389, 0.111332, 0.118773, 0.126173, 0.133499, 0.140719, 0.147811, 0.154754, 0.161535, 0.168144, 0.174579, 0.180841, 0.186932, 0.192858, 0.198628,
        0.204252, 0.209741, 0.215107, 0.220363, 0.225524, 0.230603, 0.235615, 0.240571, 0.245484, 0.250365, 0.255224, 0.260071, 0.264914, 0.269761, 0.27462, 0.279497,
        0.284399, 0.289331, 0.294298, 0.299303, 0.30435, 0.309442, 0.314582, 0.31977, 0.325011, 0.330306, 0.335657, 0.341068, 0.346541, 0.352078, 0.357684, 0.36336,
        0.36911, 0.374938, 0.380845, 0.386837, 0.392917, 0.399088, 0.405354, 0.411721, 0.418193, 0.424776, 0.431478, 0.438304, 0.445263, 0.452361, 0.459604, 0.467,
        0.474554, 0.482273, 0.490163, 0.498227, 0.506471, 0.514896, 0.523507, 0.532304, 0.54129, 0.550465, 0.559828, 0.569378, 0.579113, 0.58903, 0.599123, 0.609389,
        0.619821, 0.630412, 0.641155, 0.652041, 0.663063, 0.674213, 0.685484, 0.696868, 0.708361, 0.719956, 0.731649, 0.743435, 0.755307, 0.767264, 0.779303, 0.791424,
        0.803628, 0.815916, 0.828291, 0.840757, 0.853316, 0.865975, 0.87874, 0.891621, 0.904625, 0.917763, 0.931046, 0.944484, 0.958089, 0.971871, 0.985838, 1,
        1.01436, 1.02894, 1.04373, 1.05874, 1.07399, 1.08946, 1.10518, 1.12113, 1.13731, 1.15373, 1.17039, 1.18728, 1.2044, 1.22174, 1.2393, 1.25707,
        1.27505, 1.29322, 1.31159, 1.33013, 1.34886, 1.36776, 1.38683, 1.40607, 1.42548, 1.44506, 1.46481, 1.48472, 1.50479, 1.52503, 1.54543, 1.566,
        1.58675, 1.60769, 1.62883, 1.65017, 1.67174, 1.69354, 1.71559, 1.73792, 1.76053, 1.78346, 1.80672, 1.83035, 1.85436, 1.87879, 1.90366, 1.92901,
        1.95485, 1.98121, 2.00812, 2.03559, 2.06366, 2.09233, 2.12163, 2.15159, 2.18221, 2.21351, 2.2455, 2.27821, 2.31164, 2.3458, 2.38072, 2.4164,
        2.45286, 2.4901, 2.52814, 2.567, 2.6067, 2.64725, 2.68868, 2.73101, 2.77428, 2.81853, 2.86378, 2.91008, 2.9575, 3.00609, 3.05591, 3.10703,
        3.15954, 3.21352, 3.26904, 3.32622, 3.38514, 3.4459, 3.50863, 3.57343, 3.64042, 3.70972, 3.78146, 3.85575, 3.93271, 4.01246, 4.09511, 4.18077,
        4.26954, 4.36152, 4.45684, 4.55555, 4.65775, 4.76351, 4.87293, 4.98609, 5.10309, 5.22402, 5.349, 5.47814, 5.61157, 5.74942, 5.89181, 6.0389,
        6.19082, 6.34769, 6.50966, 6.67683, 6.84931, 7.02721, 7.2106, 7.39956, 7.59415, 7.79444, 8.00048, 8.21233, 8.43007, 8.65376, 8.88349, 9.11931,
    },
    {
        0.0253718, 0.0285119, 0.0320407, 0.0359743, 0.0403231, 0.04509, 0.0502739, 0.0558671, 0.0618538, 0.0682107, 0.0749079, 0.0819103, 0.0891793, 0.0966746, 0.104354, 0.112172,
        0.120085, 0.12805, 0.136024, 0.143971, 0.151855, 0.159646, 0.167321, 0.174861, 0.182251, 0.189483, 0.196552, 0.203459, 0.210204, 0.216793, 0.223234, 0.229536,
        0.235711, 0.241769, 0.247719, 0.25357, 0.259334, 0.265024, 0.270653, 0.276232, 0.28177, 0.287274, 0.292751, 0.29821, 0.303656, 0.309093, 0.314529, 0.319969,
        0.325418, 0.330881, 0.336362, 0.341868, 0.347404, 0.352975, 0.358586, 0.364242, 0.369948, 0.375708, 0.381526, 0.387406, 0.393353, 0.399369, 0.405458, 0.411623,
        0.417866, 0.424188, 0.430591, 0.437073, 0.443634, 0.450273, 0.456988, 0.463775, 0.470633, 0.477557, 0.484546, 0.491596, 0.498705, 0.505872, 0.513094, 0.52037,
        0.5277, 0.535081, 0.542514, 0.549998, 0.557534, 0.565121, 0.572763, 0.58046, 0.588216, 0.596032, 0.603912, 0.611861, 0.619885, 0.627989, 0.63618, 0.644465,
        0.652853, 0.661351, 0.669968, 0.678714, 0.687598, 0.696629, 0.705815, 0.715165, 0.724686, 0.734384, 0.744265, 0.754334, 0.764596, 0.775057, 0.78572, 0.796589,
        0.807668, 0.818959, 0.830467, 0.842192, 0.854138, 0.866306, 0.878698, 0.891315, 0.904154, 0.917215, 0.930493, 0.943985, 0.957687, 0.971594, 0.985699, 1,
        1.01449, 1.02917, 1.04402, 1.05906, 1.07427, 1.08965, 1.10521, 1.12094, 1.13684, 1.15292, 1.16918, 1.18561, 1.20223, 1.21904, 1.23605, 1.25325,
        1.27067, 1.28831, 1.30619, 1.32431, 1.34268, 1.36131, 1.38023, 1.39943, 1.41893, 1.43873, 1.45884, 1.47927, 1.50001, 1.52109, 1.54249, 1.56422,
        1.58627, 1.60866, 1.63137, 1.65441, 1.67777, 1.70144, 1.72543, 1.74973, 1.77433, 1.79922, 1.8244, 1.84986, 1.87559, 1.90159, 1.92783, 1.95432,
        1.98105, 2.00801, 2.0352, 2.06262, 2.09025, 2.11811, 2.1462, 2.17452, 2.20308, 2.23188, 2.26093, 2.29026, 2.31988, 2.3498, 2.38007, 2.41071,
        2.44174, 2.47322, 2.50518, 2.53766, 2.57072, 2.6044, 2.63878, 2.6739, 2.70983, 2.74665, 2.78443, 2.82324, 2.86317, 2.90429, 2.94671, 2.99052,
        3.03583, 3.08275, 3.13142, 3.18199, 3.2346, 3.28943, 3.34666, 3.4065, 3.46916, 3.53489, 3.60395, 3.67663, 3.75323, 3.83409, 3.91957, 4.01009,
        4.10605, 4.20793, 4.31621, 4.43144, 4.55418, 4.68505, 4.82468, 4.97376, 5.13301, 5.30317, 5.48499, 5.67923, 5.88659, 6.10772, 6.34318, 6.5935,
        6.85924, 7.14102, 7.43948, 7.75528, 8.08914, 8.44178, 8.81394, 9.20633, 9.61964, 10.0545, 10.5115, 10.9912, 11.4941, 12.0207, 12.5718, 13.1481,
    }
};

syb3r::models::CameraRGBResponseCurves<256> curves_Eos400D_neutral_3200K_artificial = {
    {
        0.0272449, 0.0301951, 0.0334648, 0.0370575, 0.040972, 0.0452024, 0.0497357, 0.0545526, 0.0596292, 0.0649405, 0.0704614, 0.0761646, 0.0820218, 0.0880038, 0.0940821, 0.100228,
        0.106415, 0.112615, 0.118807, 0.12497, 0.131087, 0.137145, 0.143134, 0.149045, 0.154875, 0.160623, 0.166288, 0.171873, 0.177381, 0.182817, 0.188185, 0.193493,
        0.198747, 0.203954, 0.209123, 0.21426, 0.219371, 0.224458, 0.229526, 0.234575, 0.239611, 0.244636, 0.249653, 0.254666, 0.259678, 0.264694, 0.269717, 0.274755,
        0.279813, 0.284898, 0.290016, 0.295176, 0.300386, 0.305653, 0.310988, 0.316397, 0.32189, 0.327473, 0.333155, 0.338942, 0.344841, 0.350857, 0.356997, 0.363264,
        0.369664, 0.376201, 0.382877, 0.389697, 0.39666, 0.403766, 0.411015, 0.418402, 0.425926, 0.433582, 0.441368, 0.449279, 0.457313, 0.465466, 0.473735, 0.482117,
        0.490608, 0.499204, 0.507901, 0.516695, 0.52558, 0.534554, 0.543613, 0.552755, 0.561979, 0.571285, 0.580672, 0.590141, 0.599691, 0.609323, 0.619037, 0.628837,
        0.638721, 0.648692, 0.658752, 0.668903, 0.679147, 0.689488, 0.699928, 0.71047, 0.721117, 0.731873, 0.74274, 0.753724, 0.764829, 0.776057, 0.787409, 0.798887,
        0.81049, 0.822219, 0.834072, 0.846049, 0.858152, 0.870378, 0.882732, 0.895213, 0.907825, 0.920569, 0.93345, 0.946471, 0.959634, 0.972942, 0.986397, 1,
        1.01375, 1.02765, 1.04169, 1.05588, 1.07022, 1.0847, 1.09933, 1.1141, 1.12904, 1.14413, 1.15939, 1.17481, 1.19042, 1.20621, 1.22219, 1.23837,
        1.25475, 1.27133, 1.28811, 1.30511, 1.32231, 1.33971, 1.35733, 1.37515, 1.39318, 1.41141, 1.42984, 1.44847, 1.4673, 1.48634, 1.50558, 1.52502,
        1.54468, 1.56455, 1.58463, 1.60492, 1.62541, 1.64609, 1.66696, 1.688, 1.70921, 1.73058, 1.75211, 1.77378, 1.7956, 1.81756, 1.83966, 1.86189,
        1.88426, 1.90677, 1.92943, 1.95224, 1.9752, 1.99834, 2.02165, 2.04515, 2.06885, 2.09275, 2.11688, 2.14122, 2.16581, 2.19067, 2.21581, 2.24127,
        2.26706, 2.29323, 2.31981, 2.34685, 2.37437, 2.40243, 2.43107, 2.46038, 2.49039, 2.52118, 2.55282, 2.58537, 2.61891, 2.65351, 2.68924, 2.7262,
        2.76446, 2.8041, 2.84523, 2.88794, 2.93233, 2.97849, 3.02655, 3.0766, 3.12874, 3.18308, 3.23972, 3.29877, 3.36033, 3.42452, 3.49141, 3.5611,
        3.6337, 3.70929, 3.78798, 3.86984, 3.95497, 4.04344, 4.13528, 4.23055, 4.32928, 4.43153, 4.53729, 4.64657, 4.75942, 4.87587, 4.996, 5.11984,
        5.24746, 5.37894, 5.51437, 5.65384, 5.79742, 5.94519, 6.09726, 6.25371, 6.41459, 6.58, 6.75001, 6.92468, 7.10406, 7.28821, 7.47716, 7.67101,
    },
    {
        0.0153631, 0.0175618, 0.0200751, 0.022946, 0.0261979, 0.0298479, 0.0339048, 0.0383695, 0.0432326, 0.0484814, 0.0540977, 0.0600534, 0.066314, 0.0728427, 0.0796015, 0.0865492,
        0.0936432, 0.100842, 0.108106, 0.115398, 0.122684, 0.129934, 0.13712, 0.144218, 0.151209, 0.158075, 0.164805, 0.171388, 0.177819, 0.184099, 0.190229, 0.196213,
        0.202057, 0.20777, 0.213358, 0.218834, 0.224209, 0.229497, 0.234712, 0.239867, 0.244976, 0.250051, 0.255107, 0.260155, 0.265207, 0.270272, 0.27536, 0.28048,
        0.285641, 0.290847, 0.296105, 0.301416, 0.306786, 0.312219, 0.317718, 0.323285, 0.32892, 0.334626, 0.340405, 0.346257, 0.352185, 0.358189, 0.364271, 0.37043,
        0.376668, 0.382986, 0.389383, 0.395861, 0.402421, 0.409067, 0.415798, 0.42262, 0.429533, 0.436543, 0.443653, 0.450866, 0.458188, 0.465621, 0.473169, 0.480838,
        0.488631, 0.496552, 0.504604, 0.51279, 0.521112, 0.529572, 0.538172, 0.546914, 0.555799, 0.564827, 0.573998, 0.583311, 0.592764, 0.602357, 0.612085, 0.621947,
        0.631939, 0.642057, 0.652296, 0.662654, 0.673125, 0.683707, 0.694398, 0.705193, 0.716092, 0.727091, 0.73819, 0.749386, 0.76068, 0.772072, 0.783563, 0.795156,
        0.806857, 0.818668, 0.830597, 0.842649, 0.854832, 0.867153, 0.87962, 0.892243, 0.905032, 0.917996, 0.931144, 0.944486, 0.958033, 0.971794, 0.98578, 1,
        1.01446, 1.02917, 1.04414, 1.05937, 1.07487, 1.09064, 1.10669, 1.12302, 1.13963, 1.15653, 1.1737, 1.19115, 1.20888, 1.22687, 1.24511, 1.26361,
        1.28236, 1.30135, 1.32057, 1.34001, 1.35967, 1.37955, 1.39964, 1.41993, 1.44043, 1.46114, 1.48206, 1.50319, 1.52453, 1.54609, 1.56787, 1.58988,
        1.61214, 1.63465, 1.65743, 1.68048, 1.70383, 1.72748, 1.75146, 1.77577, 1.80043, 1.82546, 1.85086, 1.87665, 1.90284, 1.92944, 1.95646, 1.98392,
        2.01183, 2.04018, 2.06898, 2.09824, 2.12796, 2.15815, 2.18881, 2.21995, 2.25157, 2.28367, 2.31626, 2.34933, 2.38288, 2.41691, 2.45142, 2.48641,
        2.52189, 2.55787, 2.59437, 2.63139, 2.66896, 2.7071, 2.74584, 2.78523, 2.82532, 2.86617, 2.90785, 2.95042, 2.99397, 3.03858, 3.08433, 3.13134,
        3.17972, 3.22957, 3.28102, 3.3342, 3.38925, 3.44631, 3.50553, 3.56706, 3.63106, 3.69769, 3.76709, 3.83943, 3.91486, 3.99355, 4.07565, 4.16132,
        4.25073, 4.34407, 4.4415, 4.54321, 4.6494, 4.76025, 4.876, 4.9969, 5.12318, 5.25512, 5.39301, 5.53713, 5.68779, 5.8453, 6.00998, 6.18215,
        6.36215, 6.55032, 6.74699, 6.95242, 7.16688, 7.39063, 7.62392, 7.86692, 8.11979, 8.38266, 8.65559, 8.93861, 9.2317, 9.5348, 9.84796, 10.1714,
    },
    {
        0.0345767, 0.0375708, 0.0408242, 0.044353, 0.0481733, 0.0522977, 0.0567364, 0.0614953, 0.0665731, 0.0719626, 0.0776529, 0.0836322, 0.0898831, 0.0963881, 0.103125, 0.110069,
        0.117193, 0.124469, 0.131864, 0.139345, 0.146876, 0.154427, 0.161972, 0.169488, 0.176955, 0.184357, 0.19168, 0.198913, 0.206048, 0.213079, 0.220004, 0.226821,
        0.23353, 0.240133, 0.246632, 0.253033, 0.25934, 0.265558, 0.271694, 0.277756, 0.283749, 0.289683, 0.295566, 0.301407, 0.307215, 0.312996, 0.318759, 0.324511,
        0.330259, 0.33601, 0.34177, 0.347547, 0.353348, 0.359182, 0.365055, 0.370976, 0.376952, 0.382988, 0.389089, 0.39526, 0.401503, 0.407822, 0.414218, 0.420692,
        0.427248, 0.433887, 0.440612, 0.447424, 0.454328, 0.461324, 0.468414, 0.475596, 0.482871, 0.490235, 0.497688, 0.505224, 0.512841, 0.520536, 0.528306, 0.536148,
        0.544061, 0.552044, 0.560097, 0.568218, 0.57641, 0.584673, 0.593009, 0.601417, 0.6099, 0.618456, 0.627087, 0.635792, 0.644573, 0.653429, 0.662364, 0.671377,
        0.680469, 0.689644, 0.698902, 0.708247, 0.717681, 0.727205, 0.736821, 0.746532, 0.756337, 0.766235, 0.776222, 0.786293, 0.796446, 0.806675, 0.816976, 0.827343,
        0.83777, 0.848253, 0.858789, 0.869372, 0.880002, 0.890678, 0.9014, 0.912167, 0.922977, 0.933832, 0.944733, 0.955683, 0.966682, 0.977733, 0.988839, 1,
        1.01122, 1.0225, 1.03384, 1.04525, 1.05673, 1.06829, 1.07994, 1.09168, 1.10352, 1.11547, 1.12755, 1.13976, 1.15212, 1.16464, 1.17733, 1.1902,
        1.20327, 1.21655, 1.23006, 1.24382, 1.25783, 1.27212, 1.28669, 1.30157, 1.31675, 1.33226, 1.34811, 1.3643, 1.38084, 1.39775, 1.41504, 1.43272,
        1.45079, 1.46926, 1.48814, 1.50743, 1.52714, 1.54725, 1.56775, 1.58865, 1.60992, 1.63154, 1.65351, 1.67581, 1.69843, 1.72134, 1.74454, 1.768,
        1.79172, 1.81568, 1.83985, 1.86422, 1.88876, 1.91346, 1.9383, 1.96326, 1.98833, 2.0135, 2.03875, 2.06408, 2.08948, 2.11495, 2.14047, 2.16605,
        2.1917, 2.21742, 2.24323, 2.26913, 2.29516, 2.32133, 2.34767, 2.3742, 2.40097, 2.428, 2.45535, 2.48305, 2.51115, 2.5397, 2.56877, 2.59841,
        2.62871, 2.65975, 2.69167, 2.72459, 2.7587, 2.79417, 2.83123, 2.87013, 2.91115, 2.95463, 3.00092, 3.05046, 3.10375, 3.16132, 3.22378, 3.29182,
        3.36621, 3.44781, 3.5376, 3.63668, 3.74629, 3.86785, 4.00298, 4.15353, 4.32163, 4.50971, 4.72059, 4.95751, 5.22423, 5.52512, 5.86523, 6.25041,
        6.68728, 7.18338, 7.74715, 8.38774, 9.11479, 9.93725, 10.864, 11.9042, 13.0678, 14.3647, 15.8046, 17.3974, 19.1548, 21.0918, 23.2251, 25.5741,
    }
};

syb3r::models::CameraRGBResponseCurves<256> curves_Eos400D_neutral_7000K_shade = {
    {
        0.0224884, 0.0253708, 0.0286226, 0.0322737, 0.0363365, 0.0408164, 0.0457118, 0.0510152, 0.0567122, 0.0627816, 0.0691963, 0.0759223, 0.0829212, 0.0901494, 0.0975623, 0.105114,
        0.11276, 0.120454, 0.128154, 0.135821, 0.143419, 0.150916, 0.158284, 0.165499, 0.172544, 0.179404, 0.186069, 0.192532, 0.198791, 0.204849, 0.210708, 0.216378,
        0.221869, 0.227191, 0.232357, 0.23738, 0.242272, 0.247046, 0.251718, 0.256301, 0.260806, 0.265249, 0.26964, 0.273994, 0.278324, 0.282643, 0.28696, 0.291287,
        0.295633, 0.300006, 0.304416, 0.308868, 0.31337, 0.317928, 0.322545, 0.327228, 0.331981, 0.336809, 0.341717, 0.346708, 0.351787, 0.356957, 0.362223, 0.367589,
        0.373059, 0.378637, 0.384327, 0.390133, 0.396058, 0.402106, 0.40828, 0.414586, 0.421025, 0.427603, 0.434323, 0.441189, 0.448204, 0.455373, 0.462699, 0.470186,
        0.477837, 0.485654, 0.493638, 0.501791, 0.510116, 0.518614, 0.527287, 0.536137, 0.545164, 0.554366, 0.563744, 0.573296, 0.583019, 0.592911, 0.602968, 0.613188,
        0.623565, 0.634095, 0.644772, 0.65559, 0.666541, 0.67762, 0.688821, 0.700138, 0.711567, 0.723101, 0.734734, 0.746463, 0.758282, 0.770186, 0.782173, 0.794239,
        0.806384, 0.818606, 0.830906, 0.843286, 0.855747, 0.868295, 0.880934, 0.893668, 0.906506, 0.919456, 0.932526, 0.945725, 0.959064, 0.972552, 0.986194, 1,
        1.01398, 1.02814, 1.04248, 1.05703, 1.07178, 1.08674, 1.10191, 1.11729, 1.13289, 1.14871, 1.16475, 1.18101, 1.19749, 1.21419, 1.23111, 1.24825,
        1.26559, 1.28313, 1.30087, 1.31881, 1.33695, 1.3553, 1.37385, 1.3926, 1.41157, 1.43075, 1.45014, 1.46975, 1.48958, 1.50965, 1.52995, 1.5505,
        1.57131, 1.59239, 1.61376, 1.63542, 1.6574, 1.6797, 1.70235, 1.72536, 1.74874, 1.77251, 1.7967, 1.82131, 1.84638, 1.87191, 1.89794, 1.92448,
        1.95155, 1.97915, 2.00731, 2.03603, 2.06532, 2.0952, 2.12568, 2.15677, 2.18848, 2.22084, 2.25384, 2.2875, 2.32182, 2.3568, 2.39245, 2.42876,
        2.46574, 2.50337, 2.54167, 2.58063, 2.62026, 2.66056, 2.70156, 2.74326, 2.78569, 2.82887, 2.87283, 2.9176, 2.96321, 3.00971, 3.05714, 3.10555,
        3.15499, 3.20554, 3.25724, 3.31018, 3.36444, 3.42009, 3.47722, 3.53592, 3.59631, 3.65848, 3.72255, 3.78862, 3.85682, 3.92726, 4.00006, 4.07535,
        4.15324, 4.23386, 4.31735, 4.40384, 4.49347, 4.58639, 4.68274, 4.78269, 4.88636, 4.99391, 5.10546, 5.22117, 5.34116, 5.46558, 5.59454, 5.72818,
        5.8666, 6.0099, 6.15819, 6.31155, 6.47003, 6.63368, 6.80253, 6.97661, 7.15596, 7.34057, 7.53041, 7.72551, 7.92587, 8.13155, 8.34259, 8.55911,
    },
    {
        0.0159884, 0.0183354, 0.021027, 0.0241068, 0.0275999, 0.0315224, 0.0358829, 0.0406823, 0.0459132, 0.0515606, 0.0576022, 0.0640083, 0.0707434, 0.0777659, 0.0850307, 0.0924896,
        0.100093, 0.107792, 0.115537, 0.123284, 0.13099, 0.138617, 0.146131, 0.153504, 0.16071, 0.167731, 0.174554, 0.181169, 0.187573, 0.193764, 0.199744, 0.20552,
        0.2111, 0.216495, 0.221717, 0.226779, 0.231697, 0.236488, 0.241167, 0.24575, 0.250251, 0.254686, 0.259069, 0.263414, 0.267733, 0.272039, 0.276345, 0.280661,
        0.284999, 0.289368, 0.293776, 0.29823, 0.302738, 0.307305, 0.311938, 0.316641, 0.321419, 0.326278, 0.331223, 0.336258, 0.341387, 0.346615, 0.351945, 0.35738,
        0.362923, 0.368577, 0.374345, 0.380229, 0.386232, 0.392358, 0.39861, 0.40499, 0.411503, 0.418151, 0.424937, 0.431866, 0.43894, 0.446162, 0.453536, 0.461064,
        0.468751, 0.476599, 0.484611, 0.49279, 0.501138, 0.509658, 0.518352, 0.527223, 0.536273, 0.545503, 0.554916, 0.56451, 0.574283, 0.584235, 0.594364, 0.604666,
        0.615138, 0.625774, 0.636571, 0.647524, 0.658627, 0.669876, 0.681264, 0.692786, 0.704436, 0.716204, 0.728086, 0.740075, 0.752163, 0.764345, 0.776614, 0.78897,
        0.801411, 0.813935, 0.826544, 0.839239, 0.852022, 0.864896, 0.877866, 0.890937, 0.904116, 0.917408, 0.930822, 0.944365, 0.958045, 0.971872, 0.985854, 1,
        1.01432, 1.02882, 1.04351, 1.05839, 1.07348, 1.08878, 1.1043, 1.12003, 1.13598, 1.15216, 1.16855, 1.18517, 1.202, 1.21906, 1.23633, 1.25381,
        1.27151, 1.28941, 1.30752, 1.32583, 1.34435, 1.36306, 1.38198, 1.40111, 1.42044, 1.43998, 1.45973, 1.47968, 1.49983, 1.52019, 1.54075, 1.56152,
        1.58251, 1.60373, 1.62519, 1.64689, 1.66887, 1.69112, 1.71368, 1.73656, 1.75979, 1.78338, 1.80736, 1.83176, 1.8566, 1.88191, 1.90771, 1.93403,
        1.96089, 1.98833, 2.01636, 2.04501, 2.07429, 2.10423, 2.13484, 2.16614, 2.19815, 2.23086, 2.2643, 2.29846, 2.33335, 2.36898, 2.40535, 2.44247,
        2.48033, 2.51893, 2.55828, 2.59839, 2.63926, 2.6809, 2.72332, 2.76655, 2.81059, 2.85547, 2.90122, 2.94787, 2.99546, 3.04401, 3.09356, 3.14417,
        3.19589, 3.24878, 3.30292, 3.35836, 3.4152, 3.4735, 3.53336, 3.59486, 3.65811, 3.7232, 3.79023, 3.85931, 3.93054, 4.00405, 4.07992, 4.15829,
        4.23926, 4.32295, 4.40948, 4.49897, 4.59153, 4.68732, 4.78646, 4.88912, 4.99545, 5.10562, 5.21978, 5.33808, 5.46069, 5.58774, 5.71935, 5.85566,
        5.99676, 6.14273, 6.29366, 6.44962, 6.61064, 6.77678, 6.94807, 7.12455, 7.30625, 7.49315, 7.68524, 7.88252, 8.08504, 8.29284, 8.506, 8.72464,
    },
    {
        0.0197987, 0.0226784, 0.0259771, 0.0297256, 0.0339412, 0.0386326, 0.0438005, 0.049436, 0.0555217, 0.0620323, 0.0689339, 0.0761868, 0.0837447, 0.0915583, 0.0995755, 0.107743,
        0.116008, 0.124318, 0.132624, 0.140881, 0.149048, 0.157089, 0.164972, 0.17267, 0.180164, 0.187438, 0.194479, 0.201281, 0.207841, 0.214161, 0.220244, 0.226101,
        0.231741, 0.237173, 0.242412, 0.247473, 0.252372, 0.257125, 0.26175, 0.266263, 0.270681, 0.275019, 0.279294, 0.283521, 0.287715, 0.291888, 0.296054, 0.300224,
        0.304411, 0.308625, 0.312876, 0.317172, 0.321522, 0.325931, 0.330404, 0.334948, 0.339565, 0.34426, 0.349035, 0.353895, 0.358843, 0.363882, 0.369015, 0.374243,
        0.37957, 0.384997, 0.390527, 0.396162, 0.401903, 0.407754, 0.413718, 0.419798, 0.425997, 0.432317, 0.438762, 0.445336, 0.452041, 0.45888, 0.465858, 0.472977,
        0.480242, 0.487655, 0.49522, 0.502942, 0.510822, 0.518866, 0.527077, 0.535458, 0.544014, 0.552746, 0.56166, 0.570755, 0.580037, 0.589507, 0.599165, 0.609014,
        0.619052, 0.629279, 0.639694, 0.650294, 0.661076, 0.672038, 0.683171, 0.694472, 0.705932, 0.717546, 0.729307, 0.741207, 0.75324, 0.765398, 0.777671, 0.790052,
        0.802535, 0.815114, 0.827782, 0.840535, 0.853368, 0.86628, 0.87927, 0.892338, 0.905485, 0.918711, 0.93202, 0.945416, 0.958905, 0.972492, 0.986188, 1,
        1.01394, 1.02801, 1.04224, 1.05662, 1.07118, 1.08592, 1.10085, 1.11599, 1.13135, 1.14694, 1.16277, 1.17884, 1.19517, 1.21176, 1.22862, 1.24575,
        1.26315, 1.28082, 1.29876, 1.31696, 1.33541, 1.35413, 1.37309, 1.3923, 1.41174, 1.43142, 1.45131, 1.47143, 1.49176, 1.5123, 1.53304, 1.55396,
        1.57508, 1.59639, 1.61787, 1.63955, 1.66141, 1.68346, 1.70572, 1.72819, 1.7509, 1.77384, 1.79706, 1.82057, 1.8444, 1.86858, 1.89313, 1.91809,
        1.9435, 1.9694, 1.99581, 2.02278, 2.05035, 2.07854, 2.1074, 2.13694, 2.16721, 2.19823, 2.23002, 2.26259, 2.29598, 2.3302, 2.36524, 2.40114,
        2.43788, 2.47548, 2.51393, 2.55324, 2.5934, 2.63444, 2.67635, 2.71916, 2.76287, 2.80753, 2.85315, 2.89976, 2.94741, 2.99613, 3.04598, 3.09701,
        3.14928, 3.20288, 3.25788, 3.31436, 3.37244, 3.43221, 3.4938, 3.55734, 3.62297, 3.69085, 3.76112, 3.83395, 3.90952, 3.98799, 4.06953, 4.15434,
        4.24258, 4.33445, 4.43012, 4.52979, 4.63362, 4.74179, 4.85446, 4.97179, 5.09394, 5.22108, 5.3534, 5.49107, 5.63429, 5.78326, 5.93817, 6.09922,
        6.26657, 6.44039, 6.62082, 6.80802, 7.00212, 7.20329, 7.41159, 7.62708, 7.84982, 8.07985, 8.31716, 8.5618, 8.81384, 9.0734, 9.34064, 9.61576,
    }
};

template class CameraResponseCurve<256>;
template class CameraRGBResponseCurves<256>;


}
}