#ifndef CAMERARESPONSECURVE_H
#define CAMERARESPONSECURVE_H

#include <Eigen/Dense>

#include <boost/filesystem.hpp>

#include <vector>
#include <string>
#include <initializer_list>

#include <iostream>

namespace tinyxml2 {
    class XMLDocument;
    class XMLElement;
}


namespace syb3r {
namespace models {

template<unsigned numLdrValues>
class CameraResponseCurve
{
    public:
        static CameraResponseCurve<numLdrValues> readFromXML(tinyxml2::XMLElement *rootNode, const char *nodeName); // these only work for extern templates
        void writeToXML(tinyxml2::XMLElement *rootNode, const char *nodeName) const;
                
        CameraResponseCurve() {
            for (unsigned i = 0; i < numLdrValues; i++)
                m_hdrValues[i] = i * 2.0f / (numLdrValues - 1);
        }

        CameraResponseCurve(std::initializer_list<float> values) {
            assert((values.size() == numLdrValues) && "Number of values in initializer list must match numLdrValues");
            for (unsigned i = 0; i < numLdrValues; i++)
                m_hdrValues[i] = values.begin()[i];
        }

        float &ldrToHdr(unsigned ldr) { return m_hdrValues[ldr]; }
        const float &ldrToHdr(unsigned ldr) const { return m_hdrValues[ldr]; }

        unsigned ldrFromHdr(float hdr) const;
    protected:
        float m_hdrValues[numLdrValues];
};

extern template class CameraResponseCurve<256>;

template<unsigned numLdrValues>
class CameraRGBResponseCurves
{
    public:
        CameraResponseCurve<numLdrValues> r;
        CameraResponseCurve<numLdrValues> g;
        CameraResponseCurve<numLdrValues> b;
        
        static CameraRGBResponseCurves<numLdrValues> readFromXML(tinyxml2::XMLElement *node); // these only work for extern templates
        static CameraRGBResponseCurves<numLdrValues> readFromXML(const boost::filesystem::path &filename);
        void writeToXML(tinyxml2::XMLElement *node) const;
        void writeToXML(const boost::filesystem::path &filename) const;        
};

extern template class CameraRGBResponseCurves<256>;


template<unsigned numLdrValues>
unsigned CameraResponseCurve<numLdrValues>::ldrFromHdr(float hdr) const
{
    int min = 0;
    int max = numLdrValues-1;
    while (min < max) {
        int center = (max + min)/2;
        float centerValue = m_hdrValues[center];
        if (hdr < centerValue)
            max = center-1;
        else
            min = center+1;
    }
    return std::max<int>(std::min<int>(min, numLdrValues-1), 0);
}

template<unsigned ldrRange>
void computeCameraResponseFromMultipleExposures(const std::vector<unsigned> &values, const std::vector<float> &exposures, float smoothness, CameraResponseCurve<ldrRange> &curve)
{
    assert(values.size() % exposures.size() == 0);

    auto computeWeight = [](unsigned v)->float{
        if (v == 0) return 0.0f;
        if (v == ldrRange-1) return 0.0f;
        return 1.0;
    };

    const unsigned numMeasurements = values.size();
    const unsigned numPixels = values.size() / exposures.size();
    const unsigned numCurvatures = ldrRange-2;
    Eigen::MatrixXd designMatrix(numMeasurements + numCurvatures + 1, numPixels + ldrRange);
    designMatrix.setZero();
    Eigen::VectorXd designVector(numMeasurements + numCurvatures + 1);
    designVector.setZero();
    for (unsigned i = 0; i < numMeasurements; i++) {
        unsigned pixel = i % numPixels;
        unsigned image = i / numPixels;

        float weight = computeWeight(values[image * numPixels + pixel]);
        designVector[i] = std::log(exposures[image]) * weight;

        designMatrix(i, pixel) = -1.0f * weight;
        designMatrix(i, numPixels + values[image * numPixels + pixel]) = 1.0f * weight;
    }

    for (unsigned i = 0; i < numCurvatures; i++) {
        float weight = computeWeight(i+1);
        designMatrix(numMeasurements+i, numPixels + i) = 1.0f * smoothness * weight;
        designMatrix(numMeasurements+i, numPixels + i+1) = -2.0f * smoothness * weight;
        designMatrix(numMeasurements+i, numPixels + i+2) = 1.0f * smoothness * weight;
        designVector[numMeasurements+i] = 0.0f * smoothness * weight;
    }

    designMatrix(numMeasurements+numCurvatures, numPixels + ldrRange/2) = 1.0f;
    designVector[numMeasurements+numCurvatures] = std::log(1.0f);

    Eigen::VectorXd solution = designMatrix.jacobiSvd(Eigen::ComputeThinU | Eigen::ComputeThinV).solve(designVector);

    for (unsigned i = 0; i < ldrRange; i++)
        curve.ldrToHdr(i) = std::exp(solution[numPixels+i]);
}

#ifdef SYB3R_HAS_EXIV2

void computeCameraResponseFromMultipleExposures(const std::vector<std::string> &jpegFilenames, unsigned numPixels, float smoothness,
                                                CameraResponseCurve<256> &redCurve, CameraResponseCurve<256> &greenCurve, CameraResponseCurve<256> &blueCurve);

#endif

///@todo Move this somewhere

extern syb3r::models::CameraRGBResponseCurves<256> curves_Eos400D_neutral_5200K_daylight;
extern syb3r::models::CameraRGBResponseCurves<256> curves_Eos400D_neutral_3200K_artificial;
extern syb3r::models::CameraRGBResponseCurves<256> curves_Eos400D_neutral_7000K_shade;

namespace Eos400D_neutral_3200K_artificial {
extern syb3r::models::CameraResponseCurve<256> redCurve;
extern syb3r::models::CameraResponseCurve<256> greenCurve;
extern syb3r::models::CameraResponseCurve<256> blueCurve;
}

namespace Eos400D_neutral_7000K_shade {
extern syb3r::models::CameraResponseCurve<256> redCurve;
extern syb3r::models::CameraResponseCurve<256> greenCurve;
extern syb3r::models::CameraResponseCurve<256> blueCurve;
}


}
}


#endif // CAMERARESPONSECURVE_H
