/*
 * Syb3r - A framework for the creation and use of synthetic benchmark datasets for SfM/MVS with Blender and Cycles.
 * Copyright (C) 2016 Andreas Ley <mail@andreas-ley.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#ifndef FLOATIMAGE_H
#define FLOATIMAGE_H

#include <Eigen/Dense>
#include <vector>

namespace syb3r {
namespace tools {

/**
 * @brief Stores an image with arbitrary, usually float based, pixel values.
 * @details If instantiated with Eigen::Vector3f, this class features the most basic tools for manipulating RGB32F images.
 */
template<typename Type>
class FloatImage
{
    public:
        /// @brief Allocate memory for w times h pixels
        /// @details The content does not get rescaled.
        void resize(unsigned w, unsigned h) {
            m_width = w;
            m_height = h;
            m_data.resize(m_width * m_height);
        }

        /// Returns number or rows in the image
        unsigned getHeight() const { return m_height; }
        /// Returns number of columns in the image
        unsigned getWidth() const { return m_width; }

        /// Direct access to the row major data
        Type &operator[](unsigned idx) { return m_data[idx]; }
        /// Direct access to the row major data
        const Type &operator[](unsigned idx) const { return m_data[idx]; }

        /// Access to a single pixel
        Type &operator()(unsigned x, unsigned y) { return m_data[x + y * m_width]; }
        /// Access to a single pixel
        const Type &operator()(unsigned x, unsigned y) const { return m_data[x + y * m_width]; }

        /// Helper function for binary operations
        template<typename OtherType, class BinaryOp>
        const FloatImage<Type>& performBinaryOp(const FloatImage<OtherType> &rhs, BinaryOp op) {
            assert(rhs.getWidth() == m_width);
            assert(rhs.getHeight() == m_height);
            for (unsigned i = 0; i < m_width*m_height; i++)
                m_data[i] = op(m_data[i], rhs.m_data[i]);
            return *this;
        }

        /// Helper function for unary operations
        template<class UnaryOp>
        const FloatImage<Type>& performUnaryOp(UnaryOp op) {
            for (unsigned i = 0; i < m_width*m_height; i++)
                m_data[i] = op(m_data[i]);
            return *this;
        }

        template<typename OtherType>
        const FloatImage<Type>& operator-=(const FloatImage<OtherType> &rhs) {
            return performBinaryOp(rhs, [](const Type &a, const OtherType &b)->Type { return a-b; });
        }

        template<typename OtherType>
        const FloatImage<Type>& operator+=(const FloatImage<OtherType> &rhs) {
            return performBinaryOp(rhs, [](const Type &a, const OtherType &b)->Type { return a+b; });
        }

        template<typename OtherType>
        const FloatImage<Type>& operator*=(const FloatImage<OtherType> &rhs) {
            return performBinaryOp(rhs, [](const Type &a, const OtherType &b)->Type { return a*b; });
        }

        template<typename OtherType>
        const FloatImage<Type>& operator/=(const FloatImage<OtherType> &rhs) {
            return performBinaryOp(rhs, [](const Type &a, const OtherType &b)->Type { return a/b; });
        }

        template<typename OtherType>
        const FloatImage<Type>& operator*=(const OtherType &factor) {
            return performUnaryOp([&factor](const Type &a)->Type { return a*factor; });
        }

        template<typename OtherType>
        const FloatImage<Type>& operator+=(const OtherType &factor) {
            return performUnaryOp([&factor](const Type &a)->Type { return a+factor; });
        }

        /// @brief Copys every n-th row and every n-th column
        /// @details No box filtering or anything, just sub sampling.
        void subsample(const FloatImage<Type> &other, unsigned stride) {
            resize(other.getWidth() / stride, other.getHeight() / stride);
            for (unsigned y = 0; y < m_height; y++)
                for (unsigned x = 0; x < m_width; x++)
                    m_data[x + y * m_width] = other(x*stride, y*stride);
        }

        /// @brief Samples from a float coordinate, using bilinear interpolation
        /// @note Keep in mind that the first pixel spans the x/y range of 0..1 and thus has it's center at 0.5, not at zero!
        Type sampleLinear(float x, float y) const {
            x = std::min(std::max(x - 0.5f, 0.0f), m_width - 1.001f);
            y = std::min(std::max(y - 0.5f, 0.0f), m_height - 1.001f);

            int ix = x;
            float fracX = x-ix;
            int iy = y;
            float fracY = y-iy;

            const Type &v00 = (*this)(ix+0, iy+0);
            const Type &v01 = (*this)(ix+0, iy+1);
            const Type &v10 = (*this)(ix+1, iy+0);
            const Type &v11 = (*this)(ix+1, iy+1);

            return (v00 * (1.0f - fracX) + v10 * fracX) * (1.0f - fracY) +
                   (v01 * (1.0f - fracX) + v11 * fracX) * fracY;
        }

        /// @brief Samples a single channel from a float coordinate, using bilinear interpolation
        /// @note Keep in mind that the first pixel spans the x/y range of 0..1 and thus has it's center at 0.5, not at zero!
        template <typename scalarType>
        scalarType sampleLinearChannel(float x, float y, unsigned channel) const {
            x = std::min(std::max(x - 0.5f, 0.0f), m_width - 1.001f);
            y = std::min(std::max(y - 0.5f, 0.0f), m_height - 1.001f);

            int ix = x;
            float fracX = x-ix;
            int iy = y;
            float fracY = y-iy;

            const scalarType &v00 = (*this)(ix+0, iy+0)[channel];
            const scalarType &v01 = (*this)(ix+0, iy+1)[channel];
            const scalarType &v10 = (*this)(ix+1, iy+0)[channel];
            const scalarType &v11 = (*this)(ix+1, iy+1)[channel];

            return (v00 * (1.0f - fracX) + v10 * fracX) * (1.0f - fracY) +
                   (v01 * (1.0f - fracX) + v11 * fracX) * fracY;
        }

        void swap(FloatImage<Type> &other) {
            std::swap(m_width, other.m_width);
            std::swap(m_height, other.m_height);
            std::swap(m_data, other.m_data);
        }
    protected:
        unsigned m_width = 0;
        unsigned m_height = 0;
        std::vector<Type> m_data;
};

/// Returns the sum over all pixels, divided by the pixel count
template<typename Type>
Type computeAverage(const FloatImage<Type> &img)
{
    Type totalSum = Type(0);
    for (unsigned y = 0; y < img.getHeight(); y++) {
        Type rowSum = Type(0);
        for (unsigned x = 0; x < img.getWidth(); x++)
            rowSum += img(x, y);
        totalSum += rowSum;
    }
    return totalSum / (img.getWidth() * img.getHeight());
}

///@brief Convolutes an image horizontally with a symmetric kernel
///@details The kernel-vector only contains one side of the kernel as it is assumed to be symmetric.
template<typename Type, typename KernelType>
void blurHorizontal(FloatImage<Type> &dst, const FloatImage<Type> &src, const std::vector<KernelType> &kernel)
{
    dst.resize(src.getWidth(), src.getHeight());
    for (unsigned y = 0; y < src.getHeight(); y++) {
        for (unsigned x = 0; x < src.getWidth(); x++) {
            Type sum = src(x, y) * kernel[0];
            for (unsigned i = 1; i < kernel.size(); i++) {
                int x0 = std::max<int>(0, (int)x - (int)i);
                int x1 = std::min<int>(src.getWidth()-1, (int)x + (int)i);

                sum += (src(x0, y) + src(x1, y)) * kernel[i];
            }
            dst(x, y) = sum;
        }
    }
}

///@brief Convolutes an image vertically with a symmetric kernel
///@details The kernel-vector only contains one side of the kernel as it is assumed to be symmetric.
template<typename Type, typename KernelType>
void blurVertical(FloatImage<Type> &dst, const FloatImage<Type> &src, const std::vector<KernelType> &kernel)
{
    dst.resize(src.getWidth(), src.getHeight());
    for (unsigned y = 0; y < src.getHeight(); y++) {
        for (unsigned x = 0; x < src.getWidth(); x++) {
            Type sum = src(x, y) * kernel[0];
            for (unsigned i = 1; i < kernel.size(); i++) {
                int y0 = std::max<int>(0, (int)y - (int)i);
                int y1 = std::min<int>(src.getHeight()-1, (int)y + (int)i);

                sum += (src(x, y0) + src(x, y1)) * kernel[i];
            }
            dst(x, y) = sum;
        }
    }
}


/// Normalizes a symmetric kernel to sum to one
void normalizeFilterKernel(std::vector<float> &kernel);
/// Normalizes a symmetric kernel so that the square of it's coefficients sum to one
void normalizeFilterKernelEnergy(std::vector<float> &kernel);



extern template class FloatImage<Eigen::Vector3f>;
extern template class FloatImage<float>;

}
}

#endif // FLOATIMAGE_H
