#include "WrapperDFMSubsBA.h"

#include <sstream>
#include <iostream>

#include <tinyxml2.h>
#include <stdexcept>

#include <boost/filesystem.hpp>
#include <Eigen/Dense>

#include <syb3r/data/Camera.h>
#include <syb3r/data/EstimatedCameraPath.h>

namespace syb3r {
namespace wrapper {

bool WrapperDFMSubsBA::runSFM(float focalLengthEstimate, const std::pair<unsigned, unsigned> &initialPair, const std::vector<std::string> &images, data::EstimatedCameraPath &estimatedCameraPath)
{
    std::stringstream command;

    command <<  executable << " --initial-internal-calib-FocalLength " << focalLengthEstimate;
    if (!radialDistortion)
        command << " --param-BA-S-radialDistortionType NoRadialDistortion";// --param-BA-S-internalCalibrationType SharedFocalLength --param-FE-minimalDoGThreshold 0.001";

    if (!sharedFocalLength)
        command << " --param-BA-S-internalCalibrationType SharedFocalLength";// --param-FE-minimalDoGThreshold 0.001";

    if (!xmlOutputFilename.empty())
        command << " --save-xml " << xmlOutputFilename;

    if (!pmvsOutputFolder.empty())
        command << " --save-pmvs " << pmvsOutputFolder;

//command << " --param-FE-minimalDoGThreshold 0.002 --print-calib";

    command << " --initial-pair " << initialPair.first << ' ' << initialPair.second << " --filenames";
    for (const auto &img : images)
        command << ' ' << img;

    bool success = run(command.str().c_str());

    if (success)
        if (!xmlOutputFilename.empty())
            parseCamerasFromXMLFile(images, xmlOutputFilename, estimatedCameraPath);

    return success;
}

void WrapperDFMSubsBA::parseLine(const char *line)
{
    std::cout << "SFM: " << line;
}

void WrapperDFMSubsBA::parseCamerasFromXMLFile(const std::vector<std::string> &images, const std::string &xmlFilename, data::EstimatedCameraPath &estimatedCameraPath)
{
    tinyxml2::XMLDocument doc;

    tinyxml2::XMLDocument xmlDoc;
    if (xmlDoc.LoadFile(xmlFilename.c_str()) != tinyxml2::XML_NO_ERROR)
        throw std::runtime_error("Could not open or parse xml file!");

    tinyxml2::XMLElement* root = xmlDoc.FirstChildElement("scan");
    if (root == nullptr) throw std::runtime_error("No 'scan' node in xml file!");

    tinyxml2::XMLElement* clipNode = root->FirstChildElement("video_clip");
    if (clipNode == nullptr) throw std::runtime_error("No 'video_clip' node in xml file!");

    for (tinyxml2::XMLElement* frameNode = clipNode->FirstChildElement("frame"); frameNode != nullptr; frameNode = frameNode->NextSiblingElement("frame")) {

        tinyxml2::XMLElement* imageNode = frameNode->FirstChildElement("image");
        if (imageNode == nullptr) throw std::runtime_error("No 'image' node in frame node!");

        const char *filename = imageNode->Attribute("filename");

        std::string filenameWOPath = boost::filesystem::path(filename).filename().string();
        unsigned srcIndex = ~0u;
        for (unsigned j = 0; j < images.size(); j++)
            if (filenameWOPath == boost::filesystem::path(images[j]).filename().string()) {
                srcIndex = j;
                break;
            }

        if (srcIndex == ~0u)
            throw std::runtime_error("One xml filename does not match any of the files given to vsfm...!");


        unsigned width, height;
        if (imageNode->QueryUnsignedAttribute("width", &width) != tinyxml2::XML_SUCCESS)
            throw std::runtime_error("No or invalid 'width' attrib in image node!");
        if (imageNode->QueryUnsignedAttribute("height", &height) != tinyxml2::XML_SUCCESS)
            throw std::runtime_error("No or invalid 'height' attrib in image node!");


        tinyxml2::XMLElement* poseNode = frameNode->FirstChildElement("pose");
        if (poseNode == nullptr) throw std::runtime_error("No 'pose' node in frame node!");

        const char *poseString = poseNode->Attribute("projection_matrix");
        if (poseString == nullptr)
            throw std::runtime_error("No 'projection_matrix' attrib in pose node!");

        std::stringstream poseStream(poseString);
        Eigen::Matrix<float, 3, 4> DFMSubsBA_P;
        for (unsigned i = 0; i < 3; i++)
            for (unsigned j = 0; j < 4; j++)
                poseStream >> DFMSubsBA_P(i, j);

        data::Camera camera = data::Camera::convertFrom_DFM_SubsBA(DFMSubsBA_P, width, height);
        estimatedCameraPath.add(camera, srcIndex);
    }
}


}
}