#include "WrapperPMVS.h"

#include <fstream>
#include <thread>

#include <boost/format.hpp>

#include <syb3r/data/Camera.h>
#include <syb3r/data/Pointcloud.h>
#include <syb3r/data/PlyPointcloudFile.h>

#include <syb3r/tools/TmpDir.h>

#include <iostream>

namespace syb3r {
namespace wrapper {

bool WrapperPMVS::runPreparedSet(const std::string &basePath, const std::string &optionsFilename, std::vector<Eigen::Vector3f> &pcld, std::vector<Eigen::Vector3f> *colors)
{
    if (generateOptionsFile) {
        std::fstream file;
        file.exceptions(std::ifstream::failbit | std::ifstream::badbit);
        file.open(basePath+'/'+optionsFilename, std::fstream::out);

        file
            << "level " << mipLevel << std::endl
            << "csize 2" << std::endl
            << "threshold 0.7" << std::endl
            << "wsize 7" << std::endl
            << "minImageNum 3" << std::endl
            << "CPU " << std::thread::hardware_concurrency() << std::endl
            << "setEdge 0" << std::endl
            << "useBound 0" << std::endl
            << "useVisData 0" << std::endl
            << "sequence -1" << std::endl
            << "maxAngle 10" << std::endl
            << "quad 2.0" << std::endl
            << "timages -1 0 " << *numImages << std::endl
            << "oimages 0" << std::endl;
    }

    std::stringstream command;

    command << executable << " " << basePath << "/ " << optionsFilename;

    bool success = run(command.str().c_str());

    if (success) {
        data::Pointcloud pointcloud;
        data::importPLYPointcloudFile(pointcloud, basePath+"/models/"+optionsFilename+".ply");
        pcld.resize(pointcloud.getCount());
        for (unsigned i = 0; i < pointcloud.getCount(); i++)
            pcld[i] = pointcloud.getPosition(i);

        if (colors != nullptr) {
            colors->resize(pointcloud.getCount());
            for (unsigned i = 0; i < pointcloud.getCount(); i++)
                (*colors)[i] = pointcloud.getColor(i);
        }
    }

    return success;
}

bool WrapperPMVS::runUnpreparedSet(const std::vector<data::Camera> &calibration, const std::vector<std::string> &images, std::vector<Eigen::Vector3f> &pcld, std::vector<Eigen::Vector3f> *colors)
{
    if (calibration.size() != images.size())
        throw std::runtime_error("calibration.size() != images.size()");

    tools::TmpDir tmpDir;
    std::cout << "tmp dir: " << tmpDir.getPath() << std::endl;

    boost::filesystem::create_directory(tmpDir.getPath() / "models");
    boost::filesystem::create_directory(tmpDir.getPath() / "txt");
    boost::filesystem::create_directory(tmpDir.getPath() / "visualize");

    for (unsigned i = 0; i < images.size(); i++) {
        boost::filesystem::path srcPath(images[i]);
        boost::filesystem::create_symlink(srcPath, tmpDir.getPath() / "visualize" / (boost::format("%08d%s") % i % srcPath.extension().string()).str());
    }

    for (unsigned i = 0; i < calibration.size(); i++) {
        std::fstream file;
        file.exceptions(std::ifstream::failbit | std::ifstream::badbit);
        file.open((boost::format("%s/txt/%08d.txt") % tmpDir.getPath().string() % i).str(), std::fstream::out);

        Eigen::Matrix<float, 3, 4> pmvsMat = calibration[i].convertTo_PMVS_Matrix();

        file
            << "CONTOUR" << std::endl
            << pmvsMat(0, 0) << " " << pmvsMat(0, 1) << " " << pmvsMat(0, 2) << " " << pmvsMat(0, 3) << std::endl
            << pmvsMat(1, 0) << " " << pmvsMat(1, 1) << " " << pmvsMat(1, 2) << " " << pmvsMat(1, 3) << std::endl
            << pmvsMat(2, 0) << " " << pmvsMat(2, 1) << " " << pmvsMat(2, 2) << " " << pmvsMat(2, 3) << std::endl;
    }

    std::fstream file;
    file.exceptions(std::ifstream::failbit | std::ifstream::badbit);
    file.open((boost::format("%s/options.txt") % tmpDir.getPath().string()).str(), std::fstream::out);

    file
        << "level " << mipLevel << std::endl
        << "csize 2" << std::endl
        << "threshold 0.7" << std::endl
        << "wsize 7" << std::endl
        << "minImageNum 3" << std::endl
        << "CPU " << std::thread::hardware_concurrency() << std::endl
        << "setEdge 0" << std::endl
        << "useBound 0" << std::endl
        << "useVisData 0" << std::endl
        << "sequence -1" << std::endl
        << "maxAngle 10" << std::endl
        << "quad 2.0" << std::endl
        << "timages -1 0 " << images.size() << std::endl
        << "oimages 0" << std::endl;

    std::stringstream command;

    command << executable << " " << tmpDir.getPath().string() << "/ " << "options.txt";

    bool success = run(command.str().c_str());

    if (success) {
        data::Pointcloud pointcloud;
        data::importPLYPointcloudFile(pointcloud, (tmpDir.getPath() / "models" / "options.txt.ply").string());
        pcld.resize(pointcloud.getCount());
        for (unsigned i = 0; i < pointcloud.getCount(); i++)
            pcld[i] = pointcloud.getPosition(i);
        if (colors != nullptr) {
            colors->resize(pointcloud.getCount());
            for (unsigned i = 0; i < pointcloud.getCount(); i++)
                (*colors)[i] = pointcloud.getColor(i);
        }
    }

    return success;

}


void WrapperPMVS::parseLine(const char *line)
{
    std::cout << "PMVS: " << line;
}

}
}
