/*
 * Syb3r - A framework for the creation and use of synthetic benchmark datasets for SfM/MVS with Blender and Cycles.
 * Copyright (C) 2016 Andreas Ley <mail@andreas-ley.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include "ImagePP_LinearBlur.h"

#include "ImagePostprocessor.h"

#include <syb3r/models/CameraShake.h>
#include <stdexcept>

namespace syb3r {
namespace synthesis {

ImagePP_LinearBlur::ImagePP_LinearBlur()
{
    exposureTimeToCameraShake.insert(std::make_pair(1.0f/50.0f, models::Eos400D_Andy_Standing_noCoffee_linear::cameraShake1_50th));
    exposureTimeToCameraShake.insert(std::make_pair(1.0f/10.0f, models::Eos400D_Andy_Standing_noCoffee_linear::cameraShake1_10th));
}


void ImagePP_LinearBlur::process(tools::FloatImage<Eigen::Vector3f> &output, const tools::FloatImage<Eigen::Vector3f> &input, ImageProperties &properties) const
{
    const float exposureTime = properties.getFloat("exposureTime");
    if (exposureTime != 0.0f) {
        output.resize(input.getWidth(), input.getHeight());
        std::gamma_distribution<float> blurLengthDist;
        
        auto it = exposureTimeToCameraShake.find(exposureTime);
        if (it != exposureTimeToCameraShake.end())
            blurLengthDist = it->second.getMotionBlurInPixels(exposureTime,
                                                                input.getWidth(),
                                                                properties.getFloat("sensorWidth"),
                                                                properties.getFloat("focalLength"));
        else
            throw std::runtime_error("Camera shake model for this exposure time not available");

        float blurLength;

        std::uniform_real_distribution<float> blurAngleDist(0.0f, M_PI);
        float blurAngle;
        {
            std::lock_guard<std::mutex> lock(m_rngMutex);
            blurLength = blurLengthDist(m_rng);
            blurAngle = blurAngleDist(m_rng);
        }

        float cosAngle = std::cos(blurAngle);
        float sinAngle = std::sin(blurAngle);

        float blurLengthHalf = blurLength * 0.5f;
        int countHalf = blurLengthHalf * 1.3f;

        for (unsigned y = 0; y < input.getHeight(); y++) {
            for (unsigned x = 0; x < input.getWidth(); x++) {
                Eigen::Vector3f sum;
                sum.setZero();

                for (int i = -countHalf; i <= countHalf; i++) {
                    float f = countHalf>0?i * blurLengthHalf / countHalf:0.0f;
                    float fx = x + 0.5f + cosAngle * f;
                    float fy = y + 0.5f + sinAngle * f;
                    Eigen::Vector3f sample = input.sampleLinear(fx, fy);
                    sum += sample;
                }
                sum /= (countHalf*2+1);
                output(x, y) = sum;
            }
        }
    } else
        output = input;
}

}
}