/*
 * Syb3r - A framework for the creation and use of synthetic benchmark datasets for SfM/MVS with Blender and Cycles.
 * Copyright (C) 2016 Andreas Ley <mail@andreas-ley.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include "ImagePP_RadialDistortion.h"

#include "ImagePostprocessor.h"

namespace syb3r {
namespace synthesis {

ImagePP_RadialDistortion::ImagePP_RadialDistortion()
{
}

void ImagePP_RadialDistortion::process(tools::FloatImage<Eigen::Vector3f> &output, const tools::FloatImage<Eigen::Vector3f> &input, ImageProperties &properties) const
{
    const Eigen::Vector3f &radialDistortionKappa2 = properties.getVec3f("radialDistortionKappa2");
    const Eigen::Vector3f &radialDistortionKappa3 = properties.getVec3f("radialDistortionKappa3");
    const Eigen::Vector3f &radialDistortionKappa4 = properties.getVec3f("radialDistortionKappa4");

    const float projectionCenterX = properties.getFloat("projectionCenterX");
    const float projectionCenterY = properties.getFloat("projectionCenterY");

    output.resize(input.getWidth(), input.getHeight());

    for (unsigned y = 0; y < input.getHeight(); y++) {
        for (unsigned x = 0; x < input.getWidth(); x++) {
            const float fx = (x + 0.5f - projectionCenterX) / input.getWidth();
            const float fy = (y + 0.5f - projectionCenterY) / input.getWidth();

            const float r2 = fx*fx + fy*fy;
            const float r3 = r2 * std::sqrt(r2);
            const float r4 = r2*r2;

            Eigen::Vector3f scale = Eigen::Vector3f(1.0f, 1.0f, 1.0f) + radialDistortionKappa2 * r2 + radialDistortionKappa3 * r3 + radialDistortionKappa4 * r4;

            for (unsigned c = 0; c < 3; c++)
                output(x, y)[c] = input.sampleLinearChannel<float>(
                            projectionCenterX + fx * scale[c] * input.getWidth(),
                            projectionCenterY + fy * scale[c] * input.getWidth(),
                            c
                );
        }
    }
}

}
}
