#include "ImagePostprocessor.h"

#include "ImagePostprocessorStep.h"
#include "ImagePP_ColorMatrix.h"
#include "ImagePP_DemosaicSimple.h"
#include "ImagePP_LinearToGamma.h"
#include "ImagePP_SensorNoise.h"
#include "ImagePP_OldSensorNoise.h"
#include "ImagePP_TonemappingCurve.h"
#include "ImagePP_LinearBlur.h"
#include "ImagePP_AutoExposure.h"
#include "ImagePP_RadialDistortion.h"

#include <OpenEXR/ImfInputFile.h>
#include <OpenEXR/ImfArray.h>
#include <OpenEXR/ImathBox.h>

#include <boost/format.hpp>

#include <syb3r/models/CameraResponseCurve.h>

#include <syb3r/data/Dataset.h>
#include <syb3r/data/Camera.h>

#include <syb3r/tools/FloatImage.h>
#include <syb3r/tools/TaskScheduler.h>
#include <syb3r/tools/RasterImage.h>

#include <Magick++.h>

#include <mutex>

namespace syb3r {
namespace synthesis {

ImagePostprocessor::ImagePostprocessor()
{
    initialProperties.set("exposureTime", 0.0f);
    initialProperties.set("sensorWidth", 26.0f/1.6f);
    initialProperties.set("focalLength", 50.0f);
    initialProperties.set("iso", 0);

    initialProperties.set("autoExposure", true);

    /// @todo read these from the camera!
    initialProperties.set("radialDistortionKappa2", Eigen::Vector3f(0.0f, 0.0f, 0.0f));
    initialProperties.set("radialDistortionKappa3", Eigen::Vector3f(0.0f, 0.0f, 0.0f));
    initialProperties.set("radialDistortionKappa4", Eigen::Vector3f(0.0f, 0.0f, 0.0f));

    initialProperties.set("jpegCompression", 95);
}

// Dtor must be explicit because of the use of forward declared classes in unique_ptr instances.
ImagePostprocessor::~ImagePostprocessor()
{
}

void ImagePostprocessor::appendStep(ImagePostprocessorStep *step)
{
    m_ppSteps.push_back(std::unique_ptr<ImagePostprocessorStep>(step));
    if (m_ppSteps.size() > 1) {
        if (m_ppSteps.size() > 2) {
            m_ppSteps[m_ppSteps.size()-1]->link(m_ppSteps[m_ppSteps.size()-2].get(), nullptr);
            m_ppSteps[m_ppSteps.size()-2]->link(m_ppSteps[m_ppSteps.size()-3].get(), m_ppSteps[m_ppSteps.size()-1].get());
        } else {
            m_ppSteps[1]->link(m_ppSteps[0].get(), nullptr);
            m_ppSteps[0]->link(nullptr, m_ppSteps[1].get());
        }
    } else {
        m_ppSteps[0]->link(nullptr, nullptr);
    }
}

/**
 * @param inputImage The image to process
 * @param output The filename where to store the resulting jpg image
 */
void ImagePostprocessor::process(const data::Image &inputImage, const std::string &output) const
{
    Imf::InputFile inputFile(inputImage.getHDRFilename().string().c_str());

    tools::FloatImage<Eigen::Vector3f> image, tmp;

    Imath::Box2i dw = inputFile.header().dataWindow();
    const unsigned width = dw.max.x - dw.min.x + 1;
    const unsigned height = dw.max.y - dw.min.y + 1;
    const int dx = dw.min.x;
    const int dy = dw.min.y;


    image.resize(width, height);
    tmp.resize(width, height);

    Imf::FrameBuffer frameBuffer;
    frameBuffer.insert("HDR.R", Imf::Slice(Imf::FLOAT, (char *) &image(-dx, -dy)[0], sizeof(Eigen::Vector3f), sizeof(Eigen::Vector3f) * width));
    frameBuffer.insert("HDR.G", Imf::Slice(Imf::FLOAT, (char *) &image(-dx, -dy)[1], sizeof(Eigen::Vector3f), sizeof(Eigen::Vector3f) * width));
    frameBuffer.insert("HDR.B", Imf::Slice(Imf::FLOAT, (char *) &image(-dx, -dy)[2], sizeof(Eigen::Vector3f), sizeof(Eigen::Vector3f) * width));

    inputFile.setFrameBuffer (frameBuffer);
    inputFile.readPixels (dw.min.y, dw.max.y);


    ImageProperties properties = initialProperties;
    properties.set("projectionCenterX", inputImage.getCamera().getInternalCalibration()(0, 2));
    properties.set("projectionCenterY", inputImage.getCamera().getInternalCalibration()(1, 2));


    for (const auto &ppStep : m_ppSteps) {
        ppStep->process(tmp, image, properties);
        image.swap(tmp);
    }

    std::vector<std::array<unsigned char, 4>> ldrImage;
    ldrImage.resize(width * height);
    for (unsigned y = 0; y < height; y++) {
        for (unsigned x = 0; x < width; x++) {
            float r = image(x, y)[0];
            float g = image(x, y)[1];
            float b = image(x, y)[2];

            int sRGB[3];
            sRGB[0] = std::min<int>(std::max<int>(r * 256.0f, 0), 255);
            sRGB[1] = std::min<int>(std::max<int>(g * 256.0f, 0), 255);
            sRGB[2] = std::min<int>(std::max<int>(b * 256.0f, 0), 255);

            ldrImage[x+y*width][0] = sRGB[0];
            ldrImage[x+y*width][1] = sRGB[1];
            ldrImage[x+y*width][2] = sRGB[2];
            ldrImage[x+y*width][3] = 0xFF;
        }
    }

    Magick::Image outputFile(width, height, "RGBA", Magick::CharPixel, &ldrImage[0][0]);
    outputFile.quality(initialProperties.getInt("jpegCompression"));
    outputFile.write(output);
}

/**
 * @param dataset All images of this dataset will be processed
 * @param outputPath Path to a directory (without trailing slash) where the images are to be stored.
 * @param outputFilenames Vector of strings in which the function will store the filenames of the processed images.
 *                          They can be related to the images in the dataset by their indices.
 */
void ImagePostprocessor::process(const data::Dataset &dataset, const std::string &outputPath, std::vector<std::string> &outputFilenames) const
{
    outputFilenames.resize(dataset.getImages().size());
    for (unsigned i = 0; i < dataset.getImages().size(); i++)
        outputFilenames[i] = (boost::format("%s/%08d.jpg") % outputPath % i).str();

    std::vector<std::string> errors;
    std::mutex errorMutex;

    tools::TaskGroup group;
    for (unsigned i = 0; i < dataset.getImages().size(); i++) {
        group.add([this, i, &dataset, &outputFilenames, &errors, &errorMutex]{
            try {
                process(dataset.getImages()[i], outputFilenames[i]);
            } catch (const std::exception &e) {
                std::lock_guard<std::mutex> lock(errorMutex);
                errors.push_back(e.what());
            }
        }, tools::TaskScheduler::get());
    }
    tools::TaskScheduler::get().waitFor(&group);

    if (!errors.empty()) {
        std::cout << "There were errors:" << std::endl;
        for (const std::string &s : errors)
            std::cout << s << std::endl;

        throw std::runtime_error("There were errors in the image post processing step!");
    }
}


void ImagePostprocessor::setupDefaultEOS400()
{
    appendStep(new ImagePP_LinearBlur());
    appendStep(new ImagePP_RadialDistortion());

    Eigen::Matrix3f eos400ToRgb = ImagePP_ColorMatrix::XYZ_to_sRGB * ImagePP_ColorMatrix::CanonEOS400_to_XYZ;
    appendStep(new ImagePP_ColorMatrix(eos400ToRgb.inverse()));

    appendStep(new ImagePP_SensorNoise(0.002f, 0.025f));
    appendStep(new ImagePP_DemosaicSimple());

    appendStep(new ImagePP_ColorMatrix(eos400ToRgb));
    appendStep(new ImagePP_LinearToGamma());
}

void ImagePostprocessor::setupOldEOS400()
{
    appendStep(new ImagePP_LinearBlur());
    appendStep(new ImagePP_RadialDistortion());
    appendStep(new ImagePP_AutoExposure(true));
    appendStep(new ImagePP_OldSensorNoise());
    appendStep(new ImagePP_TonemappingCurve(syb3r::models::curves_Eos400D_neutral_5200K_daylight));
}

}
}
