#ifndef IMAGEPOSTPROCESSOR_H
#define IMAGEPOSTPROCESSOR_H

#include <memory>

#include <string>
#include <Eigen/Dense>
#include <map>
#include <string>

namespace syb3r {

namespace data {
    class Dataset;
    class Image;
}

namespace synthesis {

class ImagePostprocessorStep;

/**
 * @brief Stores basic information in key/value pairs.
 */
class ImageProperties
{
    public:
        inline void set(std::string key, const Eigen::Vector3f &value) {
            m_vec3fAttribs[std::move(key)] = value;
        }
        inline void set(std::string key, float value) {
            m_floatAttribs[std::move(key)] = value;
        }
        inline void set(std::string key, int value) {
            m_intAttribs[std::move(key)] = value;
        }
        inline void set(std::string key, bool value) {
            m_boolAttribs[std::move(key)] = value;
        }

        const Eigen::Vector3f &getVec3f(const std::string &key) const {
            return m_vec3fAttribs.at(key);
        }
        float getFloat(const std::string &key) const {
            return m_floatAttribs.at(key);
        }
        int getInt(const std::string &key) const {
            return m_intAttribs.at(key);
        }
        bool getBool(const std::string &key) const {
            return m_boolAttribs.at(key);
        }
    protected:
        std::map<std::string, Eigen::Vector3f> m_vec3fAttribs;
        std::map<std::string, float> m_floatAttribs;
        std::map<std::string, int> m_intAttribs;
        std::map<std::string, bool> m_boolAttribs;
};

/**
 * @brief Reads the hdr images from cycles, bakes all post processing effects into them and stores them as jpg files
 * @todo Put jpg compression into an exchangeable step as well.
 */
class ImagePostprocessor
{
    public:
        ImagePostprocessor();
        ~ImagePostprocessor();

        /// @brief The initial set of properties that the processing of each image starts with.
        /// @details The @ref ImagePostprocessorStep "post processing steps" can modify these properties to pass information to subsequent steps.
        ImageProperties initialProperties;

        /// @brief Append another @ref ImagePostprocessorStep "post processing step" to the chain.
        /// @details The ImagePostprocessor takes ownership of the instance.
        void appendStep(ImagePostprocessorStep *step);

        /// Process a single image
        void process(const data::Image &inputImage, const std::string &output) const;
        /// @brief Process all images of a dataset in a multithreaded way
        /// @details All resulting images are stored in a single directory, the filenames are passed back as a vector of strings.
        void process(const data::Dataset &dataset, const std::string &outputPath, std::vector<std::string> &outputFilenames) const;

        void setupDefaultEOS400();
        void setupOldEOS400();
    protected:
        std::vector<std::unique_ptr<ImagePostprocessorStep>> m_ppSteps;
};

}
}

#endif // IMAGEPOSTPROCESSOR_H
