#include "ScriptGenerator.h"

#include <stdexcept>

namespace syb3r {
namespace synthesis {

void ScripGenerator::generate(std::ostream &stream)
{
    if (outputDir.empty())
        throw std::runtime_error("Invalid output dir!");

    if (outputDir[outputDir.size()-1] != '/')
        outputDir += '/';

    stream
        << "import bpy, mathutils, os" << std::endl
        << "scene = bpy.context.scene" << std::endl << std::endl
        << "renderHDR = " << (renderHDR?"True":"False") << std::endl
        << "renderAttribs = " << (renderAttribs?"True":"False") << std::endl
        << "outputDir = \"" << outputDir << "\"" << std::endl << std::endl
        << "if not os.path.exists(outputDir):" << std::endl
        << "    os.makedirs(outputDir)" << std::endl;


    if (preview) {
        stream
            << "# preview mode!" << std::endl
            << "scene.render.resolution_percentage = 10" << std::endl << std::endl;
            //<< "scene.cycles.samples = scene.cycles.samples / 4" << std::endl
            //<< "scene.cycles.aa_samples = scene.cycles.aa_samples / 4" << std::endl
    }

    stream
        << "# Hooks" << std::endl
        << preRenderHook << std::endl;

    stream
        << "# Setting compute device" << std::endl;
    switch (computeDevice) {
        case COMPUTEDEVICE_GPU:
            stream << "scene.cycles.device = 'GPU'"<<std::endl
                   << "scene.render.tile_x = 256"<<std::endl
                   << "scene.render.tile_y = 256"<<std::endl<<std::endl;
        break;
        case COMPUTEDEVICE_CPU:
            stream << "scene.cycles.device = 'CPU'"<<std::endl
                   << "scene.render.tile_x = 16"<<std::endl
                   << "scene.render.tile_y = 16"<<std::endl<<std::endl;
        break;
    }

    if (cameras.empty()) {
        stream << "cameraList = [ scene.camera ]" << std::endl;
    } else {
        stream << "cameraList = [";

        for (unsigned i = 0; i < cameras.size(); i++) {
            if (i > 0)
                stream << ", ";

            stream << "bpy.data.objects['" << cameras[i] <<"']";
        }

        stream << "]" << std::endl;
    }

    stream
        << "def setOutputFilenames(scene, cameraIdx):" << std::endl
        << "    scene.node_tree.nodes['syb3r_hdr_file_output'].base_path = outputDir+\"hdr_%02d_########.exr\" % (cameraIdx)" << std::endl
        << "    scene.node_tree.nodes['syb3r_attrib_file_output'].base_path = outputDir+\"attrib_%02d_########.exr\" % (cameraIdx)" << std::endl << std::endl;

    stream
        << "hdrRenderLayer = scene.render.layers[\"syb3r_hdr_layer\"]" << std::endl
        << "attribRenderLayer = scene.render.layers[\"syb3r_attrib_layer\"]" << std::endl << std::endl

        << "# For some reason node type information (RGBA, Vector, Value==Scalar) of the compositor gets lost in the console version of blender" << std::endl
        << "# Force another propagation of type information through node updates" << std::endl
        << "for n in scene.node_tree.nodes:" << std::endl
        << "    n.update()" << std::endl << std::endl

        << "rotateAroundX = mathutils.Matrix()" << std::endl
        << "rotateAroundX[1][1] = -1.0" << std::endl
        << "rotateAroundX[2][2] = -1.0" << std::endl
        << std::endl
        << "sceneXmlSpec = open(outputDir + \"scene.xml\", 'w');" << std::endl
        << "sceneXmlSpec.write('<?xml version=\"1.0\" encoding=\"UTF-8\"?>\\n')" << std::endl
        << "sceneXmlSpec.write('<Scene version=\"1.0\">\\n')" << std::endl
        << "assert scene.unit_settings.system == 'METRIC'" << std::endl
        << "sceneXmlSpec.write('    <SceneScale units_to_meters=\"%f\"/>\\n' % scene.unit_settings.scale_length)" << std::endl
        << "sceneXmlSpec.write('    <Cameras>\\n')" << std::endl
        << "# Output camera calibration" << std::endl
        << "tmpCurrFrame = scene.frame_current" << std::endl
        << "tmpCurrCamera = scene.camera" << std::endl
        << "for cameraIndex,cameraObj in enumerate(cameraList):" << std::endl
        << "    scene.camera = cameraObj" << std::endl
        << "    camera = cameraObj.data" << std::endl << std::endl
        << "    for frameIdx in range(scene.frame_start, scene.frame_end+scene.frame_step, scene.frame_step):" << std::endl
        << "        scene.frame_set(frameIdx)" << std::endl
        << "        sceneXmlSpec.write('        <Cam frame=\"%i\" name=\"%s\">\\n' % (frameIdx, cameraObj.name))" << std::endl
        << "        sceneXmlSpec.write('            <hdr_image filename=\"hdr_%02d_%08d.exr\"/>\\n' % (cameraIndex,frameIdx))" << std::endl
        << "        sceneXmlSpec.write('            <attrib_image filename=\"attrib_%02d_%08d.exr\"/>\\n' % (cameraIndex,frameIdx))" << std::endl
        << "        assert cameraObj.scale[0] == 1.0" << std::endl
        << "        assert cameraObj.scale[1] == 1.0" << std::endl
        << "        assert cameraObj.scale[2] == 1.0" << std::endl
        << "        externalCalib = rotateAroundX * cameraObj.matrix_world.inverted()" << std::endl
        << "        sceneXmlSpec.write('            <ext_calib mat_00=\"%f\" mat_01=\"%f\" mat_02=\"%f\" mat_03=\"%f\" mat_10=\"%f\" mat_11=\"%f\" mat_12=\"%f\" mat_13=\"%f\" mat_20=\"%f\" mat_21=\"%f\" mat_22=\"%f\" mat_23=\"%f\"/>\\n' % (" << std::endl
        << "                                               externalCalib[0][0], externalCalib[0][1], externalCalib[0][2], externalCalib[0][3]," << std::endl
        << "                                               externalCalib[1][0], externalCalib[1][1], externalCalib[1][2], externalCalib[1][3]," << std::endl
        << "                                               externalCalib[2][0], externalCalib[2][1], externalCalib[2][2], externalCalib[2][3]))" << std::endl
        << "        internalCalib = mathutils.Matrix()" << std::endl
        << "        assert camera.lens_unit == 'MILLIMETERS'" << std::endl
        << "        assert scene.render.pixel_aspect_x == 1.0" << std::endl
        << "        assert scene.render.pixel_aspect_y == 1.0" << std::endl
        << "        assert camera.sensor_fit == 'HORIZONTAL'" << std::endl
        << "        focalLength = camera.lens / camera.sensor_width" << std::endl
        << "        imageWidth = scene.render.resolution_x * scene.render.resolution_percentage / 100" << std::endl
        << "        imageHeight = scene.render.resolution_y * scene.render.resolution_percentage / 100" << std::endl
        << "        internalCalib[0][0] = focalLength * imageWidth" << std::endl
        << "        internalCalib[1][1] = focalLength * imageWidth" << std::endl
        << "        internalCalib[0][1] = 0.0" << std::endl
        << "        internalCalib[0][2] = imageWidth * 0.5 - imageWidth * camera.shift_x" << std::endl
        << "        internalCalib[1][2] = imageHeight * 0.5 + imageWidth * camera.shift_y" << std::endl
        << "        sceneXmlSpec.write('            <resolution width=\"%i\" height=\"%i\"/>\\n' % (imageWidth, imageHeight))" << std::endl
        << "        sceneXmlSpec.write('            <int_calib mat_00=\"%f\" mat_01=\"%f\" mat_02=\"%f\" mat_10=\"%f\" mat_11=\"%f\" mat_12=\"%f\" mat_20=\"%f\" mat_21=\"%f\" mat_22=\"%f\"/>\\n' % (" << std::endl
        << "                                               internalCalib[0][0], internalCalib[0][1], internalCalib[0][2]," << std::endl
        << "                                               internalCalib[1][0], internalCalib[1][1], internalCalib[1][2]," << std::endl
        << "                                               internalCalib[2][0], internalCalib[2][1], internalCalib[2][2]))" << std::endl
        << "        sceneXmlSpec.write('            <sensor width_mm=\"%f\"/>\\n' % camera.sensor_width)" << std::endl
        << "        if camera.cycles.aperture_size > 0:" << std::endl
        << "            f_number = camera.lens / 1000 / (camera.cycles.aperture_size * 2 * scene.unit_settings.scale_length)" << std::endl
        << "            sceneXmlSpec.write('            <lens focal_length_mm=\"%f\" f_number=\"%f\"/>\\n' % (camera.lens, f_number))" << std::endl
        << "        else:" << std::endl
        << "            sceneXmlSpec.write('            <lens focal_length_mm=\"%f\"/>\\n' % camera.lens)" << std::endl
        << "        sceneXmlSpec.write('        </Cam>\\n')" << std::endl
        << std::endl
        << "sceneXmlSpec.write('    </Cameras>\\n')" << std::endl
        << "sceneXmlSpec.write('</Scene>\\n')" << std::endl
        << "sceneXmlSpec.close()" << std::endl
        << std::endl
        << "scene.frame_set(tmpCurrFrame)" << std::endl << std::endl

        << " # For some reason bpy.ops.render.render( write_still=False, animation=True ) does produce still images, so we provide our own renderAnim function" << std::endl
        << "def renderAnim(type):" << std::endl
        << "    tmpCurrFrame = scene.frame_current" << std::endl
        << "    tmpCurrCamera = scene.camera" << std::endl
        << "    for cameraIndex,cameraObj in enumerate(cameraList):" << std::endl
        << "        setOutputFilenames(scene, cameraIndex)" << std::endl
        << "        scene.camera = cameraObj" << std::endl
        << "        for frameIdx in range(scene.frame_start, scene.frame_end+scene.frame_step, scene.frame_step):" << std::endl
        << "            scene.frame_set(frameIdx)" << std::endl
        << "            preRenderHook(scene, type)" << std::endl
        << "            bpy.ops.render.render( write_still=False, animation=False )" << std::endl
        << "    scene.frame_set(tmpCurrFrame)" << std::endl
        << "    scene.camera = tmpCurrCamera" << std::endl
        << std::endl << std::endl

        << "def disableAllRenderLayers(scene):" << std::endl
        << "    for layer in scene.render.layers:" << std::endl
        << "        layer.use = False" << std::endl
        << std::endl << std::endl

        << "# Render full pass" << std::endl
        << "if renderHDR:" << std::endl
        << "    disableAllRenderLayers(scene)" << std::endl
        << "    hdrRenderLayer.use = True" << std::endl
        << "    renderAnim('HDR')" << std::endl << std::endl

        << "# Disable blur and render attrib pass" << std::endl
        << "if renderAttribs:" << std::endl
        << "    motionBlurEnabled = scene.render.use_motion_blur" << std::endl
        << "    scene.render.use_motion_blur = False" << std::endl
        << "    aperture = camera.cycles.aperture_size" << std::endl
        << "    camera.cycles.aperture_size = 0" << std::endl << std::endl
        << "    filterWidth = scene.cycles.filter_width" << std::endl << std::endl
        << "    scene.cycles.filter_width = 0.01 # actually not necessary" << std::endl << std::endl

        << "    disableAllRenderLayers(scene)" << std::endl
        << "    attribRenderLayer.use = True" << std::endl
        << "    renderAnim('ATTRIB')" << std::endl << std::endl

        << "    # Restore blur" << std::endl << std::endl

        << "    scene.render.use_motion_blur = motionBlurEnabled" << std::endl
        << "    camera.cycles.aperture_size = aperture" << std::endl
        << "    scene.cycles.filter_width = filterWidth" << std::endl;
}


}
}
