#include <boost/test/unit_test.hpp>

#include <boost/test/floating_point_comparison.hpp>

#include <syb3r/math/CameraProjectionMath.h>

#include <random>

#include <iomanip>
#include <iostream>


BOOST_AUTO_TEST_CASE(testCameraProjectionMath)
{
    std::default_random_engine generator(133700);
    std::uniform_real_distribution<float> distribution(0.0f, 1.0f);

    auto generateProjectionViewMatrix = [](const Eigen::Vector3f &focalPoint, const Eigen::Matrix3f &rotation, const Eigen::Matrix3f &internal)->Eigen::Matrix<float, 3, 4>{
        Eigen::Vector3f revFocalPoint = rotation * focalPoint;

        Eigen::Matrix4f external;
        external(0, 0) = rotation(0, 0);
        external(0, 1) = rotation(0, 1);
        external(0, 2) = rotation(0, 2);
        external(0, 3) = -revFocalPoint[0];

        external(1, 0) = rotation(1, 0);
        external(1, 1) = rotation(1, 1);
        external(1, 2) = rotation(1, 2);
        external(1, 3) = -revFocalPoint[1];

        external(2, 0) = rotation(2, 0);
        external(2, 1) = rotation(2, 1);
        external(2, 2) = rotation(2, 2);
        external(2, 3) = -revFocalPoint[2];

        external(3, 0) = 0.0f;
        external(3, 1) = 0.0f;
        external(3, 2) = 0.0f;
        external(3, 3) = 1.0f;

        Eigen::Matrix<float, 3, 4> P;
        P(0, 0) = 1.0f;
        P(0, 1) = 0.0f;
        P(0, 2) = 0.0f;
        P(0, 3) = 0.0f;

        P(1, 0) = 0.0f;
        P(1, 1) = 1.0f;
        P(1, 2) = 0.0f;
        P(1, 3) = 0.0f;

        P(2, 0) = 0.0f;
        P(2, 1) = 0.0f;
        P(2, 2) = 1.0f;
        P(2, 3) = 0.0f;

        return internal * P * external;
    };

    for (unsigned test = 0; test < 100; test++) {
        Eigen::Matrix3f internal;
        internal.setZero();
        internal(0, 0) = 0.8f + distribution(generator) * 0.4f;
        internal(0, 1) = -0.1f + distribution(generator) * 0.2f;
        internal(1, 1) = 0.8f + distribution(generator) * 0.4f;

        internal(0, 2) = -0.1f + distribution(generator) * 0.2f;
        internal(1, 2) = -0.1f + distribution(generator) * 0.2f;
        internal(2, 2) = 1.0f;

        Eigen::Vector3f cameraPosition(distribution(generator) * 10.0f, distribution(generator) * 10.0f, distribution(generator) * 10.0f);

        Eigen::Matrix3f rotation = Eigen::AngleAxis<float>(distribution(generator) * M_PI * 2.0f, Eigen::Vector3f(1.0f, 0.0f, 0.0f)) *
                                   Eigen::AngleAxis<float>(distribution(generator) * M_PI * 2.0f, Eigen::Vector3f(0.0f, 1.0f, 0.0f)) *
                                   Eigen::AngleAxis<float>(distribution(generator) * M_PI * 2.0f, Eigen::Vector3f(0.0f, 0.0f, 1.0f)).matrix();


        Eigen::Matrix<float, 3, 4> P = generateProjectionViewMatrix(cameraPosition, rotation, internal);

        Eigen::Vector3f computedCameraPosition1 = syb3r::math::computeFocalPoint(P);
        BOOST_CHECK((computedCameraPosition1 - cameraPosition).norm() < 1e-4f);


        Eigen::Vector3f computedCameraPosition2;
        Eigen::Matrix3f computedCameraRotation;
        Eigen::Matrix3f computedCameraInternal;
        syb3r::math::decomposeProjectionMatrix(P, computedCameraInternal, computedCameraRotation, computedCameraPosition2);
        BOOST_CHECK((computedCameraPosition2 - cameraPosition).norm() < 1e-4f);

        for (unsigned i = 0; i < 3; i++)
            for (unsigned j = 0; j < 3; j++) {
                BOOST_CHECK(std::abs(computedCameraInternal(i, j) - internal(i, j)) < 1e-5f);
            }

        for (unsigned i = 0; i < 3; i++)
            for (unsigned j = 0; j < 3; j++)
                BOOST_CHECK(std::abs(computedCameraRotation(i, j) - rotation(i, j)) < 1e-5f);



        Eigen::Matrix4f augmentedP = syb3r::math::augmentProjectionMatrix(P);
        Eigen::Vector3f randomPoint(distribution(generator) * 10.0f, distribution(generator) * 10.0f, distribution(generator) * 10.0f);
        float z = (rotation * (randomPoint - cameraPosition))[2];
        float rcpZ = 1.0f / z;

        Eigen::Vector4f projectedPoint = augmentedP * Eigen::Vector4f(randomPoint[0], randomPoint[1], randomPoint[2], 1.0f);
        projectedPoint /= projectedPoint[3];
        BOOST_CHECK(std::abs(projectedPoint[2] - rcpZ) < 1e-3f);
    }


}